@@ -4,13 +4,12 @@
 
 #include "src/api/api-natives.h"
 
-#include "src/api/api-inl.h"
 #include "src/common/message-template.h"
 #include "src/execution/isolate-inl.h"
+#include "src/heap/heap-inl.h"
+#include "src/logging/runtime-call-stats-scope.h"
 #include "src/objects/api-callbacks.h"
-#include "src/objects/hash-table-inl.h"
 #include "src/objects/lookup.h"
-#include "src/objects/property-cell.h"
 #include "src/objects/templates.h"
 
 namespace v8 {
@@ -109,7 +108,7 @@ MaybeHandle<Object> DefineDataProperty(Isolate* isolate,
   ASSIGN_RETURN_ON_EXCEPTION(isolate, value,
                              Instantiate(isolate, prop_data, name), Object);
 
-  LookupIterator::Key key(isolate, name);
+  PropertyKey key(isolate, name);
   LookupIterator it(isolate, object, key, LookupIterator::OWN_SKIP_INTERCEPTOR);
 
 #ifdef DEBUG
@@ -242,7 +241,7 @@ MaybeHandle<JSObject> ConfigureInstance(Isolate* isolate, Handle<JSObject> obj,
       PropertyAttributes attributes = details.attributes();
       PropertyKind kind = details.kind();
 
-      if (kind == kData) {
+      if (kind == PropertyKind::kData) {
         auto prop_data = handle(properties->get(i++), isolate);
         RETURN_ON_EXCEPTION(
             isolate,
@@ -261,7 +260,7 @@ MaybeHandle<JSObject> ConfigureInstance(Isolate* isolate, Handle<JSObject> obj,
       // context.
       PropertyDetails details(Smi::cast(properties->get(i++)));
       PropertyAttributes attributes = details.attributes();
-      DCHECK_EQ(kData, details.kind());
+      DCHECK_EQ(PropertyKind::kData, details.kind());
 
       v8::Intrinsic intrinsic =
           static_cast<v8::Intrinsic>(Smi::ToInt(properties->get(i++)));
@@ -384,7 +383,7 @@ bool IsSimpleInstantiation(Isolate* isolate, ObjectTemplateInfo info,
   if (fun.shared().function_data(kAcquireLoad) != info.constructor())
     return false;
   if (info.immutable_proto()) return false;
-  return fun.context().native_context() == isolate->raw_native_context();
+  return fun.native_context() == isolate->raw_native_context();
 }
 
 MaybeHandle<JSObject> InstantiateObject(Isolate* isolate,
@@ -526,12 +525,14 @@ MaybeHandle<JSFunction> InstantiateFunction(
                                   Handle<HeapObject>::cast(parent_prototype));
     }
   }
-  InstanceType function_type =
-      (!data->needs_access_check() &&
-       data->GetNamedPropertyHandler().IsUndefined(isolate) &&
-       data->GetIndexedPropertyHandler().IsUndefined(isolate))
-          ? JS_API_OBJECT_TYPE
-          : JS_SPECIAL_API_OBJECT_TYPE;
+  InstanceType function_type = JS_SPECIAL_API_OBJECT_TYPE;
+  if (!data->needs_access_check() &&
+      data->GetNamedPropertyHandler().IsUndefined(isolate) &&
+      data->GetIndexedPropertyHandler().IsUndefined(isolate)) {
+    function_type = FLAG_embedder_instance_types && data->HasInstanceType()
+                        ? static_cast<InstanceType>(data->InstanceType())
+                        : JS_API_OBJECT_TYPE;
+  }
 
   Handle<JSFunction> function = ApiNatives::CreateApiFunction(
       isolate, native_context, data, prototype, function_type, maybe_name);
@@ -621,7 +622,8 @@ MaybeHandle<JSObject> ApiNatives::InstantiateRemoteObject(
 void ApiNatives::AddDataProperty(Isolate* isolate, Handle<TemplateInfo> info,
                                  Handle<Name> name, Handle<Object> value,
                                  PropertyAttributes attributes) {
-  PropertyDetails details(kData, attributes, PropertyConstness::kMutable);
+  PropertyDetails details(PropertyKind::kData, attributes,
+                          PropertyConstness::kMutable);
   auto details_handle = handle(details.AsSmi(), isolate);
   Handle<Object> data[] = {name, details_handle, value};
   AddPropertyToPropertyList(isolate, info, arraysize(data), data);
@@ -632,7 +634,8 @@ void ApiNatives::AddDataProperty(Isolate* isolate, Handle<TemplateInfo> info,
                                  PropertyAttributes attributes) {
   auto value = handle(Smi::FromInt(intrinsic), isolate);
   auto intrinsic_marker = isolate->factory()->true_value();
-  PropertyDetails details(kData, attributes, PropertyConstness::kMutable);
+  PropertyDetails details(PropertyKind::kData, attributes,
+                          PropertyConstness::kMutable);
   auto details_handle = handle(details.AsSmi(), isolate);
   Handle<Object> data[] = {name, intrinsic_marker, details_handle, value};
   AddPropertyToPropertyList(isolate, info, arraysize(data), data);
@@ -646,7 +649,8 @@ void ApiNatives::AddAccessorProperty(Isolate* isolate,
                                      PropertyAttributes attributes) {
   if (!getter.is_null()) getter->set_published(true);
   if (!setter.is_null()) setter->set_published(true);
-  PropertyDetails details(kAccessor, attributes, PropertyConstness::kMutable);
+  PropertyDetails details(PropertyKind::kAccessor, attributes,
+                          PropertyConstness::kMutable);
   auto details_handle = handle(details.AsSmi(), isolate);
   Handle<Object> data[] = {name, details_handle, getter, setter};
   AddPropertyToPropertyList(isolate, info, arraysize(data), data);
