@@ -89,6 +89,7 @@ int InstructionScheduler::GetTargetInstructionFlags(
     case kArm64Ubfx:
     case kArm64Ubfx32:
     case kArm64Ubfiz32:
+    case kArm64Sbfiz:
     case kArm64Bfi:
     case kArm64Rbit:
     case kArm64Rbit32:
@@ -181,8 +182,6 @@ int InstructionScheduler::GetTargetInstructionFlags(
     case kArm64F64x2PromoteLowF32x4:
     case kArm64F32x4SConvertI32x4:
     case kArm64F32x4UConvertI32x4:
-    case kArm64F32x4RecipApprox:
-    case kArm64F32x4RecipSqrtApprox:
     case kArm64F32x4Qfma:
     case kArm64F32x4Qfms:
     case kArm64F32x4Pmin:
@@ -205,6 +204,8 @@ int InstructionScheduler::GetTargetInstructionFlags(
     case kArm64INe:
     case kArm64IGtS:
     case kArm64IGeS:
+    case kArm64ILtS:
+    case kArm64ILeS:
     case kArm64I64x2ShrU:
     case kArm64I64x2BitMask:
     case kArm64I32x4SConvertF32x4:
@@ -225,6 +226,8 @@ int InstructionScheduler::GetTargetInstructionFlags(
     case kArm64IGeU:
     case kArm64I32x4BitMask:
     case kArm64I32x4DotI16x8S:
+    case kArm64I16x8DotI8x16S:
+    case kArm64I32x4DotI8x16AddS:
     case kArm64I32x4TruncSatF64x2SZero:
     case kArm64I32x4TruncSatF64x2UZero:
     case kArm64IExtractLaneU:
@@ -315,6 +318,7 @@ int InstructionScheduler::GetTargetInstructionFlags(
     case kArm64LdarDecompressTaggedSigned:
     case kArm64LdarDecompressTaggedPointer:
     case kArm64LdarDecompressAnyTagged:
+    case kArm64LdrDecodeSandboxedPointer:
     case kArm64Peek:
     case kArm64LoadSplat:
     case kArm64LoadLane:
@@ -338,6 +342,7 @@ int InstructionScheduler::GetTargetInstructionFlags(
     case kArm64Str:
     case kArm64StrCompressTagged:
     case kArm64StlrCompressTagged:
+    case kArm64StrEncodeSandboxedPointer:
     case kArm64DmbIsh:
     case kArm64DsbIsb:
     case kArm64StoreLane:
@@ -408,6 +413,7 @@ int InstructionScheduler::GetInstructionLatency(const Instruction* instr) {
     case kArm64Sxth32:
     case kArm64Sxtw:
     case kArm64Ubfiz32:
+    case kArm64Sbfiz:
     case kArm64Ubfx:
     case kArm64Ubfx32:
       return 1;
