@@ -550,64 +550,97 @@ LoopTree* LoopFinder::BuildLoopTree(Graph* graph, TickCounter* tick_counter,
 
 #if V8_ENABLE_WEBASSEMBLY
 // static
-ZoneUnorderedSet<Node*>* LoopFinder::FindSmallUnnestedLoopFromHeader(
-    Node* loop_header, Zone* zone, size_t max_size) {
+ZoneUnorderedSet<Node*>* LoopFinder::FindSmallInnermostLoopFromHeader(
+    Node* loop_header, Zone* zone, size_t max_size, bool calls_are_large) {
   auto* visited = zone->New<ZoneUnorderedSet<Node*>>(zone);
   std::vector<Node*> queue;
 
-  DCHECK(loop_header->opcode() == IrOpcode::kLoop);
+  DCHECK_EQ(loop_header->opcode(), IrOpcode::kLoop);
 
   queue.push_back(loop_header);
 
+#define ENQUEUE_USES(use_name, condition)                                      \
+  for (Node * use_name : node->uses()) {                                       \
+    if (condition && visited->count(use_name) == 0) queue.push_back(use_name); \
+  }
+
   while (!queue.empty()) {
     Node* node = queue.back();
     queue.pop_back();
-    // Terminate is not part of the loop, and neither are its uses.
-    if (node->opcode() == IrOpcode::kTerminate) {
-      DCHECK_EQ(node->InputAt(1), loop_header);
+    if (node->opcode() == IrOpcode::kEnd) {
+      // We reached the end of the graph. The end node is not part of the loop.
       continue;
     }
     visited->insert(node);
     if (visited->size() > max_size) return nullptr;
     switch (node->opcode()) {
+      case IrOpcode::kLoop:
+        // Found nested loop.
+        if (node != loop_header) return nullptr;
+        ENQUEUE_USES(use, true);
+        break;
       case IrOpcode::kLoopExit:
-        DCHECK_EQ(node->InputAt(1), loop_header);
+        // Found nested loop.
+        if (node->InputAt(1) != loop_header) return nullptr;
         // LoopExitValue/Effect uses are inside the loop. The rest are not.
-        for (Node* use : node->uses()) {
-          if (use->opcode() == IrOpcode::kLoopExitEffect ||
-              use->opcode() == IrOpcode::kLoopExitValue) {
-            if (visited->count(use) == 0) queue.push_back(use);
-          }
-        }
+        ENQUEUE_USES(use, (use->opcode() == IrOpcode::kLoopExitEffect ||
+                           use->opcode() == IrOpcode::kLoopExitValue))
         break;
       case IrOpcode::kLoopExitEffect:
       case IrOpcode::kLoopExitValue:
-        DCHECK_EQ(NodeProperties::GetControlInput(node)->InputAt(1),
-                  loop_header);
+        if (NodeProperties::GetControlInput(node)->InputAt(1) != loop_header) {
+          // Found nested loop.
+          return nullptr;
+        }
         // All uses are outside the loop, do nothing.
         break;
+      // If {calls_are_large}, call nodes are considered to have unbounded size,
+      // i.e. >max_size, with the exception of certain wasm builtins.
       case IrOpcode::kTailCall:
       case IrOpcode::kJSWasmCall:
       case IrOpcode::kJSCall:
-        // Call nodes are considered to have unbounded size, i.e. >max_size.
-        // An exception is the call to the stack guard builtin at the beginning
-        // of many loops.
-        return nullptr;
+        if (calls_are_large) return nullptr;
+        ENQUEUE_USES(use, true)
+        break;
       case IrOpcode::kCall: {
-        Node* callee = node->InputAt(0);
-        if (callee->opcode() == IrOpcode::kRelocatableInt32Constant ||
-            callee->opcode() == IrOpcode::kRelocatableInt64Constant) {
-          auto info = OpParameter<RelocatablePtrConstantInfo>(callee->op());
-          if (info.value() != v8::internal::wasm::WasmCode::kWasmStackGuard) {
-            return nullptr;
-          }
+        if (!calls_are_large) {
+          ENQUEUE_USES(use, true);
+          break;
         }
-        V8_FALLTHROUGH;
+        Node* callee = node->InputAt(0);
+        if (callee->opcode() != IrOpcode::kRelocatableInt32Constant &&
+            callee->opcode() != IrOpcode::kRelocatableInt64Constant) {
+          return nullptr;
+        }
+        intptr_t info =
+            OpParameter<RelocatablePtrConstantInfo>(callee->op()).value();
+        using WasmCode = v8::internal::wasm::WasmCode;
+        constexpr intptr_t unrollable_builtins[] = {
+            // Exists in every stack check.
+            WasmCode::kWasmStackGuard,
+            // Fast table operations.
+            WasmCode::kWasmTableGet, WasmCode::kWasmTableSet,
+            WasmCode::kWasmTableGetFuncRef, WasmCode::kWasmTableSetFuncRef,
+            WasmCode::kWasmTableGrow,
+            // Atomics.
+            WasmCode::kWasmAtomicNotify, WasmCode::kWasmI32AtomicWait32,
+            WasmCode::kWasmI32AtomicWait64, WasmCode::kWasmI64AtomicWait32,
+            WasmCode::kWasmI64AtomicWait64,
+            // Exceptions.
+            WasmCode::kWasmAllocateFixedArray, WasmCode::kWasmThrow,
+            WasmCode::kWasmRethrow, WasmCode::kWasmRethrowExplicitContext,
+            // Fast wasm-gc operations.
+            WasmCode::kWasmRefFunc};
+        if (std::count(unrollable_builtins,
+                       unrollable_builtins + arraysize(unrollable_builtins),
+                       info) == 0) {
+          return nullptr;
+        }
+        ENQUEUE_USES(use, true)
+        break;
       }
       default:
-        for (Node* use : node->uses()) {
-          if (visited->count(use) == 0) queue.push_back(use);
-        }
+        ENQUEUE_USES(use, true)
         break;
     }
   }
@@ -662,7 +695,6 @@ bool LoopFinder::HasMarkedExits(LoopTree* loop_tree,
         }
         if (unmarked_exit) {
           if (FLAG_trace_turbo_loop) {
-            Node* loop_node = loop_tree->GetLoopControl(loop);
             PrintF(
                 "Cannot peel loop %i. Loop exit without explicit mark: Node %i "
                 "(%s) is inside loop, but its use %i (%s) is outside.\n",
