@@ -4,8 +4,8 @@
 
 #include "src/execution/microtask-queue.h"
 
-#include <stddef.h>
 #include <algorithm>
+#include <cstddef>
 
 #include "src/api/api-inl.h"
 #include "src/base/logging.h"
@@ -74,7 +74,7 @@ Address MicrotaskQueue::CallEnqueueMicrotask(Isolate* isolate,
   Microtask microtask = Microtask::cast(Object(raw_microtask));
   reinterpret_cast<MicrotaskQueue*>(microtask_queue_pointer)
       ->EnqueueMicrotask(microtask);
-  return ReadOnlyRoots(isolate).undefined_value().ptr();
+  return Smi::zero().ptr();
 }
 
 void MicrotaskQueue::EnqueueMicrotask(v8::Isolate* v8_isolate,
@@ -110,23 +110,21 @@ void MicrotaskQueue::EnqueueMicrotask(Microtask microtask) {
   ++size_;
 }
 
-void MicrotaskQueue::PerformCheckpoint(v8::Isolate* v8_isolate) {
-  if (!IsRunningMicrotasks() && !GetMicrotasksScopeDepth() &&
-      !HasMicrotasksSuppressions()) {
-    std::unique_ptr<MicrotasksScope> microtasks_scope;
-    if (microtasks_policy_ == v8::MicrotasksPolicy::kScoped) {
-      // If we're using microtask scopes to schedule microtask execution, V8
-      // API calls will check that there's always a microtask scope on the
-      // stack. As the microtasks we're about to execute could invoke embedder
-      // callbacks which then calls back into V8, we create an artificial
-      // microtask scope here to avoid running into the CallDepthScope check.
-      microtasks_scope.reset(new v8::MicrotasksScope(
-          v8_isolate, this, v8::MicrotasksScope::kDoNotRunMicrotasks));
-    }
-    Isolate* isolate = reinterpret_cast<Isolate*>(v8_isolate);
-    RunMicrotasks(isolate);
-    isolate->ClearKeptObjects();
+void MicrotaskQueue::PerformCheckpointInternal(v8::Isolate* v8_isolate) {
+  DCHECK(ShouldPerfomCheckpoint());
+  std::unique_ptr<MicrotasksScope> microtasks_scope;
+  if (microtasks_policy_ == v8::MicrotasksPolicy::kScoped) {
+    // If we're using microtask scopes to schedule microtask execution, V8
+    // API calls will check that there's always a microtask scope on the
+    // stack. As the microtasks we're about to execute could invoke embedder
+    // callbacks which then calls back into V8, we create an artificial
+    // microtask scope here to avoid running into the CallDepthScope check.
+    microtasks_scope.reset(new v8::MicrotasksScope(
+        v8_isolate, this, v8::MicrotasksScope::kDoNotRunMicrotasks));
   }
+  Isolate* isolate = reinterpret_cast<Isolate*>(v8_isolate);
+  RunMicrotasks(isolate);
+  isolate->ClearKeptObjects();
 }
 
 namespace {
@@ -158,7 +156,6 @@ int MicrotaskQueue::RunMicrotasks(Isolate* isolate) {
   intptr_t base_count = finished_microtask_count_;
 
   HandleScope handle_scope(isolate);
-  MaybeHandle<Object> maybe_exception;
 
   MaybeHandle<Object> maybe_result;
 
@@ -172,8 +169,7 @@ int MicrotaskQueue::RunMicrotasks(Isolate* isolate) {
     TRACE_EVENT_BEGIN0("v8.execute", "RunMicrotasks");
     {
       TRACE_EVENT_CALL_STATS_SCOPED(isolate, "v8", "V8.RunMicrotasks");
-      maybe_result = Execution::TryRunMicrotasks(isolate, this,
-                                                 &maybe_exception);
+      maybe_result = Execution::TryRunMicrotasks(isolate, this);
       processed_microtask_count =
           static_cast<int>(finished_microtask_count_ - base_count);
     }
@@ -181,15 +177,15 @@ int MicrotaskQueue::RunMicrotasks(Isolate* isolate) {
                      processed_microtask_count);
   }
 
-  // If execution is terminating, clean up and propagate that to TryCatch scope.
-  if (maybe_result.is_null() && maybe_exception.is_null()) {
+  if (isolate->is_execution_terminating()) {
+    DCHECK(isolate->has_scheduled_exception());
+    DCHECK(maybe_result.is_null());
     delete[] ring_buffer_;
     ring_buffer_ = nullptr;
     capacity_ = 0;
     size_ = 0;
     start_ = 0;
-    DCHECK(isolate->has_scheduled_exception());
-    isolate->SetTerminationOnExternalTryCatch();
+    isolate->OnTerminationDuringRunMicrotasks();
     OnCompleted(isolate);
     return -1;
   }
@@ -226,10 +222,6 @@ void MicrotaskQueue::IterateMicrotasks(RootVisitor* visitor) {
   }
 }
 
-int MicrotaskQueue::GetMicrotasksScopeDepth() const {
-  return microtasks_depth_;
-}
-
 void MicrotaskQueue::AddMicrotasksCompletedCallback(
     MicrotasksCompletedCallbackWithData callback, void* data) {
   CallbackWithData callback_with_data(callback, data);
@@ -250,7 +242,7 @@ void MicrotaskQueue::RemoveMicrotasksCompletedCallback(
   microtasks_completed_callbacks_.erase(pos);
 }
 
-void MicrotaskQueue::FireMicrotasksCompletedCallback(Isolate* isolate) const {
+void MicrotaskQueue::OnCompleted(Isolate* isolate) const {
   std::vector<CallbackWithData> callbacks(microtasks_completed_callbacks_);
   for (auto& callback : callbacks) {
     callback.first(reinterpret_cast<v8::Isolate*>(isolate), callback.second);
@@ -263,10 +255,6 @@ Microtask MicrotaskQueue::get(intptr_t index) const {
   return Microtask::cast(microtask);
 }
 
-void MicrotaskQueue::OnCompleted(Isolate* isolate) {
-  FireMicrotasksCompletedCallback(isolate);
-}
-
 void MicrotaskQueue::ResizeBuffer(intptr_t new_capacity) {
   DCHECK_LE(size_, new_capacity);
   Address* new_ring_buffer = new Address[new_capacity];
