@@ -4,7 +4,6 @@
 
 #include "src/heap/cppgc/marking-verifier.h"
 
-#include "include/cppgc/internal/caged-heap-local-data.h"
 #include "src/base/logging.h"
 #include "src/heap/cppgc/gc-info-table.h"
 #include "src/heap/cppgc/heap-object-header.h"
@@ -12,6 +11,10 @@
 #include "src/heap/cppgc/marking-visitor.h"
 #include "src/heap/cppgc/object-view.h"
 
+#if defined(CPPGC_CAGED_HEAP)
+#include "include/cppgc/internal/caged-heap-local-data.h"
+#endif  // defined(CPPGC_CAGED_HEAP)
+
 namespace cppgc {
 namespace internal {
 
@@ -51,7 +54,13 @@ void MarkingVerifierBase::Run(
 // marking. This means that the conservative iteration below may find more
 // objects then the regular marker. The difference is benign as the delta of
 // objects is not reachable from user code but it prevents verification.
-#if !defined(THREAD_SANITIZER)
+// We also avoid verifying the stack when pointer compression is enabled.
+// Currently, verification happens after compaction, V8 compaction can change
+// slots on stack, which could lead to false positives in verifier. Those are
+// more likely with checking compressed pointers on stack.
+// TODO(chromium:1325007): Investigate if Oilpan verification can be moved
+// before V8 compaction or compaction never runs with stack.
+#if !defined(THREAD_SANITIZER) && !defined(CPPGC_POINTER_COMPRESSION)
   if (stack_state == Heap::Config::StackState::kMayContainHeapPointers) {
     in_construction_objects_ = &in_construction_objects_stack_;
     heap_.stack()->IteratePointersUnsafe(this, stack_end);
@@ -106,8 +115,9 @@ bool MarkingVerifierBase::VisitHeapObjectHeader(HeapObjectHeader& header) {
 
 #if defined(CPPGC_YOUNG_GENERATION)
   if (collection_type_ == Heap::Config::CollectionType::kMinor) {
-    const auto age = heap_.caged_heap().local_data().age_table.GetAge(
-        heap_.caged_heap().OffsetFromAddress(header.ObjectStart()));
+    auto& caged_heap = CagedHeap::Instance();
+    const auto age = CagedHeapLocalData::Get().age_table.GetAge(
+        caged_heap.OffsetFromAddress(header.ObjectStart()));
     if (age == AgeTable::Age::kOld) {
       // Do not verify old objects.
       return true;
