@@ -60,16 +60,20 @@ namespace internal {
 
 #define CPPGC_FOR_ALL_HISTOGRAM_CONCURRENT_SCOPES(V) \
   V(ConcurrentMark)                                  \
-  V(ConcurrentSweep)
+  V(ConcurrentSweep)                                 \
+  V(ConcurrentWeakCallback)
 
 #define CPPGC_FOR_ALL_CONCURRENT_SCOPES(V) V(ConcurrentMarkProcessEphemerons)
 
 // Sink for various time and memory statistics.
 class V8_EXPORT_PRIVATE StatsCollector final {
-  using CollectionType = GarbageCollector::Config::CollectionType;
   using IsForcedGC = GarbageCollector::Config::IsForcedGC;
 
  public:
+  using CollectionType = GarbageCollector::Config::CollectionType;
+  using MarkingType = GarbageCollector::Config::MarkingType;
+  using SweepingType = GarbageCollector::Config::SweepingType;
+
 #if defined(CPPGC_DECLARE_ENUM)
   static_assert(false, "CPPGC_DECLARE_ENUM macro is already defined");
 #endif
@@ -105,6 +109,8 @@ class V8_EXPORT_PRIVATE StatsCollector final {
 
     size_t epoch = -1;
     CollectionType collection_type = CollectionType::kMajor;
+    MarkingType marking_type = MarkingType::kAtomic;
+    SweepingType sweeping_type = SweepingType::kAtomic;
     IsForcedGC is_forced_gc = IsForcedGC::kNotForced;
     // Marked bytes collected during marking.
     size_t marked_bytes = 0;
@@ -218,8 +224,9 @@ class V8_EXPORT_PRIVATE StatsCollector final {
   using DisabledConcurrentScope = InternalScope<kDisabled, kConcurrentThread>;
   using EnabledConcurrentScope = InternalScope<kEnabled, kConcurrentThread>;
 
-  // Observer for allocated object size. May be used to implement heap growing
-  // heuristics.
+  // Observer for allocated object size. May e.g. be used to implement heap
+  // growing heuristics. Observers may register/unregister observers at any
+  // time when being invoked.
   class AllocationObserver {
    public:
     // Called after observing at least
@@ -267,22 +274,28 @@ class V8_EXPORT_PRIVATE StatsCollector final {
   void NotifySafePointForTesting();
 
   // Indicates a new garbage collection cycle.
-  void NotifyMarkingStarted(CollectionType, IsForcedGC);
+  void NotifyMarkingStarted(CollectionType, MarkingType, IsForcedGC);
   // Indicates that marking of the current garbage collection cycle is
   // completed.
   void NotifyMarkingCompleted(size_t marked_bytes);
   // Indicates the end of a garbage collection cycle. This means that sweeping
   // is finished at this point.
-  void NotifySweepingCompleted();
+  void NotifySweepingCompleted(SweepingType);
 
   size_t allocated_memory_size() const;
   // Size of live objects in bytes  on the heap. Based on the most recent marked
   // bytes and the bytes allocated since last marking.
   size_t allocated_object_size() const;
 
-  // Returns the most recent marked bytes count. Should not be called during
+  // Returns the overall marked bytes count, i.e. if young generation is
+  // enabled, it returns the accumulated number. Should not be called during
   // marking.
   size_t marked_bytes() const;
+
+  // Returns the marked bytes for the current cycle. Should only be called
+  // within GC cycle.
+  size_t marked_bytes_on_current_cycle() const;
+
   // Returns the overall duration of the most recent marking phase. Should not
   // be called during marking.
   v8::base::TimeDelta marking_time() const;
@@ -339,6 +352,10 @@ class V8_EXPORT_PRIVATE StatsCollector final {
   size_t tracked_live_bytes_ = 0;
 #endif  // CPPGC_VERIFY_HEAP
 
+  // The number of bytes marked so far. For young generation (with sticky bits)
+  // keeps track of marked bytes across multiple GC cycles.
+  size_t marked_bytes_so_far_ = 0;
+
   int64_t memory_allocated_bytes_ = 0;
   int64_t memory_freed_bytes_since_end_of_marking_ = 0;
   std::atomic<size_t> discarded_bytes_{0};
@@ -346,6 +363,7 @@ class V8_EXPORT_PRIVATE StatsCollector final {
   // vector to allow fast iteration of observers. Register/Unregisters only
   // happens on startup/teardown.
   std::vector<AllocationObserver*> allocation_observers_;
+  bool allocation_observer_deleted_ = false;
 
   GarbageCollectionState gc_state_ = GarbageCollectionState::kNotRunning;
 
@@ -363,8 +381,19 @@ class V8_EXPORT_PRIVATE StatsCollector final {
 
 template <typename Callback>
 void StatsCollector::ForAllAllocationObservers(Callback callback) {
-  for (AllocationObserver* observer : allocation_observers_) {
-    callback(observer);
+  // Iterate using indices to allow push_back() of new observers.
+  for (size_t i = 0; i < allocation_observers_.size(); ++i) {
+    auto* observer = allocation_observers_[i];
+    if (observer) {
+      callback(observer);
+    }
+  }
+  if (allocation_observer_deleted_) {
+    allocation_observers_.erase(
+        std::remove(allocation_observers_.begin(), allocation_observers_.end(),
+                    nullptr),
+        allocation_observers_.end());
+    allocation_observer_deleted_ = false;
   }
 }
 
