@@ -22,8 +22,16 @@ void HeapAllocator::Setup() {
   for (int i = FIRST_SPACE; i <= LAST_SPACE; ++i) {
     spaces_[i] = heap_->space(i);
   }
+
+  space_for_maps_ = spaces_[MAP_SPACE]
+                        ? static_cast<PagedSpace*>(spaces_[MAP_SPACE])
+                        : static_cast<PagedSpace*>(spaces_[OLD_SPACE]);
+
   shared_old_allocator_ = heap_->shared_old_allocator_.get();
-  shared_map_allocator_ = heap_->shared_map_allocator_.get();
+  shared_map_allocator_ = heap_->shared_map_allocator_
+                              ? heap_->shared_map_allocator_.get()
+                              : shared_old_allocator_;
+  shared_lo_space_ = heap_->shared_lo_space();
 }
 
 void HeapAllocator::SetReadOnlySpace(ReadOnlySpace* read_only_space) {
@@ -41,10 +49,12 @@ AllocationResult HeapAllocator::AllocateRawLargeInternal(
       return lo_space()->AllocateRaw(size_in_bytes);
     case AllocationType::kCode:
       return code_lo_space()->AllocateRaw(size_in_bytes);
+    case AllocationType::kSharedOld:
+      return shared_lo_space()->AllocateRawBackground(
+          heap_->main_thread_local_heap(), size_in_bytes);
     case AllocationType::kMap:
     case AllocationType::kReadOnly:
     case AllocationType::kSharedMap:
-    case AllocationType::kSharedOld:
       UNREACHABLE();
   }
 }
@@ -82,7 +92,13 @@ AllocationResult HeapAllocator::AllocateRawWithLightRetrySlowPath(
     if (IsSharedAllocationType(allocation)) {
       heap_->CollectSharedGarbage(GarbageCollectionReason::kAllocationFailure);
     } else {
-      heap_->CollectGarbage(AllocationTypeToGCSpace(allocation),
+      AllocationSpace space_to_gc = AllocationTypeToGCSpace(allocation);
+      if (v8_flags.minor_mc && i > 0) {
+        // Repeated young gen GCs won't have any additional effect. Do a full GC
+        // instead.
+        space_to_gc = AllocationSpace::OLD_SPACE;
+      }
+      heap_->CollectGarbage(space_to_gc,
                             GarbageCollectionReason::kAllocationFailure);
     }
     result = AllocateRaw(size, allocation, origin, alignment);
@@ -100,7 +116,6 @@ AllocationResult HeapAllocator::AllocateRawWithRetryOrFailSlowPath(
       AllocateRawWithLightRetrySlowPath(size, allocation, origin, alignment);
   if (!result.IsFailure()) return result;
 
-  heap_->isolate()->counters()->gc_last_resort_from_handles()->Increment();
   if (IsSharedAllocationType(allocation)) {
     heap_->CollectSharedGarbage(GarbageCollectionReason::kLastResort);
 
@@ -121,8 +136,8 @@ AllocationResult HeapAllocator::AllocateRawWithRetryOrFailSlowPath(
     return result;
   }
 
-  v8::internal::V8::FatalProcessOutOfMemory(heap_->isolate(),
-                                            "CALL_AND_RETRY_LAST", true);
+  V8::FatalProcessOutOfMemory(heap_->isolate(), "CALL_AND_RETRY_LAST",
+                              V8::kHeapOOM);
 }
 
 #ifdef DEBUG
@@ -135,16 +150,31 @@ void HeapAllocator::IncrementObjectCounters() {
 #endif  // DEBUG
 
 #ifdef V8_ENABLE_ALLOCATION_TIMEOUT
+// static
+void HeapAllocator::InitializeOncePerProcess() {
+  SetAllocationGcInterval(v8_flags.gc_interval);
+}
+
+// static
+void HeapAllocator::SetAllocationGcInterval(int allocation_gc_interval) {
+  allocation_gc_interval_.store(allocation_gc_interval,
+                                std::memory_order_relaxed);
+}
+
+// static
+std::atomic<int> HeapAllocator::allocation_gc_interval_{-1};
 
 void HeapAllocator::SetAllocationTimeout(int allocation_timeout) {
-  allocation_timeout_ = allocation_timeout;
+  // See `allocation_timeout_` for description. We map negative values to 0 to
+  // avoid underflows as allocation decrements this value as well.
+  allocation_timeout_ = std::max(0, allocation_timeout);
 }
 
 void HeapAllocator::UpdateAllocationTimeout() {
-  if (FLAG_random_gc_interval > 0) {
+  if (v8_flags.random_gc_interval > 0) {
     const int new_timeout = allocation_timeout_ <= 0
                                 ? heap_->isolate()->fuzzer_rng()->NextInt(
-                                      FLAG_random_gc_interval + 1)
+                                      v8_flags.random_gc_interval + 1)
                                 : allocation_timeout_;
     // Reset the allocation timeout, but make sure to allow at least a few
     // allocations after a collection. The reason for this is that we have a lot
@@ -152,8 +182,12 @@ void HeapAllocator::UpdateAllocationTimeout() {
     // allow the subsequent allocation attempts to go through.
     constexpr int kFewAllocationsHeadroom = 6;
     allocation_timeout_ = std::max(kFewAllocationsHeadroom, new_timeout);
-  } else if (FLAG_gc_interval >= 0) {
-    allocation_timeout_ = FLAG_gc_interval;
+    return;
+  }
+
+  int interval = allocation_gc_interval_.load(std::memory_order_relaxed);
+  if (interval >= 0) {
+    allocation_timeout_ = interval;
   }
 }
 
