@@ -14,10 +14,29 @@
 namespace v8 {
 namespace internal {
 
+// static
+void StubCache::ClearCallback(v8::Isolate* isolate, v8::GCType type,
+                              v8::GCCallbackFlags flags, void* data) {
+  StubCache* cache = static_cast<StubCache*>(data);
+  cache->Clear();
+}
+
 StubCache::StubCache(Isolate* isolate) : isolate_(isolate) {
-  // Ensure the nullptr (aka Smi::kZero) which StubCache::Get() returns
+  // Ensure the nullptr (aka Smi::zero()) which StubCache::Get() returns
   // when the entry is not found is not considered as a handler.
   DCHECK(!IC::IsHandler(MaybeObject()));
+
+  // The stub caches are not traversed during GC; clear them to force
+  // their lazy re-initialization. This must be done after the
+  // GC, because it relies on the new address of certain old space
+  // objects (empty string, illegal builtin).
+
+  isolate_->heap()->AddGCEpilogueCallback(ClearCallback,
+                                          kGCTypeMarkSweepCompact, this);
+}
+
+StubCache::~StubCache() {
+  isolate_->heap()->RemoveGCEpilogueCallback(ClearCallback, this);
 }
 
 void StubCache::Initialize() {
@@ -31,8 +50,8 @@ void StubCache::Initialize() {
 // is scaled by 1 << kCacheIndexShift.
 int StubCache::PrimaryOffset(Name name, Map map) {
   // Compute the hash of the name (use entire hash field).
-  DCHECK(name.HasHashCode());
-  uint32_t field = name.hash_field();
+  uint32_t field = name.raw_hash_field();
+  DCHECK(Name::IsHashFieldComputed(field));
   // Using only the low bits in 64-bit mode is unlikely to increase the
   // risk of collision even if the heap is spread over an area larger than
   // 4Gb (and not at all if it isn't).
@@ -44,12 +63,14 @@ int StubCache::PrimaryOffset(Name name, Map map) {
 }
 
 // Hash algorithm for the secondary table.  This algorithm is replicated in
-// assembler for every architecture.  Returns an index into the table that
-// is scaled by 1 << kCacheIndexShift.
-int StubCache::SecondaryOffset(Name name, int seed) {
-  // Use the seed from the primary cache in the secondary cache.
+// assembler. This hash should be sufficiently different from the primary one
+// in order to avoid collisions for minified code with short names.
+// Returns an index into the table that is scaled by 1 << kCacheIndexShift.
+int StubCache::SecondaryOffset(Name name, Map old_map) {
   uint32_t name_low32bits = static_cast<uint32_t>(name.ptr());
-  uint32_t key = (seed - name_low32bits) + kSecondaryMagic;
+  uint32_t map_low32bits = static_cast<uint32_t>(old_map.ptr());
+  uint32_t key = (map_low32bits + name_low32bits);
+  key = key + (key >> kSecondaryKeyShift);
   return key & ((kSecondaryTableSize - 1) << kCacheIndexShift);
 }
 
@@ -57,8 +78,8 @@ int StubCache::PrimaryOffsetForTesting(Name name, Map map) {
   return PrimaryOffset(name, map);
 }
 
-int StubCache::SecondaryOffsetForTesting(Name name, int seed) {
-  return SecondaryOffset(name, seed);
+int StubCache::SecondaryOffsetForTesting(Name name, Map map) {
+  return SecondaryOffset(name, map);
 }
 
 #ifdef DEBUG
@@ -71,7 +92,6 @@ bool CommonStubCacheChecks(StubCache* stub_cache, Name name, Map map,
   DCHECK(!Heap::InYoungGeneration(name));
   DCHECK(!Heap::InYoungGeneration(handler));
   DCHECK(name.IsUniqueName());
-  DCHECK(name.HasHashCode());
   if (handler->ptr() != kNullAddress) DCHECK(IC::IsHandler(handler));
   return true;
 }
@@ -90,15 +110,13 @@ void StubCache::Set(Name name, Map map, MaybeObject handler) {
   // If the primary entry has useful data in it, we retire it to the
   // secondary cache before overwriting it.
   if (old_handler != MaybeObject::FromObject(
-                         isolate()->builtins()->builtin(Builtins::kIllegal)) &&
+                         isolate()->builtins()->code(Builtin::kIllegal)) &&
       !primary->map.IsSmi()) {
     Map old_map =
         Map::cast(StrongTaggedValue::ToObject(isolate(), primary->map));
-    int seed = PrimaryOffset(
-        Name::cast(StrongTaggedValue::ToObject(isolate(), primary->key)),
-        old_map);
-    int secondary_offset = SecondaryOffset(
-        Name::cast(StrongTaggedValue::ToObject(isolate(), primary->key)), seed);
+    Name old_name =
+        Name::cast(StrongTaggedValue::ToObject(isolate(), primary->key));
+    int secondary_offset = SecondaryOffset(old_name, old_map);
     Entry* secondary = entry(secondary_, secondary_offset);
     *secondary = *primary;
   }
@@ -117,7 +135,7 @@ MaybeObject StubCache::Get(Name name, Map map) {
   if (primary->key == name && primary->map == map) {
     return TaggedValue::ToMaybeObject(isolate(), primary->value);
   }
-  int secondary_offset = SecondaryOffset(name, primary_offset);
+  int secondary_offset = SecondaryOffset(name, map);
   Entry* secondary = entry(secondary_, secondary_offset);
   if (secondary->key == name && secondary->map == map) {
     return TaggedValue::ToMaybeObject(isolate(), secondary->value);
@@ -126,8 +144,8 @@ MaybeObject StubCache::Get(Name name, Map map) {
 }
 
 void StubCache::Clear() {
-  MaybeObject empty = MaybeObject::FromObject(
-      isolate_->builtins()->builtin(Builtins::kIllegal));
+  MaybeObject empty =
+      MaybeObject::FromObject(isolate_->builtins()->code(Builtin::kIllegal));
   Name empty_string = ReadOnlyRoots(isolate()).empty_string();
   for (int i = 0; i < kPrimaryTableSize; i++) {
     primary_[i].key = StrongTaggedValue(empty_string);
