@@ -21,29 +21,16 @@ namespace maglev {
 
 class Graph;
 
-class MaglevCompiler {
+class MaglevCompiler : public AllStatic {
  public:
   // May be called from any thread.
-  static void Compile(MaglevCompilationUnit* toplevel_compilation_unit);
+  static void Compile(LocalIsolate* local_isolate,
+                      MaglevCompilationInfo* compilation_info);
 
   // Called on the main thread after Compile has completed.
   // TODO(v8:7700): Move this to a different class?
   static MaybeHandle<CodeT> GenerateCode(
-      MaglevCompilationUnit* toplevel_compilation_unit);
-
- private:
-  explicit MaglevCompiler(MaglevCompilationUnit* toplevel_compilation_unit)
-      : toplevel_compilation_unit_(toplevel_compilation_unit) {}
-
-  void Compile();
-
-  compiler::JSHeapBroker* broker() const {
-    return toplevel_compilation_unit_->broker();
-  }
-  Zone* zone() { return toplevel_compilation_unit_->zone(); }
-  Isolate* isolate() { return toplevel_compilation_unit_->isolate(); }
-
-  MaglevCompilationUnit* const toplevel_compilation_unit_;
+      MaglevCompilationInfo* compilation_info);
 };
 
 }  // namespace maglev
