@@ -11,7 +11,8 @@
 
 #include "src/base/numbers/dtoa.h"
 #include "src/base/numbers/strtod.h"
-#include "src/base/platform/wrappers.h"
+#include "src/base/small-vector.h"
+#include "src/bigint/bigint.h"
 #include "src/common/assert-scope.h"
 #include "src/handles/handles.h"
 #include "src/heap/factory.h"
@@ -139,7 +140,7 @@ class SimpleStringBuilder {
 };
 
 inline double JunkStringValue() {
-  return bit_cast<double, uint64_t>(kQuietNaNMask);
+  return base::bit_cast<double, uint64_t>(kQuietNaNMask);
 }
 
 inline double SignedZero(bool negative) {
@@ -310,38 +311,35 @@ class StringToIntHelper {
 
  protected:
   // Subclasses must implement these:
-  virtual void AllocateResult() = 0;
-  virtual void ResultMultiplyAdd(uint32_t multiplier, uint32_t part) = 0;
+  virtual void ParseOneByte(const uint8_t* start) = 0;
+  virtual void ParseTwoByte(const base::uc16* start) = 0;
 
   // Subclasses must call this to do all the work.
   void ParseInt();
 
-  // Subclasses may override this.
-  virtual bool CheckTermination() { return false; }
-  virtual void HandleSpecialCases() {}
-
   // Subclass constructors should call these for configuration before calling
   // ParseInt().
   void set_allow_binary_and_octal_prefixes() {
     allow_binary_and_octal_prefixes_ = true;
   }
   void set_disallow_trailing_junk() { allow_trailing_junk_ = false; }
+  bool allow_trailing_junk() { return allow_trailing_junk_; }
 
   bool IsOneByte() const {
     return raw_one_byte_subject_ != nullptr ||
            String::IsOneByteRepresentationUnderneath(*subject_);
   }
 
-  base::Vector<const uint8_t> GetOneByteVector() {
+  base::Vector<const uint8_t> GetOneByteVector(
+      const DisallowGarbageCollection& no_gc) {
     if (raw_one_byte_subject_ != nullptr) {
       return base::Vector<const uint8_t>(raw_one_byte_subject_, length_);
     }
-    DisallowGarbageCollection no_gc;
     return subject_->GetFlatContent(no_gc).ToOneByteVector();
   }
 
-  base::Vector<const base::uc16> GetTwoByteVector() {
-    DisallowGarbageCollection no_gc;
+  base::Vector<const base::uc16> GetTwoByteVector(
+      const DisallowGarbageCollection& no_gc) {
     return subject_->GetFlatContent(no_gc).ToUC16Vector();
   }
 
@@ -357,8 +355,6 @@ class StringToIntHelper {
  private:
   template <class Char>
   void DetectRadixInternal(Char current, int length);
-  template <class Char>
-  bool ParseChunkInternal(Char start);
 
   IsolateT* isolate_;
   Handle<String> subject_;
@@ -375,46 +371,18 @@ class StringToIntHelper {
 
 template <typename IsolateT>
 void StringToIntHelper<IsolateT>::ParseInt() {
-  {
-    DisallowGarbageCollection no_gc;
-    if (IsOneByte()) {
-      base::Vector<const uint8_t> vector = GetOneByteVector();
-      DetectRadixInternal(vector.begin(), vector.length());
-    } else {
-      base::Vector<const base::uc16> vector = GetTwoByteVector();
-      DetectRadixInternal(vector.begin(), vector.length());
-    }
+  DisallowGarbageCollection no_gc;
+  if (IsOneByte()) {
+    base::Vector<const uint8_t> vector = GetOneByteVector(no_gc);
+    DetectRadixInternal(vector.begin(), vector.length());
+    if (state_ != State::kRunning) return;
+    ParseOneByte(vector.begin());
+  } else {
+    base::Vector<const base::uc16> vector = GetTwoByteVector(no_gc);
+    DetectRadixInternal(vector.begin(), vector.length());
+    if (state_ != State::kRunning) return;
+    ParseTwoByte(vector.begin());
   }
-  if (state_ != State::kRunning) return;
-  AllocateResult();
-  HandleSpecialCases();
-  if (state_ != State::kRunning) return;
-  do {
-    {
-      DisallowGarbageCollection no_gc;
-      if (IsOneByte()) {
-        base::Vector<const uint8_t> vector = GetOneByteVector();
-        DCHECK_EQ(length_, vector.length());
-        if (ParseChunkInternal(vector.begin())) {
-          break;
-        }
-      } else {
-        base::Vector<const base::uc16> vector = GetTwoByteVector();
-        DCHECK_EQ(length_, vector.length());
-        if (ParseChunkInternal(vector.begin())) {
-          break;
-        }
-      }
-    }
-
-    // The flat vector handle is temporarily released after parsing 10kb
-    // in order to invoke interrupts which may in turn invoke GC.
-    if (CheckTermination()) {
-      set_state(State::kError);
-      break;
-    }
-  } while (true);
-  DCHECK_NE(state_, State::kRunning);
 }
 
 template <typename IsolateT>
@@ -493,26 +461,122 @@ void StringToIntHelper<IsolateT>::DetectRadixInternal(Char current,
   }
 
   DCHECK(radix_ >= 2 && radix_ <= 36);
-  STATIC_ASSERT(String::kMaxLength <= INT_MAX);
+  static_assert(String::kMaxLength <= INT_MAX);
   cursor_ = static_cast<int>(current - start);
 }
 
-template <typename IsolateT>
-template <class Char>
-bool StringToIntHelper<IsolateT>::ParseChunkInternal(Char start) {
-  const int kChunkSize = 10240;
-  Char current = start + cursor_;
-  Char end = start + length_;
-  Char break_pos = current + kChunkSize;
+class NumberParseIntHelper : public StringToIntHelper<Isolate> {
+ public:
+  NumberParseIntHelper(Isolate* isolate, Handle<String> string, int radix)
+      : StringToIntHelper(isolate, string, radix) {}
 
+  template <class Char>
+  void ParseInternal(Char start) {
+    Char current = start + cursor();
+    Char end = start + length();
+
+    if (radix() == 10) return HandleBaseTenCase(current, end);
+    if (base::bits::IsPowerOfTwo(radix())) {
+      result_ = HandlePowerOfTwoCase(current, end);
+      set_state(State::kDone);
+      return;
+    }
+    return HandleGenericCase(current, end);
+  }
+  void ParseOneByte(const uint8_t* start) final { return ParseInternal(start); }
+  void ParseTwoByte(const base::uc16* start) final {
+    return ParseInternal(start);
+  }
+
+  double GetResult() {
+    ParseInt();
+    switch (state()) {
+      case State::kJunk:
+      case State::kEmpty:
+        return JunkStringValue();
+      case State::kZero:
+        return SignedZero(negative());
+      case State::kDone:
+        return negative() ? -result_ : result_;
+      case State::kError:
+      case State::kRunning:
+        break;
+    }
+    UNREACHABLE();
+  }
+
+ private:
+  template <class Char>
+  void HandleGenericCase(Char current, Char end);
+
+  template <class Char>
+  double HandlePowerOfTwoCase(Char current, Char end) {
+    const bool allow_trailing_junk = true;
+    // GetResult() will take care of the sign bit, so ignore it for now.
+    const bool negative = false;
+    switch (radix()) {
+      case 2:
+        return InternalStringToIntDouble<1>(current, end, negative,
+                                            allow_trailing_junk);
+      case 4:
+        return InternalStringToIntDouble<2>(current, end, negative,
+                                            allow_trailing_junk);
+      case 8:
+        return InternalStringToIntDouble<3>(current, end, negative,
+                                            allow_trailing_junk);
+
+      case 16:
+        return InternalStringToIntDouble<4>(current, end, negative,
+                                            allow_trailing_junk);
+
+      case 32:
+        return InternalStringToIntDouble<5>(current, end, negative,
+                                            allow_trailing_junk);
+      default:
+        UNREACHABLE();
+    }
+  }
+
+  template <class Char>
+  void HandleBaseTenCase(Char current, Char end) {
+    // Parsing with strtod.
+    const int kMaxSignificantDigits = 309;  // Doubles are less than 1.8e308.
+    // The buffer may contain up to kMaxSignificantDigits + 1 digits and a zero
+    // end.
+    const int kBufferSize = kMaxSignificantDigits + 2;
+    char buffer[kBufferSize];
+    int buffer_pos = 0;
+    while (*current >= '0' && *current <= '9') {
+      if (buffer_pos <= kMaxSignificantDigits) {
+        // If the number has more than kMaxSignificantDigits it will be parsed
+        // as infinity.
+        DCHECK_LT(buffer_pos, kBufferSize);
+        buffer[buffer_pos++] = static_cast<char>(*current);
+      }
+      ++current;
+      if (current == end) break;
+    }
+
+    SLOW_DCHECK(buffer_pos < kBufferSize);
+    buffer[buffer_pos] = '\0';
+    base::Vector<const char> buffer_vector(buffer, buffer_pos);
+    result_ = Strtod(buffer_vector, 0);
+    set_state(State::kDone);
+  }
+
+  double result_ = 0;
+};
+
+template <class Char>
+void NumberParseIntHelper::HandleGenericCase(Char current, Char end) {
   // The following code causes accumulating rounding error for numbers greater
   // than ~2^56. It's explicitly allowed in the spec: "if R is not 2, 4, 8, 10,
   // 16, or 32, then mathInt may be an implementation-dependent approximation to
   // the mathematical integer value" (15.1.2.2).
 
-  int lim_0 = '0' + (radix_ < 10 ? radix_ : 10);
-  int lim_a = 'a' + (radix_ - 10);
-  int lim_A = 'A' + (radix_ - 10);
+  int lim_0 = '0' + (radix() < 10 ? radix() : 10);
+  int lim_a = 'a' + (radix() - 10);
+  int lim_A = 'A' + (radix() - 10);
 
   // NOTE: The code for computing the value may seem a bit complex at
   // first glance. It is structured to use 32-bit multiply-and-add
@@ -542,9 +606,9 @@ bool StringToIntHelper<IsolateT>::ParseChunkInternal(Char start) {
       // will not overflow the multiplier, we stop parsing the part
       // by leaving the loop.
       const uint32_t kMaximumMultiplier = 0xFFFFFFFFU / 36;
-      uint32_t m = multiplier * static_cast<uint32_t>(radix_);
+      uint32_t m = multiplier * static_cast<uint32_t>(radix());
       if (m > kMaximumMultiplier) break;
-      part = part * radix_ + d;
+      part = part * radix() + d;
       multiplier = m;
       DCHECK(multiplier > part);
 
@@ -554,132 +618,14 @@ bool StringToIntHelper<IsolateT>::ParseChunkInternal(Char start) {
         break;
       }
     }
-
-    // Update the value and skip the part in the string.
-    ResultMultiplyAdd(multiplier, part);
-
-    // Set final state
-    if (done) {
-      if (!allow_trailing_junk_ && AdvanceToNonspace(&current, end)) {
-        set_state(State::kJunk);
-      } else {
-        set_state(State::kDone);
-      }
-      return true;
-    }
-  } while (current < break_pos);
-
-  cursor_ = static_cast<int>(current - start);
-  return false;
-}
-
-class NumberParseIntHelper : public StringToIntHelper<Isolate> {
- public:
-  NumberParseIntHelper(Isolate* isolate, Handle<String> string, int radix)
-      : StringToIntHelper(isolate, string, radix) {}
-
-  double GetResult() {
-    ParseInt();
-    switch (state()) {
-      case State::kJunk:
-      case State::kEmpty:
-        return JunkStringValue();
-      case State::kZero:
-        return SignedZero(negative());
-      case State::kDone:
-        return negative() ? -result_ : result_;
-      case State::kError:
-      case State::kRunning:
-        break;
-    }
-    UNREACHABLE();
-  }
-
- protected:
-  void AllocateResult() override {}
-  void ResultMultiplyAdd(uint32_t multiplier, uint32_t part) override {
     result_ = result_ * multiplier + part;
+  } while (!done);
+
+  if (!allow_trailing_junk() && AdvanceToNonspace(&current, end)) {
+    return set_state(State::kJunk);
   }
-
- private:
-  void HandleSpecialCases() override {
-    bool is_power_of_two = base::bits::IsPowerOfTwo(radix());
-    if (!is_power_of_two && radix() != 10) return;
-    DisallowGarbageCollection no_gc;
-    if (IsOneByte()) {
-      base::Vector<const uint8_t> vector = GetOneByteVector();
-      DCHECK_EQ(length(), vector.length());
-      result_ = is_power_of_two ? HandlePowerOfTwoCase(vector.begin())
-                                : HandleBaseTenCase(vector.begin());
-    } else {
-      base::Vector<const base::uc16> vector = GetTwoByteVector();
-      DCHECK_EQ(length(), vector.length());
-      result_ = is_power_of_two ? HandlePowerOfTwoCase(vector.begin())
-                                : HandleBaseTenCase(vector.begin());
-    }
-    set_state(State::kDone);
-  }
-
-  template <class Char>
-  double HandlePowerOfTwoCase(Char start) {
-    Char current = start + cursor();
-    Char end = start + length();
-    const bool allow_trailing_junk = true;
-    // GetResult() will take care of the sign bit, so ignore it for now.
-    const bool negative = false;
-    switch (radix()) {
-      case 2:
-        return InternalStringToIntDouble<1>(current, end, negative,
-                                            allow_trailing_junk);
-      case 4:
-        return InternalStringToIntDouble<2>(current, end, negative,
-                                            allow_trailing_junk);
-      case 8:
-        return InternalStringToIntDouble<3>(current, end, negative,
-                                            allow_trailing_junk);
-
-      case 16:
-        return InternalStringToIntDouble<4>(current, end, negative,
-                                            allow_trailing_junk);
-
-      case 32:
-        return InternalStringToIntDouble<5>(current, end, negative,
-                                            allow_trailing_junk);
-      default:
-        UNREACHABLE();
-    }
-  }
-
-  template <class Char>
-  double HandleBaseTenCase(Char start) {
-    // Parsing with strtod.
-    Char current = start + cursor();
-    Char end = start + length();
-    const int kMaxSignificantDigits = 309;  // Doubles are less than 1.8e308.
-    // The buffer may contain up to kMaxSignificantDigits + 1 digits and a zero
-    // end.
-    const int kBufferSize = kMaxSignificantDigits + 2;
-    char buffer[kBufferSize];
-    int buffer_pos = 0;
-    while (*current >= '0' && *current <= '9') {
-      if (buffer_pos <= kMaxSignificantDigits) {
-        // If the number has more than kMaxSignificantDigits it will be parsed
-        // as infinity.
-        DCHECK_LT(buffer_pos, kBufferSize);
-        buffer[buffer_pos++] = static_cast<char>(*current);
-      }
-      ++current;
-      if (current == end) break;
-    }
-
-    SLOW_DCHECK(buffer_pos < kBufferSize);
-    buffer[buffer_pos] = '\0';
-    base::Vector<const char> buffer_vector(buffer, buffer_pos);
-    return Strtod(buffer_vector, 0);
-  }
-
-  double result_ = 0;
-};
+  return set_state(State::kDone);
+}
 
 // Converts a string to a double value. Assumes the Iterator supports
 // the following operations:
@@ -724,19 +670,19 @@ double InternalStringToDouble(Iterator current, EndMark end, int flags,
   int insignificant_digits = 0;
   bool nonzero_digit_dropped = false;
 
-  enum Sign { NONE, NEGATIVE, POSITIVE };
+  enum class Sign { kNone, kNegative, kPositive };
 
-  Sign sign = NONE;
+  Sign sign = Sign::kNone;
 
   if (*current == '+') {
     // Ignore leading sign.
     ++current;
     if (current == end) return JunkStringValue();
-    sign = POSITIVE;
+    sign = Sign::kPositive;
   } else if (*current == '-') {
     ++current;
     if (current == end) return JunkStringValue();
-    sign = NEGATIVE;
+    sign = Sign::kNegative;
   }
 
   static const char kInfinityString[] = "Infinity";
@@ -750,20 +696,20 @@ double InternalStringToDouble(Iterator current, EndMark end, int flags,
     }
 
     DCHECK_EQ(buffer_pos, 0);
-    return (sign == NEGATIVE) ? -V8_INFINITY : V8_INFINITY;
+    return (sign == Sign::kNegative) ? -V8_INFINITY : V8_INFINITY;
   }
 
   bool leading_zero = false;
   if (*current == '0') {
     ++current;
-    if (current == end) return SignedZero(sign == NEGATIVE);
+    if (current == end) return SignedZero(sign == Sign::kNegative);
 
     leading_zero = true;
 
     // It could be hexadecimal value.
     if ((flags & ALLOW_HEX) && (*current == 'x' || *current == 'X')) {
       ++current;
-      if (current == end || !isDigit(*current, 16) || sign != NONE) {
+      if (current == end || !isDigit(*current, 16) || sign != Sign::kNone) {
         return JunkStringValue();  // "0x".
       }
 
@@ -773,7 +719,7 @@ double InternalStringToDouble(Iterator current, EndMark end, int flags,
       // It could be an explicit octal value.
     } else if ((flags & ALLOW_OCTAL) && (*current == 'o' || *current == 'O')) {
       ++current;
-      if (current == end || !isDigit(*current, 8) || sign != NONE) {
+      if (current == end || !isDigit(*current, 8) || sign != Sign::kNone) {
         return JunkStringValue();  // "0o".
       }
 
@@ -783,7 +729,7 @@ double InternalStringToDouble(Iterator current, EndMark end, int flags,
       // It could be a binary value.
     } else if ((flags & ALLOW_BINARY) && (*current == 'b' || *current == 'B')) {
       ++current;
-      if (current == end || !isBinaryDigit(*current) || sign != NONE) {
+      if (current == end || !isBinaryDigit(*current) || sign != Sign::kNone) {
         return JunkStringValue();  // "0b".
       }
 
@@ -794,7 +740,7 @@ double InternalStringToDouble(Iterator current, EndMark end, int flags,
     // Ignore leading zeros in the integer part.
     while (*current == '0') {
       ++current;
-      if (current == end) return SignedZero(sign == NEGATIVE);
+      if (current == end) return SignedZero(sign == Sign::kNegative);
     }
   }
 
@@ -839,7 +785,7 @@ double InternalStringToDouble(Iterator current, EndMark end, int flags,
       // leading zeros (if any).
       while (*current == '0') {
         ++current;
-        if (current == end) return SignedZero(sign == NEGATIVE);
+        if (current == end) return SignedZero(sign == Sign::kNegative);
         exponent--;  // Move this 0 into the exponent.
       }
     }
@@ -880,9 +826,9 @@ double InternalStringToDouble(Iterator current, EndMark end, int flags,
         return JunkStringValue();
       }
     }
-    char sign = '+';
+    char exponent_sign = '+';
     if (*current == '+' || *current == '-') {
-      sign = static_cast<char>(*current);
+      exponent_sign = static_cast<char>(*current);
       ++current;
       if (current == end) {
         if (allow_trailing_junk) {
@@ -916,7 +862,7 @@ double InternalStringToDouble(Iterator current, EndMark end, int flags,
       ++current;
     } while (current != end && *current >= '0' && *current <= '9');
 
-    exponent += (sign == '-' ? -num : num);
+    exponent += (exponent_sign == '-' ? -num : num);
   }
 
   if (!allow_trailing_junk && AdvanceToNonspace(&current, end)) {
@@ -928,7 +874,8 @@ parsing_done:
 
   if (octal) {
     return InternalStringToIntDouble<3>(buffer, buffer + buffer_pos,
-                                        sign == NEGATIVE, allow_trailing_junk);
+                                        sign == Sign::kNegative,
+                                        allow_trailing_junk);
   }
 
   if (nonzero_digit_dropped) {
@@ -941,7 +888,7 @@ parsing_done:
 
   double converted =
       Strtod(base::Vector<const char>(buffer, buffer_pos), exponent);
-  return (sign == NEGATIVE) ? -converted : converted;
+  return (sign == Sign::kNegative) ? -converted : converted;
 }
 
 double StringToDouble(const char* str, int flags, double empty_string_val) {
@@ -989,6 +936,11 @@ class StringToBigIntHelper : public StringToIntHelper<IsolateT> {
     this->set_allow_binary_and_octal_prefixes();
   }
 
+  void ParseOneByte(const uint8_t* start) final { return ParseInternal(start); }
+  void ParseTwoByte(const base::uc16* start) final {
+    return ParseInternal(start);
+  }
+
   MaybeHandle<BigInt> GetResult() {
     this->ParseInt();
     if (behavior_ == Behavior::kStringToBigInt && this->sign() != Sign::kNone &&
@@ -1009,7 +961,8 @@ class StringToBigIntHelper : public StringToIntHelper<IsolateT> {
       case State::kZero:
         return BigInt::Zero(this->isolate(), allocation_type());
       case State::kDone:
-        return BigInt::Finalize<Isolate>(result_, this->negative());
+        return BigInt::Allocate(this->isolate(), &accumulator_,
+                                this->negative(), allocation_type());
       case State::kEmpty:
       case State::kRunning:
         break;
@@ -1017,27 +970,44 @@ class StringToBigIntHelper : public StringToIntHelper<IsolateT> {
     UNREACHABLE();
   }
 
- protected:
-  void AllocateResult() override {
-    // We have to allocate a BigInt that's big enough to fit the result.
-    // Conseratively assume that all remaining digits are significant.
-    // Optimization opportunity: Would it makes sense to scan for trailing
-    // junk before allocating the result?
-    int charcount = this->length() - this->cursor();
-    MaybeHandle<FreshlyAllocatedBigInt> maybe =
-        BigInt::AllocateFor(this->isolate(), this->radix(), charcount,
-                            kDontThrow, allocation_type());
-    if (!maybe.ToHandle(&result_)) {
-      this->set_state(State::kError);
+  // Used for converting BigInt literals. The scanner has already checked
+  // that the literal is valid and not too big, so this always succeeds.
+  std::unique_ptr<char[]> DecimalString(bigint::Processor* processor) {
+    DCHECK_EQ(behavior_, Behavior::kLiteral);
+    this->ParseInt();
+    if (this->state() == State::kZero) {
+      // Input may have been "0x0" or similar.
+      return std::unique_ptr<char[]>(new char[2]{'0', '\0'});
     }
+    DCHECK_EQ(this->state(), State::kDone);
+    int num_digits = accumulator_.ResultLength();
+    base::SmallVector<bigint::digit_t, 8> digit_storage(num_digits);
+    bigint::RWDigits digits(digit_storage.data(), num_digits);
+    processor->FromString(digits, &accumulator_);
+    int num_chars = bigint::ToStringResultLength(digits, 10, false);
+    std::unique_ptr<char[]> out(new char[num_chars + 1]);
+    processor->ToString(out.get(), &num_chars, digits, 10, false);
+    out[num_chars] = '\0';
+    return out;
   }
 
-  void ResultMultiplyAdd(uint32_t multiplier, uint32_t part) override {
-    BigInt::InplaceMultiplyAdd(*result_, static_cast<uintptr_t>(multiplier),
-                               static_cast<uintptr_t>(part));
-  }
+ private:
+  template <class Char>
+  void ParseInternal(Char start) {
+    using Result = bigint::FromStringAccumulator::Result;
+    Char current = start + this->cursor();
+    Char end = start + this->length();
+    current = accumulator_.Parse(current, end, this->radix());
 
-  bool CheckTermination() override;
+    Result result = accumulator_.result();
+    if (result == Result::kMaxSizeExceeded) {
+      return this->set_state(State::kError);
+    }
+    if (!this->allow_trailing_junk() && AdvanceToNonspace(&current, end)) {
+      return this->set_state(State::kJunk);
+    }
+    return this->set_state(State::kDone);
+  }
 
   AllocationType allocation_type() {
     // For literals, we pretenure the allocated BigInt, since it's about
@@ -1046,23 +1016,10 @@ class StringToBigIntHelper : public StringToIntHelper<IsolateT> {
                                            : AllocationType::kYoung;
   }
 
- private:
-  Handle<FreshlyAllocatedBigInt> result_;
+  bigint::FromStringAccumulator accumulator_{BigInt::kMaxLength};
   Behavior behavior_;
 };
 
-template <typename IsolateT>
-bool StringToBigIntHelper<IsolateT>::CheckTermination() {
-  return false;
-}
-
-template <>
-bool StringToBigIntHelper<Isolate>::CheckTermination() {
-  StackLimitCheck interrupt_check(isolate());
-  return interrupt_check.InterruptRequested() &&
-         isolate()->stack_guard()->HandleInterrupts().IsException(isolate());
-}
-
 MaybeHandle<BigInt> StringToBigInt(Isolate* isolate, Handle<String> string) {
   string = String::Flatten(isolate, string);
   StringToBigIntHelper<Isolate> helper(isolate, string);
@@ -1082,6 +1039,13 @@ template EXPORT_TEMPLATE_DEFINE(V8_EXPORT_PRIVATE)
     MaybeHandle<BigInt> BigIntLiteral(LocalIsolate* isolate,
                                       const char* string);
 
+std::unique_ptr<char[]> BigIntLiteralToDecimal(
+    LocalIsolate* isolate, base::Vector<const uint8_t> literal) {
+  StringToBigIntHelper<LocalIsolate> helper(nullptr, literal.begin(),
+                                            literal.length());
+  return helper.DecimalString(isolate->bigint_processor());
+}
+
 const char* DoubleToCString(double v, base::Vector<char> buffer) {
   switch (FPCLASSIFY_NAMESPACE::fpclassify(v)) {
     case FP_NAN:
@@ -1428,7 +1392,7 @@ char* DoubleToRadixCString(double value, int radix) {
             }
             char c = buffer[fraction_cursor];
             // Reconstruct digit.
-            int digit = c > '9' ? (c - 'a' + 10) : (c - '0');
+            digit = c > '9' ? (c - 'a' + 10) : (c - '0');
             if (digit + 1 < radix) {
               buffer[fraction_cursor++] = chars[digit + 1];
               break;
@@ -1490,7 +1454,7 @@ base::Optional<double> TryStringToDouble(LocalIsolate* isolate,
   const int flags = ALLOW_HEX | ALLOW_OCTAL | ALLOW_BINARY;
   auto buffer = std::make_unique<base::uc16[]>(max_length_for_conversion);
   SharedStringAccessGuardIfNeeded access_guard(isolate);
-  String::WriteToFlat(*object, buffer.get(), 0, length, access_guard);
+  String::WriteToFlat(*object, buffer.get(), 0, length, isolate, access_guard);
   base::Vector<const base::uc16> v(buffer.get(), length);
   return StringToDouble(v, flags);
 }
@@ -1541,7 +1505,7 @@ bool IsSpecialIndex(String string) {
   }
   // Slow path: test DoubleToString(StringToDouble(string)) == string.
   base::Vector<const uint16_t> vector(buffer, length);
-  double d = StringToDouble(vector, NO_FLAGS);
+  double d = StringToDouble(vector, NO_CONVERSION_FLAGS);
   if (std::isnan(d)) return false;
   // Compute reverse string.
   char reverse_buffer[kBufferSize + 1];  // Result will be /0 terminated.
