@@ -5,13 +5,12 @@
 #ifndef V8_OBJECTS_FIXED_ARRAY_INL_H_
 #define V8_OBJECTS_FIXED_ARRAY_INL_H_
 
-#include "src/objects/fixed-array.h"
-
 #include "src/handles/handles-inl.h"
 #include "src/heap/heap-write-barrier-inl.h"
 #include "src/numbers/conversions.h"
 #include "src/objects/bigint.h"
 #include "src/objects/compressed-slots.h"
+#include "src/objects/fixed-array.h"
 #include "src/objects/map.h"
 #include "src/objects/maybe-object-inl.h"
 #include "src/objects/objects-inl.h"
@@ -84,7 +83,7 @@ bool FixedArray::is_the_hole(Isolate* isolate, int index) {
   return get(isolate, index).IsTheHole(isolate);
 }
 
-#if !defined(_WIN32) || defined(_WIN64)
+#if !defined(_WIN32) || (defined(_WIN64) && _MSC_VER < 1930 && __cplusplus < 201703L)
 void FixedArray::set(int index, Smi value) {
   DCHECK_NE(map(), GetReadOnlyRoots().fixed_cow_array_map());
   DCHECK_LT(static_cast<unsigned>(index), static_cast<unsigned>(length()));
@@ -145,6 +144,30 @@ void FixedArray::set(int index, Smi value, RelaxedStoreTag tag) {
   set(index, value, tag, SKIP_WRITE_BARRIER);
 }
 
+Object FixedArray::get(int index, SeqCstAccessTag) const {
+  PtrComprCageBase cage_base = GetPtrComprCageBase(*this);
+  return get(cage_base, index);
+}
+
+Object FixedArray::get(PtrComprCageBase cage_base, int index,
+                       SeqCstAccessTag) const {
+  DCHECK_LT(static_cast<unsigned>(index), static_cast<unsigned>(length()));
+  return SEQ_CST_READ_FIELD(*this, OffsetOfElementAt(index));
+}
+
+void FixedArray::set(int index, Object value, SeqCstAccessTag,
+                     WriteBarrierMode mode) {
+  DCHECK_NE(map(), GetReadOnlyRoots().fixed_cow_array_map());
+  DCHECK_LT(static_cast<unsigned>(index), static_cast<unsigned>(length()));
+  SEQ_CST_WRITE_FIELD(*this, OffsetOfElementAt(index), value);
+  CONDITIONAL_WRITE_BARRIER(*this, OffsetOfElementAt(index), value, mode);
+}
+
+void FixedArray::set(int index, Smi value, SeqCstAccessTag tag) {
+  DCHECK(Object(value).IsSmi());
+  set(index, value, tag, SKIP_WRITE_BARRIER);
+}
+
 Object FixedArray::get(int index, AcquireLoadTag) const {
   PtrComprCageBase cage_base = GetPtrComprCageBase(*this);
   return get(cage_base, index);
@@ -203,6 +226,21 @@ void FixedArray::set_the_hole(ReadOnlyRoots ro_roots, int index) {
   FixedArray::NoWriteBarrierSet(*this, index, ro_roots.the_hole_value());
 }
 
+Object FixedArray::swap(int index, Object value, SeqCstAccessTag,
+                        WriteBarrierMode mode) {
+  DCHECK_NE(map(), GetReadOnlyRoots().fixed_cow_array_map());
+  DCHECK_LT(static_cast<unsigned>(index), static_cast<unsigned>(length()));
+  Object previous_value =
+      SEQ_CST_SWAP_FIELD(*this, OffsetOfElementAt(index), value);
+  CONDITIONAL_WRITE_BARRIER(*this, OffsetOfElementAt(index), value, mode);
+  return previous_value;
+}
+
+Object FixedArray::swap(int index, Smi value, SeqCstAccessTag tag) {
+  DCHECK(Object(value).IsSmi());
+  return swap(index, value, tag, SKIP_WRITE_BARRIER);
+}
+
 void FixedArray::FillWithHoles(int from, int to) {
   for (int i = from; i < to; i++) {
     set_the_hole(i);
@@ -446,6 +484,16 @@ void WeakFixedArray::Set(int index, MaybeObject value, WriteBarrierMode mode) {
   set_objects(index, value, mode);
 }
 
+Handle<WeakFixedArray> WeakFixedArray::EnsureSpace(Isolate* isolate,
+                                                   Handle<WeakFixedArray> array,
+                                                   int length) {
+  if (array->length() < length) {
+    int grow_by = length - array->length();
+    array = isolate->factory()->CopyWeakFixedArrayAndGrow(array, grow_by);
+  }
+  return array;
+}
+
 MaybeObjectSlot WeakFixedArray::data_start() {
   return RawMaybeWeakField(kObjectsOffset);
 }
@@ -481,6 +529,10 @@ void WeakArrayList::Set(int index, MaybeObject value, WriteBarrierMode mode) {
   set_objects(index, value, mode);
 }
 
+void WeakArrayList::Set(int index, Smi value) {
+  Set(index, MaybeObject::FromSmi(value), SKIP_WRITE_BARRIER);
+}
+
 MaybeObjectSlot WeakArrayList::data_start() {
   return RawMaybeWeakField(kObjectsOffset);
 }
@@ -547,58 +599,60 @@ void ArrayList::Clear(int index, Object undefined) {
 
 int ByteArray::Size() { return RoundUp(length() + kHeaderSize, kTaggedSize); }
 
-byte ByteArray::get(int index) const {
-  DCHECK(index >= 0 && index < this->length());
-  return ReadField<byte>(kHeaderSize + index * kCharSize);
+byte ByteArray::get(int offset) const {
+  DCHECK_GE(offset, 0);
+  DCHECK_LT(offset, length());
+  return ReadField<byte>(kHeaderSize + offset);
 }
 
-void ByteArray::set(int index, byte value) {
-  DCHECK(index >= 0 && index < this->length());
-  WriteField<byte>(kHeaderSize + index * kCharSize, value);
+void ByteArray::set(int offset, byte value) {
+  DCHECK_GE(offset, 0);
+  DCHECK_LT(offset, length());
+  WriteField<byte>(kHeaderSize + offset, value);
 }
 
-void ByteArray::copy_in(int index, const byte* buffer, int length) {
-  DCHECK(index >= 0 && length >= 0 && length <= kMaxInt - index &&
-         index + length <= this->length());
-  Address dst_addr = field_address(kHeaderSize + index * kCharSize);
-  memcpy(reinterpret_cast<void*>(dst_addr), buffer, length);
+int ByteArray::get_int(int offset) const {
+  DCHECK_GE(offset, 0);
+  DCHECK_LE(offset + sizeof(int), length());
+  return ReadField<int>(kHeaderSize + offset);
 }
 
-void ByteArray::copy_out(int index, byte* buffer, int length) {
-  DCHECK(index >= 0 && length >= 0 && length <= kMaxInt - index &&
-         index + length <= this->length());
-  Address src_addr = field_address(kHeaderSize + index * kCharSize);
-  memcpy(buffer, reinterpret_cast<void*>(src_addr), length);
+void ByteArray::set_int(int offset, int value) {
+  DCHECK_GE(offset, 0);
+  DCHECK_LE(offset + sizeof(int), length());
+  WriteField<int>(kHeaderSize + offset, value);
 }
 
-int ByteArray::get_int(int index) const {
-  DCHECK(index >= 0 && index < this->length() / kIntSize);
-  return ReadField<int>(kHeaderSize + index * kIntSize);
+Address ByteArray::get_sandboxed_pointer(int offset) const {
+  DCHECK_GE(offset, 0);
+  DCHECK_LE(offset + sizeof(Address), length());
+  PtrComprCageBase sandbox_base = GetPtrComprCageBase(*this);
+  return ReadSandboxedPointerField(kHeaderSize + offset, sandbox_base);
 }
 
-void ByteArray::set_int(int index, int value) {
-  DCHECK(index >= 0 && index < this->length() / kIntSize);
-  WriteField<int>(kHeaderSize + index * kIntSize, value);
+void ByteArray::set_sandboxed_pointer(int offset, Address value) {
+  DCHECK_GE(offset, 0);
+  DCHECK_LE(offset + sizeof(Address), length());
+  PtrComprCageBase sandbox_base = GetPtrComprCageBase(*this);
+  WriteSandboxedPointerField(kHeaderSize + offset, sandbox_base, value);
 }
 
-uint32_t ByteArray::get_uint32(int index) const {
-  DCHECK(index >= 0 && index < this->length() / kUInt32Size);
-  return ReadField<uint32_t>(kHeaderSize + index * kUInt32Size);
+void ByteArray::copy_in(int offset, const byte* buffer, int slice_length) {
+  DCHECK_GE(offset, 0);
+  DCHECK_GE(slice_length, 0);
+  DCHECK_LE(slice_length, kMaxInt - offset);
+  DCHECK_LE(offset + slice_length, length());
+  Address dst_addr = field_address(kHeaderSize + offset);
+  memcpy(reinterpret_cast<void*>(dst_addr), buffer, slice_length);
 }
 
-void ByteArray::set_uint32(int index, uint32_t value) {
-  DCHECK(index >= 0 && index < this->length() / kUInt32Size);
-  WriteField<uint32_t>(kHeaderSize + index * kUInt32Size, value);
-}
-
-uint32_t ByteArray::get_uint32_relaxed(int index) const {
-  DCHECK(index >= 0 && index < this->length() / kUInt32Size);
-  return RELAXED_READ_UINT32_FIELD(*this, kHeaderSize + index * kUInt32Size);
-}
-
-void ByteArray::set_uint32_relaxed(int index, uint32_t value) {
-  DCHECK(index >= 0 && index < this->length() / kUInt32Size);
-  RELAXED_WRITE_UINT32_FIELD(*this, kHeaderSize + index * kUInt32Size, value);
+void ByteArray::copy_out(int offset, byte* buffer, int slice_length) {
+  DCHECK_GE(offset, 0);
+  DCHECK_GE(slice_length, 0);
+  DCHECK_LE(slice_length, kMaxInt - offset);
+  DCHECK_LE(offset + slice_length, length());
+  Address src_addr = field_address(kHeaderSize + offset);
+  memcpy(buffer, reinterpret_cast<void*>(src_addr), slice_length);
 }
 
 void ByteArray::clear_padding() {
@@ -613,8 +667,6 @@ ByteArray ByteArray::FromDataStartAddress(Address address) {
 
 int ByteArray::DataSize() const { return RoundUp(length(), kTaggedSize); }
 
-int ByteArray::ByteArraySize() { return SizeFor(this->length()); }
-
 byte* ByteArray::GetDataStartAddress() {
   return reinterpret_cast<byte*>(address() + kHeaderSize);
 }
@@ -623,6 +675,48 @@ byte* ByteArray::GetDataEndAddress() {
   return GetDataStartAddress() + length();
 }
 
+template <typename T>
+FixedIntegerArray<T>::FixedIntegerArray(Address ptr) : ByteArray(ptr) {
+  DCHECK_EQ(ByteArray::length() % sizeof(T), 0);
+}
+
+template <typename T>
+FixedIntegerArray<T> FixedIntegerArray<T>::cast(Object object) {
+  return FixedIntegerArray<T>(object.ptr());
+}
+
+// static
+template <typename T>
+Handle<FixedIntegerArray<T>> FixedIntegerArray<T>::New(
+    Isolate* isolate, int length, AllocationType allocation) {
+  int byte_length;
+  CHECK(!base::bits::SignedMulOverflow32(length, sizeof(T), &byte_length));
+  return Handle<FixedIntegerArray<T>>::cast(
+      isolate->factory()->NewByteArray(byte_length, allocation));
+}
+
+template <typename T>
+T FixedIntegerArray<T>::get(int index) const {
+  static_assert(std::is_integral<T>::value);
+  DCHECK_GE(index, 0);
+  DCHECK_LT(index, length());
+  return ReadField<T>(kHeaderSize + index * sizeof(T));
+}
+
+template <typename T>
+void FixedIntegerArray<T>::set(int index, T value) {
+  static_assert(std::is_integral<T>::value);
+  DCHECK_GE(index, 0);
+  DCHECK_LT(index, length());
+  WriteField<T>(kHeaderSize + index * sizeof(T), value);
+}
+
+template <typename T>
+int FixedIntegerArray<T>::length() const {
+  DCHECK_EQ(ByteArray::length() % sizeof(T), 0);
+  return ByteArray::length() / sizeof(T);
+}
+
 template <class T>
 PodArray<T>::PodArray(Address ptr) : ByteArray(ptr) {}
 
@@ -635,8 +729,10 @@ PodArray<T> PodArray<T>::cast(Object object) {
 template <class T>
 Handle<PodArray<T>> PodArray<T>::New(Isolate* isolate, int length,
                                      AllocationType allocation) {
+  int byte_length;
+  CHECK(!base::bits::SignedMulOverflow32(length, sizeof(T), &byte_length));
   return Handle<PodArray<T>>::cast(
-      isolate->factory()->NewByteArray(length * sizeof(T), allocation));
+      isolate->factory()->NewByteArray(byte_length, allocation));
 }
 
 template <class T>
@@ -663,7 +759,6 @@ void TemplateList::set(int index, Object value) {
 }  // namespace internal
 }  // namespace v8
 
-#include "src/base/platform/wrappers.h"
 #include "src/objects/object-macros-undef.h"
 
 #endif  // V8_OBJECTS_FIXED_ARRAY_INL_H_
