@@ -34,6 +34,7 @@ namespace internal {
 Handle<String> String::SlowFlatten(Isolate* isolate, Handle<ConsString> cons,
                                    AllocationType allocation) {
   DCHECK_NE(cons->second().length(), 0);
+  DCHECK(!cons->InSharedHeap());
 
   // TurboFan can create cons strings with empty first parts.
   while (cons->first().length() == 0) {
@@ -43,20 +44,31 @@ Handle<String> String::SlowFlatten(Isolate* isolate, Handle<ConsString> cons,
     if (cons->second().IsConsString() && !cons->second().IsFlat()) {
       cons = handle(ConsString::cast(cons->second()), isolate);
     } else {
-      return String::Flatten(isolate, handle(cons->second(), isolate));
+      return String::Flatten(isolate, handle(cons->second(), isolate),
+                             allocation);
     }
   }
 
   DCHECK(AllowGarbageCollection::IsAllowed());
   int length = cons->length();
-  allocation =
-      ObjectInYoungGeneration(*cons) ? allocation : AllocationType::kOld;
+  if (allocation != AllocationType::kSharedOld) {
+    allocation =
+        ObjectInYoungGeneration(*cons) ? allocation : AllocationType::kOld;
+  }
   Handle<SeqString> result;
   if (cons->IsOneByteRepresentation()) {
     Handle<SeqOneByteString> flat =
         isolate->factory()
             ->NewRawOneByteString(length, allocation)
             .ToHandleChecked();
+    // When the ConsString had a forwarding index, it is possible that it was
+    // transitioned to a ThinString (and eventually shortcutted to
+    // InternalizedString) during GC.
+    if (V8_UNLIKELY(v8_flags.always_use_string_forwarding_table &&
+                    !cons->IsConsString())) {
+      DCHECK(cons->IsInternalizedString() || cons->IsThinString());
+      return String::Flatten(isolate, cons, allocation);
+    }
     DisallowGarbageCollection no_gc;
     WriteToFlat(*cons, flat->GetChars(no_gc), 0, length);
     result = flat;
@@ -65,16 +77,64 @@ Handle<String> String::SlowFlatten(Isolate* isolate, Handle<ConsString> cons,
         isolate->factory()
             ->NewRawTwoByteString(length, allocation)
             .ToHandleChecked();
+    // When the ConsString had a forwarding index, it is possible that it was
+    // transitioned to a ThinString (and eventually shortcutted to
+    // InternalizedString) during GC.
+    if (V8_UNLIKELY(v8_flags.always_use_string_forwarding_table &&
+                    !cons->IsConsString())) {
+      DCHECK(cons->IsInternalizedString() || cons->IsThinString());
+      return String::Flatten(isolate, cons, allocation);
+    }
     DisallowGarbageCollection no_gc;
     WriteToFlat(*cons, flat->GetChars(no_gc), 0, length);
     result = flat;
   }
-  cons->set_first(*result);
-  cons->set_second(ReadOnlyRoots(isolate).empty_string());
+  {
+    DisallowGarbageCollection no_gc;
+    auto raw_cons = *cons;
+    raw_cons.set_first(*result);
+    raw_cons.set_second(ReadOnlyRoots(isolate).empty_string());
+  }
   DCHECK(result->IsFlat());
   return result;
 }
 
+Handle<String> String::SlowShare(Isolate* isolate, Handle<String> source) {
+  DCHECK(v8_flags.shared_string_table);
+  Handle<String> flat = Flatten(isolate, source, AllocationType::kSharedOld);
+
+  // Do not recursively call Share, so directly compute the sharing strategy for
+  // the flat string, which could already be a copy or an existing string from
+  // e.g. a shortcut ConsString.
+  MaybeHandle<Map> new_map;
+  switch (isolate->factory()->ComputeSharingStrategyForString(flat, &new_map)) {
+    case StringTransitionStrategy::kCopy:
+      break;
+    case StringTransitionStrategy::kInPlace:
+      // A relaxed write is sufficient here, because at this point the string
+      // has not yet escaped the current thread.
+      DCHECK(flat->InSharedHeap());
+      flat->set_map_no_write_barrier(*new_map.ToHandleChecked());
+      return flat;
+    case StringTransitionStrategy::kAlreadyTransitioned:
+      return flat;
+  }
+
+  int length = flat->length();
+  if (flat->IsOneByteRepresentation()) {
+    Handle<SeqOneByteString> copy =
+        isolate->factory()->NewRawSharedOneByteString(length).ToHandleChecked();
+    DisallowGarbageCollection no_gc;
+    WriteToFlat(*flat, copy->GetChars(no_gc), 0, length);
+    return copy;
+  }
+  Handle<SeqTwoByteString> copy =
+      isolate->factory()->NewRawSharedTwoByteString(length).ToHandleChecked();
+  DisallowGarbageCollection no_gc;
+  WriteToFlat(*flat, copy->GetChars(no_gc), 0, length);
+  return copy;
+}
+
 namespace {
 
 template <class StringClass>
@@ -112,6 +172,29 @@ void MigrateExternalString(Isolate* isolate, String string,
   }
 }
 
+template <typename IsolateT>
+Map ComputeThinStringMap(IsolateT* isolate, StringShape from_string_shape,
+                         bool one_byte) {
+  ReadOnlyRoots roots(isolate);
+  if (from_string_shape.IsShared()) {
+    return one_byte ? roots.shared_thin_one_byte_string_map()
+                    : roots.shared_thin_string_map();
+  }
+  return one_byte ? roots.thin_one_byte_string_map() : roots.thin_string_map();
+}
+
+void InitExternalPointerFieldsDuringExternalization(String string, Map new_map,
+                                                    Isolate* isolate) {
+  string.InitExternalPointerField<kExternalStringResourceTag>(
+      ExternalString::kResourceOffset, isolate, kNullAddress);
+  bool is_uncached = (new_map.instance_type() & kUncachedExternalStringMask) ==
+                     kUncachedExternalStringTag;
+  if (!is_uncached) {
+    string.InitExternalPointerField<kExternalStringResourceDataTag>(
+        ExternalString::kResourceDataOffset, isolate, kNullAddress);
+  }
+}
+
 }  // namespace
 
 template <typename IsolateT>
@@ -120,29 +203,42 @@ void String::MakeThin(IsolateT* isolate, String internalized) {
   DCHECK_NE(*this, internalized);
   DCHECK(internalized.IsInternalizedString());
 
-  if (this->IsExternalString()) {
+  Map initial_map = map(kAcquireLoad);
+  StringShape initial_shape(initial_map);
+
+  DCHECK(!initial_shape.IsThin());
+
+#ifdef DEBUG
+  // Check that shared strings can only transition to ThinStrings on the main
+  // thread when no other thread is active.
+  // The exception is during serialization, as no strings have escaped the
+  // thread yet.
+  if (initial_shape.IsShared() && !isolate->has_active_deserializer()) {
+    isolate->AsIsolate()->global_safepoint()->AssertActive();
+  }
+#endif
+
+  bool has_pointers = initial_shape.IsIndirect();
+  int old_size = SizeFromMap(initial_map);
+  Map target_map = ComputeThinStringMap(isolate, initial_shape,
+                                        internalized.IsOneByteRepresentation());
+  if (initial_shape.IsExternal()) {
     MigrateExternalString(isolate->AsIsolate(), *this, internalized);
   }
 
-  bool has_pointers = StringShape(*this).IsIndirect();
-
-  int old_size = this->Size();
-  bool one_byte = internalized.IsOneByteRepresentation();
-  Handle<Map> map = one_byte ? isolate->factory()->thin_one_byte_string_map()
-                             : isolate->factory()->thin_string_map();
   // Update actual first and then do release store on the map word. This ensures
   // that the concurrent marker will read the pointer when visiting a
   // ThinString.
   ThinString thin = ThinString::unchecked_cast(*this);
   thin.set_actual(internalized);
+  set_map_safe_transition(target_map, kReleaseStore);
+
   DCHECK_GE(old_size, ThinString::kSize);
-  this->set_map(*map, kReleaseStore);
-  Address thin_end = thin.address() + ThinString::kSize;
   int size_delta = old_size - ThinString::kSize;
   if (size_delta != 0) {
     if (!Heap::IsLargeObject(thin)) {
-      isolate->heap()->CreateFillerObjectAt(
-          thin_end, size_delta,
+      isolate->heap()->NotifyObjectSizeChange(
+          thin, old_size, ThinString::kSize,
           has_pointers ? ClearRecordedSlots::kYes : ClearRecordedSlots::kNo);
     } else {
       // We don't need special handling for the combination IsLargeObject &&
@@ -152,8 +248,143 @@ void String::MakeThin(IsolateT* isolate, String internalized) {
   }
 }
 
-template void String::MakeThin(Isolate* isolate, String internalized);
-template void String::MakeThin(LocalIsolate* isolate, String internalized);
+template EXPORT_TEMPLATE_DEFINE(V8_EXPORT_PRIVATE) void String::MakeThin(
+    Isolate* isolate, String internalized);
+template EXPORT_TEMPLATE_DEFINE(V8_EXPORT_PRIVATE) void String::MakeThin(
+    LocalIsolate* isolate, String internalized);
+
+template <typename T>
+bool String::MarkForExternalizationDuringGC(Isolate* isolate, T* resource) {
+  uint32_t raw_hash = raw_hash_field(kAcquireLoad);
+  if (IsExternalForwardingIndex(raw_hash)) return false;
+  if (IsInternalizedForwardingIndex(raw_hash)) {
+    const int forwarding_index = ForwardingIndexValueBits::decode(raw_hash);
+    if (!isolate->string_forwarding_table()->TryUpdateExternalResource(
+            forwarding_index, resource)) {
+      // The external resource was concurrently updated by another thread.
+      return false;
+    }
+    raw_hash = Name::IsExternalForwardingIndexBit::update(raw_hash, true);
+    set_raw_hash_field(raw_hash, kReleaseStore);
+    return true;
+  }
+  // We need to store the hash in the forwarding table, as all non-external
+  // shared strings are in-place internalizable. In case the string gets
+  // internalized, we have to ensure that we can get the hash from the
+  // forwarding table to satisfy the invariant that all internalized strings
+  // have a computed hash value.
+  if (!IsHashFieldComputed(raw_hash)) {
+    raw_hash = EnsureRawHash();
+  }
+  DCHECK(IsHashFieldComputed(raw_hash));
+  int forwarding_index =
+      isolate->string_forwarding_table()->AddExternalResourceAndHash(
+          *this, resource, raw_hash);
+  set_raw_hash_field(String::CreateExternalForwardingIndex(forwarding_index),
+                     kReleaseStore);
+
+  return true;
+}
+
+namespace {
+
+template <bool is_one_byte>
+Map ComputeExternalStringMap(Isolate* isolate, String string, int size) {
+  ReadOnlyRoots roots(isolate);
+  StringShape shape(string, isolate);
+  const bool is_internalized = shape.IsInternalized();
+  const bool is_shared = shape.IsShared();
+  if constexpr (is_one_byte) {
+    if (size < ExternalString::kSizeOfAllExternalStrings) {
+      if (is_internalized) {
+        return roots.uncached_external_one_byte_internalized_string_map();
+      } else {
+        return is_shared ? roots.shared_uncached_external_one_byte_string_map()
+                         : roots.uncached_external_one_byte_string_map();
+      }
+    } else {
+      if (is_internalized) {
+        return roots.external_one_byte_internalized_string_map();
+      } else {
+        return is_shared ? roots.shared_external_one_byte_string_map()
+                         : roots.external_one_byte_string_map();
+      }
+    }
+  } else {
+    if (size < ExternalString::kSizeOfAllExternalStrings) {
+      if (is_internalized) {
+        return roots.uncached_external_internalized_string_map();
+      } else {
+        return is_shared ? roots.shared_uncached_external_string_map()
+                         : roots.uncached_external_string_map();
+      }
+    } else {
+      if (is_internalized) {
+        return roots.external_internalized_string_map();
+      } else {
+        return is_shared ? roots.shared_external_string_map()
+                         : roots.external_string_map();
+      }
+    }
+  }
+}
+
+}  // namespace
+
+template <typename T>
+void String::MakeExternalDuringGC(Isolate* isolate, T* resource) {
+  isolate->heap()->safepoint()->AssertActive();
+  DCHECK_NE(isolate->heap()->gc_state(), Heap::NOT_IN_GC);
+
+  constexpr bool is_one_byte =
+      std::is_base_of_v<v8::String::ExternalOneByteStringResource, T>;
+  int size = this->Size();  // Byte size of the original string.
+  DCHECK_GE(size, ExternalString::kUncachedSize);
+
+  // Morph the string to an external string by replacing the map and
+  // reinitializing the fields.  This won't work if the space the existing
+  // string occupies is too small for a regular external string.  Instead, we
+  // resort to an uncached external string instead, omitting the field caching
+  // the address of the backing store.  When we encounter uncached external
+  // strings in generated code, we need to bailout to runtime.
+  Map new_map = ComputeExternalStringMap<is_one_byte>(isolate, *this, size);
+
+  // Byte size of the external String object.
+  int new_size = this->SizeFromMap(new_map);
+
+  // Shared strings are never indirect or large.
+  DCHECK(!isolate->heap()->IsLargeObject(*this));
+  DCHECK(!StringShape(*this).IsIndirect());
+
+  isolate->heap()->NotifyObjectSizeChange(*this, size, new_size,
+                                          ClearRecordedSlots::kNo);
+
+  // The external pointer slots must be initialized before the new map is
+  // installed. Otherwise, a GC marking thread may see the new map before the
+  // slots are initialized and attempt to mark the (invalid) external pointers
+  // table entries as alive.
+  InitExternalPointerFieldsDuringExternalization(*this, new_map, isolate);
+
+  // We are storing the new map using release store after creating a filler in
+  // the NotifyObjectSizeChange call for the left-over space to avoid races with
+  // the sweeper thread.
+  this->set_map(new_map, kReleaseStore);
+
+  if constexpr (is_one_byte) {
+    ExternalOneByteString self = ExternalOneByteString::cast(*this);
+    self.SetResource(isolate, resource);
+  } else {
+    ExternalTwoByteString self = ExternalTwoByteString::cast(*this);
+    self.SetResource(isolate, resource);
+  }
+  isolate->heap()->RegisterExternalString(*this);
+}
+
+template EXPORT_TEMPLATE_DEFINE(V8_EXPORT_PRIVATE) void String::
+    MakeExternalDuringGC(Isolate* isolate,
+                         v8::String::ExternalOneByteStringResource*);
+template EXPORT_TEMPLATE_DEFINE(V8_EXPORT_PRIVATE) void String::
+    MakeExternalDuringGC(Isolate* isolate, v8::String::ExternalStringResource*);
 
 bool String::MakeExternal(v8::String::ExternalStringResource* resource) {
   // Disallow garbage collection to avoid possible GC vs string access deadlock.
@@ -164,7 +395,7 @@ bool String::MakeExternal(v8::String::ExternalStringResource* resource) {
   DCHECK(this->SupportsExternalization());
   DCHECK(resource->IsCacheable());
 #ifdef ENABLE_SLOW_DCHECKS
-  if (FLAG_enable_slow_asserts) {
+  if (v8_flags.enable_slow_asserts) {
     // Assert that the resource and the string are equivalent.
     DCHECK(static_cast<size_t>(this->length()) == resource->length());
     base::ScopedVector<base::uc16> smart_chars(this->length());
@@ -180,14 +411,12 @@ bool String::MakeExternal(v8::String::ExternalStringResource* resource) {
   // string.
   if (IsReadOnlyHeapObject(*this)) return false;
   Isolate* isolate = GetIsolateFromWritableObject(*this);
+  if (IsShared(isolate)) {
+    return MarkForExternalizationDuringGC(isolate, resource);
+  }
   bool is_internalized = this->IsInternalizedString();
   bool has_pointers = StringShape(*this).IsIndirect();
 
-  if (has_pointers) {
-    isolate->heap()->NotifyObjectLayoutChange(*this, no_gc,
-                                              InvalidateRecordedSlots::kYes);
-  }
-
   base::SharedMutexGuard<base::kExclusive> shared_mutex_guard(
       isolate->internalized_string_access());
   // Morph the string to an external string by replacing the map and
@@ -196,24 +425,20 @@ bool String::MakeExternal(v8::String::ExternalStringResource* resource) {
   // resort to an uncached external string instead, omitting the field caching
   // the address of the backing store.  When we encounter uncached external
   // strings in generated code, we need to bailout to runtime.
-  Map new_map;
-  ReadOnlyRoots roots(isolate);
-  if (size < ExternalString::kSizeOfAllExternalStrings) {
-    if (is_internalized) {
-      new_map = roots.uncached_external_internalized_string_map();
-    } else {
-      new_map = roots.uncached_external_string_map();
-    }
-  } else {
-    new_map = is_internalized ? roots.external_internalized_string_map()
-                              : roots.external_string_map();
-  }
+  constexpr bool is_one_byte = false;
+  Map new_map = ComputeExternalStringMap<is_one_byte>(isolate, *this, size);
 
   // Byte size of the external String object.
   int new_size = this->SizeFromMap(new_map);
+
+  if (has_pointers) {
+    isolate->heap()->NotifyObjectLayoutChange(
+        *this, no_gc, InvalidateRecordedSlots::kYes, new_size);
+  }
+
   if (!isolate->heap()->IsLargeObject(*this)) {
-    isolate->heap()->CreateFillerObjectAt(
-        this->address() + new_size, size - new_size,
+    isolate->heap()->NotifyObjectSizeChange(
+        *this, size, new_size,
         has_pointers ? ClearRecordedSlots::kYes : ClearRecordedSlots::kNo);
   } else {
     // We don't need special handling for the combination IsLargeObject &&
@@ -221,12 +446,18 @@ bool String::MakeExternal(v8::String::ExternalStringResource* resource) {
     DCHECK(!has_pointers);
   }
 
-  // We are storing the new map using release store after creating a filler for
-  // the left-over space to avoid races with the sweeper thread.
+  // The external pointer slots must be initialized before the new map is
+  // installed. Otherwise, a GC marking thread may see the new map before the
+  // slots are initialized and attempt to mark the (invalid) external pointers
+  // table entries as alive.
+  InitExternalPointerFieldsDuringExternalization(*this, new_map, isolate);
+
+  // We are storing the new map using release store after creating a filler in
+  // the NotifyObjectSizeChange call for the left-over space to avoid races with
+  // the sweeper thread.
   this->set_map(new_map, kReleaseStore);
 
   ExternalTwoByteString self = ExternalTwoByteString::cast(*this);
-  self.AllocateExternalPointerEntries(isolate);
   self.SetResource(isolate, resource);
   isolate->heap()->RegisterExternalString(*this);
   // Force regeneration of the hash value.
@@ -243,7 +474,7 @@ bool String::MakeExternal(v8::String::ExternalOneByteStringResource* resource) {
   DCHECK(this->SupportsExternalization());
   DCHECK(resource->IsCacheable());
 #ifdef ENABLE_SLOW_DCHECKS
-  if (FLAG_enable_slow_asserts) {
+  if (v8_flags.enable_slow_asserts) {
     // Assert that the resource and the string are equivalent.
     DCHECK(static_cast<size_t>(this->length()) == resource->length());
     if (this->IsTwoByteRepresentation()) {
@@ -264,14 +495,12 @@ bool String::MakeExternal(v8::String::ExternalOneByteStringResource* resource) {
   // string.
   if (IsReadOnlyHeapObject(*this)) return false;
   Isolate* isolate = GetIsolateFromWritableObject(*this);
+  if (IsShared(isolate)) {
+    return MarkForExternalizationDuringGC(isolate, resource);
+  }
   bool is_internalized = this->IsInternalizedString();
   bool has_pointers = StringShape(*this).IsIndirect();
 
-  if (has_pointers) {
-    isolate->heap()->NotifyObjectLayoutChange(*this, no_gc,
-                                              InvalidateRecordedSlots::kYes);
-  }
-
   base::SharedMutexGuard<base::kExclusive> shared_mutex_guard(
       isolate->internalized_string_access());
   // Morph the string to an external string by replacing the map and
@@ -280,24 +509,20 @@ bool String::MakeExternal(v8::String::ExternalOneByteStringResource* resource) {
   // resort to an uncached external string instead, omitting the field caching
   // the address of the backing store.  When we encounter uncached external
   // strings in generated code, we need to bailout to runtime.
-  Map new_map;
-  ReadOnlyRoots roots(isolate);
-  if (size < ExternalString::kSizeOfAllExternalStrings) {
-    new_map = is_internalized
-                  ? roots.uncached_external_one_byte_internalized_string_map()
-                  : roots.uncached_external_one_byte_string_map();
-  } else {
-    new_map = is_internalized
-                  ? roots.external_one_byte_internalized_string_map()
-                  : roots.external_one_byte_string_map();
-  }
+  constexpr bool is_one_byte = true;
+  Map new_map = ComputeExternalStringMap<is_one_byte>(isolate, *this, size);
 
   if (!isolate->heap()->IsLargeObject(*this)) {
     // Byte size of the external String object.
     int new_size = this->SizeFromMap(new_map);
 
-    isolate->heap()->CreateFillerObjectAt(
-        this->address() + new_size, size - new_size,
+    if (has_pointers) {
+      isolate->heap()->NotifyObjectLayoutChange(
+          *this, no_gc, InvalidateRecordedSlots::kYes, new_size);
+    }
+
+    isolate->heap()->NotifyObjectSizeChange(
+        *this, size, new_size,
         has_pointers ? ClearRecordedSlots::kYes : ClearRecordedSlots::kNo);
   } else {
     // We don't need special handling for the combination IsLargeObject &&
@@ -305,12 +530,18 @@ bool String::MakeExternal(v8::String::ExternalOneByteStringResource* resource) {
     DCHECK(!has_pointers);
   }
 
-  // We are storing the new map using release store after creating a filler for
-  // the left-over space to avoid races with the sweeper thread.
+  // The external pointer slots must be initialized before the new map is
+  // installed. Otherwise, a GC marking thread may see the new map before the
+  // slots are initialized and attempt to mark the (invalid) external pointers
+  // table entries as alive.
+  InitExternalPointerFieldsDuringExternalization(*this, new_map, isolate);
+
+  // We are storing the new map using release store after creating a filler in
+  // the NotifyObjectSizeChange call for the left-over space to avoid races with
+  // the sweeper thread.
   this->set_map(new_map, kReleaseStore);
 
   ExternalOneByteString self = ExternalOneByteString::cast(*this);
-  self.AllocateExternalPointerEntries(isolate);
   self.SetResource(isolate, resource);
   isolate->heap()->RegisterExternalString(*this);
   // Force regeneration of the hash value.
@@ -347,7 +578,6 @@ bool String::SupportsExternalization() {
 const char* String::PrefixForDebugPrint() const {
   StringShape shape(*this);
   if (IsTwoByteRepresentation()) {
-    StringShape shape(*this);
     if (shape.IsInternalized()) {
       return "u#";
     } else if (shape.IsCons()) {
@@ -360,7 +590,6 @@ const char* String::PrefixForDebugPrint() const {
       return "u\"";
     }
   } else {
-    StringShape shape(*this);
     if (shape.IsInternalized()) {
       return "#";
     } else if (shape.IsCons()) {
@@ -524,7 +753,7 @@ Handle<Object> String::ToNumber(Isolate* isolate, Handle<String> subject) {
         subject->EnsureHash();  // Force hash calculation.
         DCHECK_EQ(subject->raw_hash_field(), raw_hash_field);
 #endif
-        subject->set_raw_hash_field(raw_hash_field);
+        subject->set_raw_hash_field_if_empty(raw_hash_field);
       }
       return handle(Smi::FromInt(d), isolate);
     }
@@ -535,63 +764,42 @@ Handle<Object> String::ToNumber(Isolate* isolate, Handle<String> subject) {
   return isolate->factory()->NewNumber(StringToDouble(isolate, subject, flags));
 }
 
-String::FlatContent String::GetFlatContent(
-    const DisallowGarbageCollection& no_gc) {
-#if DEBUG
-  // Check that this method is called only from the main thread.
-  {
-    Isolate* isolate;
-    // We don't have to check read only strings as those won't move.
-    DCHECK_IMPLIES(GetIsolateFromHeapObject(*this, &isolate),
-                   ThreadId::Current() == isolate->thread_id());
-  }
-#endif
+String::FlatContent String::SlowGetFlatContent(
+    const DisallowGarbageCollection& no_gc,
+    const SharedStringAccessGuardIfNeeded& access_guard) {
   USE(no_gc);
   PtrComprCageBase cage_base = GetPtrComprCageBase(*this);
-  int length = this->length();
-  StringShape shape(*this, cage_base);
   String string = *this;
+  StringShape shape(string, cage_base);
   int offset = 0;
-  if (shape.representation_tag() == kConsStringTag) {
+
+  // Extract cons- and sliced strings.
+  if (shape.IsCons()) {
     ConsString cons = ConsString::cast(string);
-    if (cons.second(cage_base).length() != 0) {
-      return FlatContent(no_gc);
-    }
+    if (!cons.IsFlat(cage_base)) return FlatContent(no_gc);
     string = cons.first(cage_base);
     shape = StringShape(string, cage_base);
-  } else if (shape.representation_tag() == kSlicedStringTag) {
+  } else if (shape.IsSliced()) {
     SlicedString slice = SlicedString::cast(string);
     offset = slice.offset();
     string = slice.parent(cage_base);
     shape = StringShape(string, cage_base);
-    DCHECK(shape.representation_tag() != kConsStringTag &&
-           shape.representation_tag() != kSlicedStringTag);
   }
-  if (shape.representation_tag() == kThinStringTag) {
+
+  DCHECK(!shape.IsCons());
+  DCHECK(!shape.IsSliced());
+
+  // Extract thin strings.
+  if (shape.IsThin()) {
     ThinString thin = ThinString::cast(string);
     string = thin.actual(cage_base);
     shape = StringShape(string, cage_base);
-    DCHECK(!shape.IsCons());
-    DCHECK(!shape.IsSliced());
-  }
-  if (shape.encoding_tag() == kOneByteStringTag) {
-    const uint8_t* start;
-    if (shape.representation_tag() == kSeqStringTag) {
-      start = SeqOneByteString::cast(string).GetChars(no_gc);
-    } else {
-      start = ExternalOneByteString::cast(string).GetChars(cage_base);
-    }
-    return FlatContent(start + offset, length, no_gc);
-  } else {
-    DCHECK_EQ(shape.encoding_tag(), kTwoByteStringTag);
-    const base::uc16* start;
-    if (shape.representation_tag() == kSeqStringTag) {
-      start = SeqTwoByteString::cast(string).GetChars(no_gc);
-    } else {
-      start = ExternalTwoByteString::cast(string).GetChars(cage_base);
-    }
-    return FlatContent(start + offset, length, no_gc);
   }
+
+  DCHECK(shape.IsDirect());
+  return TryGetFlatContentFromDirectString(cage_base, no_gc, string, offset,
+                                           length(), access_guard)
+      .value();
 }
 
 std::unique_ptr<char[]> String::ToCString(AllowNullsFlag allow_nulls,
@@ -664,7 +872,7 @@ void String::WriteToFlat(String source, sinkchar* sink, int start, int length,
     DCHECK_LT(0, length);
     DCHECK_LE(0, start);
     DCHECK_LE(length, source.length());
-    switch (StringShape(source, cage_base).full_representation_tag()) {
+    switch (StringShape(source, cage_base).representation_and_encoding_tag()) {
       case kOneByteStringTag | kExternalStringTag:
         CopyChars(
             sink,
@@ -846,10 +1054,12 @@ bool String::SlowEquals(
 
   // Fast check: if hash code is computed for both strings
   // a fast negative check can be performed.
-  if (HasHashCode() && other.HasHashCode()) {
+  uint32_t this_hash;
+  uint32_t other_hash;
+  if (TryGetHash(&this_hash) && other.TryGetHash(&other_hash)) {
 #ifdef ENABLE_SLOW_DCHECKS
-    if (FLAG_enable_slow_asserts) {
-      if (hash() != other.hash()) {
+    if (v8_flags.enable_slow_asserts) {
+      if (this_hash != other_hash) {
         bool found_difference = false;
         for (int i = 0; i < len; i++) {
           if (Get(i) != other.Get(i)) {
@@ -861,7 +1071,7 @@ bool String::SlowEquals(
       }
     }
 #endif
-    if (hash() != other.hash()) return false;
+    if (this_hash != other_hash) return false;
   }
 
   // We know the strings are both non-empty. Compare the first chars
@@ -886,26 +1096,30 @@ bool String::SlowEquals(
 bool String::SlowEquals(Isolate* isolate, Handle<String> one,
                         Handle<String> two) {
   // Fast check: negative check with lengths.
-  int one_length = one->length();
+  const int one_length = one->length();
   if (one_length != two->length()) return false;
   if (one_length == 0) return true;
 
   // Fast check: if at least one ThinString is involved, dereference it/them
   // and restart.
   if (one->IsThinString() || two->IsThinString()) {
-    if (one->IsThinString())
+    if (one->IsThinString()) {
       one = handle(ThinString::cast(*one).actual(), isolate);
-    if (two->IsThinString())
+    }
+    if (two->IsThinString()) {
       two = handle(ThinString::cast(*two).actual(), isolate);
+    }
     return String::Equals(isolate, one, two);
   }
 
   // Fast check: if hash code is computed for both strings
   // a fast negative check can be performed.
-  if (one->HasHashCode() && two->HasHashCode()) {
+  uint32_t one_hash;
+  uint32_t two_hash;
+  if (one->TryGetHash(&one_hash) && two->TryGetHash(&two_hash)) {
 #ifdef ENABLE_SLOW_DCHECKS
-    if (FLAG_enable_slow_asserts) {
-      if (one->hash() != two->hash()) {
+    if (v8_flags.enable_slow_asserts) {
+      if (one_hash != two_hash) {
         bool found_difference = false;
         for (int i = 0; i < one_length; i++) {
           if (one->Get(i) != two->Get(i)) {
@@ -917,7 +1131,7 @@ bool String::SlowEquals(Isolate* isolate, Handle<String> one,
       }
     }
 #endif
-    if (one->hash() != two->hash()) return false;
+    if (one_hash != two_hash) return false;
   }
 
   // We know the strings are both non-empty. Compare the first chars
@@ -934,12 +1148,17 @@ bool String::SlowEquals(Isolate* isolate, Handle<String> one,
   if (flat1.IsOneByte() && flat2.IsOneByte()) {
     return CompareCharsEqual(flat1.ToOneByteVector().begin(),
                              flat2.ToOneByteVector().begin(), one_length);
-  } else {
-    for (int i = 0; i < one_length; i++) {
-      if (flat1.Get(i) != flat2.Get(i)) return false;
-    }
-    return true;
+  } else if (flat1.IsTwoByte() && flat2.IsTwoByte()) {
+    return CompareCharsEqual(flat1.ToUC16Vector().begin(),
+                             flat2.ToUC16Vector().begin(), one_length);
+  } else if (flat1.IsOneByte() && flat2.IsTwoByte()) {
+    return CompareCharsEqual(flat1.ToOneByteVector().begin(),
+                             flat2.ToUC16Vector().begin(), one_length);
+  } else if (flat1.IsTwoByte() && flat2.IsOneByte()) {
+    return CompareCharsEqual(flat1.ToUC16Vector().begin(),
+                             flat2.ToOneByteVector().begin(), one_length);
   }
+  UNREACHABLE();
 }
 
 // static
@@ -1362,6 +1581,35 @@ bool String::HasOneBytePrefix(base::Vector<const char> str) {
 
 namespace {
 
+template <typename Char>
+bool IsIdentifierVector(const base::Vector<Char>& vec) {
+  if (vec.empty()) {
+    return false;
+  }
+  if (!IsIdentifierStart(vec[0])) {
+    return false;
+  }
+  for (size_t i = 1; i < vec.size(); ++i) {
+    if (!IsIdentifierPart(vec[i])) {
+      return false;
+    }
+  }
+  return true;
+}
+
+}  // namespace
+
+// static
+bool String::IsIdentifier(Isolate* isolate, Handle<String> str) {
+  str = String::Flatten(isolate, str);
+  DisallowGarbageCollection no_gc;
+  String::FlatContent flat = str->GetFlatContent(no_gc);
+  return flat.IsOneByte() ? IsIdentifierVector(flat.ToOneByteVector())
+                          : IsIdentifierVector(flat.ToUC16Vector());
+}
+
+namespace {
+
 template <typename Char>
 uint32_t HashString(String string, size_t start, int length, uint64_t seed,
                     PtrComprCageBase cage_base,
@@ -1391,15 +1639,21 @@ uint32_t HashString(String string, size_t start, int length, uint64_t seed,
 
 }  // namespace
 
-uint32_t String::ComputeAndSetHash() {
+uint32_t String::ComputeAndSetRawHash() {
   DCHECK(!SharedStringAccessGuardIfNeeded::IsNeeded(*this));
-  return ComputeAndSetHash(SharedStringAccessGuardIfNeeded::NotNeeded());
+  return ComputeAndSetRawHash(SharedStringAccessGuardIfNeeded::NotNeeded());
 }
-uint32_t String::ComputeAndSetHash(
+
+uint32_t String::ComputeAndSetRawHash(
     const SharedStringAccessGuardIfNeeded& access_guard) {
   DisallowGarbageCollection no_gc;
   // Should only be called if hash code has not yet been computed.
-  DCHECK(!HasHashCode());
+  //
+  // If in-place internalizable strings are shared, there may be calls to
+  // ComputeAndSetRawHash in parallel. Since only flat strings are in-place
+  // internalizable and their contents do not change, the result hash is the
+  // same. The raw hash field is stored with relaxed ordering.
+  DCHECK_IMPLIES(!v8_flags.shared_string_table, !HasHashCode());
 
   // Store the hash code in the object.
   uint64_t seed = HashSeed(GetReadOnlyRoots());
@@ -1421,8 +1675,10 @@ uint32_t String::ComputeAndSetHash(
     string = ThinString::cast(string).actual(cage_base);
     shape = StringShape(string, cage_base);
     if (length() == string.length()) {
-      set_raw_hash_field(string.raw_hash_field());
-      return hash();
+      uint32_t raw_hash = string.raw_hash_field();
+      DCHECK(IsHashFieldComputed(raw_hash));
+      set_raw_hash_field(raw_hash);
+      return raw_hash;
     }
   }
   uint32_t raw_hash_field =
@@ -1431,22 +1687,21 @@ uint32_t String::ComputeAndSetHash(
                                 access_guard)
           : HashString<uint16_t>(string, start, length(), seed, cage_base,
                                  access_guard);
-  set_raw_hash_field(raw_hash_field);
-
-  // Check the hash code is there.
-  DCHECK(HasHashCode());
-  uint32_t result = raw_hash_field >> kHashShift;
-  DCHECK_NE(result, 0);  // Ensure that the hash value of 0 is never computed.
-  return result;
+  set_raw_hash_field_if_empty(raw_hash_field);
+  // Check the hash code is there (or a forwarding index if the string was
+  // internalized/externalized in parallel).
+  DCHECK(HasHashCode() || HasForwardingIndex(kAcquireLoad));
+  // Ensure that the hash value of 0 is never computed.
+  DCHECK_NE(HashBits::decode(raw_hash_field), 0);
+  return raw_hash_field;
 }
 
 bool String::SlowAsArrayIndex(uint32_t* index) {
   DisallowGarbageCollection no_gc;
   int length = this->length();
   if (length <= kMaxCachedArrayIndexLength) {
-    EnsureHash();  // Force computation of hash code.
-    uint32_t field = raw_hash_field();
-    if ((field & kIsNotIntegerIndexMask) != 0) return false;
+    uint32_t field = EnsureRawHash();  // Force computation of hash code.
+    if (!IsIntegerIndex(field)) return false;
     *index = ArrayIndexValueBits::decode(field);
     return true;
   }
@@ -1459,9 +1714,8 @@ bool String::SlowAsIntegerIndex(size_t* index) {
   DisallowGarbageCollection no_gc;
   int length = this->length();
   if (length <= kMaxCachedArrayIndexLength) {
-    EnsureHash();  // Force computation of hash code.
-    uint32_t field = raw_hash_field();
-    if ((field & kIsNotIntegerIndexMask) != 0) return false;
+    uint32_t field = EnsureRawHash();  // Force computation of hash code.
+    if (!IsIntegerIndex(field)) return false;
     *index = ArrayIndexValueBits::decode(field);
     return true;
   }
@@ -1501,18 +1755,18 @@ Handle<String> SeqString::Truncate(Handle<SeqString> string, int new_length) {
     new_size = SeqTwoByteString::SizeFor(new_length);
   }
 
-  int delta = old_size - new_size;
-
+#if DEBUG
   Address start_of_string = string->address();
   DCHECK(IsAligned(start_of_string, kObjectAlignment));
   DCHECK(IsAligned(start_of_string + new_size, kObjectAlignment));
+#endif
 
   Heap* heap = Heap::FromWritableHeapObject(*string);
   if (!heap->IsLargeObject(*string)) {
     // Sizes are pointer size aligned, so that we can use filler objects
     // that are a multiple of pointer size.
-    heap->CreateFillerObjectAt(start_of_string + new_size, delta,
-                               ClearRecordedSlots::kNo);
+    heap->NotifyObjectSizeChange(*string, old_size, new_size,
+                                 ClearRecordedSlots::kNo);
   }
   // We are storing the new length using release store after creating a filler
   // for the left-over space to avoid races with the sweeper thread.
@@ -1521,16 +1775,25 @@ Handle<String> SeqString::Truncate(Handle<SeqString> string, int new_length) {
   return string;
 }
 
-void SeqOneByteString::clear_padding() {
-  int data_size = SeqString::kHeaderSize + length() * kOneByteSize;
-  memset(reinterpret_cast<void*>(address() + data_size), 0,
-         SizeFor(length()) - data_size);
+SeqString::DataAndPaddingSizes SeqString::GetDataAndPaddingSizes() const {
+  if (IsSeqOneByteString()) {
+    return SeqOneByteString::cast(*this).GetDataAndPaddingSizes();
+  }
+  return SeqTwoByteString::cast(*this).GetDataAndPaddingSizes();
 }
 
-void SeqTwoByteString::clear_padding() {
+SeqString::DataAndPaddingSizes SeqOneByteString::GetDataAndPaddingSizes()
+    const {
+  int data_size = SeqString::kHeaderSize + length() * kOneByteSize;
+  int padding_size = SizeFor(length()) - data_size;
+  return DataAndPaddingSizes{data_size, padding_size};
+}
+
+SeqString::DataAndPaddingSizes SeqTwoByteString::GetDataAndPaddingSizes()
+    const {
   int data_size = SeqString::kHeaderSize + length() * base::kUC16Size;
-  memset(reinterpret_cast<void*>(address() + data_size), 0,
-         SizeFor(length()) - data_size);
+  int padding_size = SizeFor(length()) - data_size;
+  return DataAndPaddingSizes{data_size, padding_size};
 }
 
 uint16_t ConsString::Get(
@@ -1756,7 +2019,7 @@ const byte* String::AddressOfCharacterAt(
   }
   CHECK_LE(0, start_index);
   CHECK_LE(start_index, subject.length());
-  switch (shape.full_representation_tag()) {
+  switch (shape.representation_and_encoding_tag()) {
     case kOneByteStringTag | kSeqStringTag:
       return reinterpret_cast<const byte*>(
           SeqOneByteString::cast(subject).GetChars(no_gc) + start_index);
@@ -1793,18 +2056,18 @@ namespace {
 
 DEFINE_TORQUE_GENERATED_STRING_INSTANCE_TYPE()
 
-STATIC_ASSERT(kStringRepresentationMask == RepresentationBits::kMask);
+static_assert(kStringRepresentationMask == RepresentationBits::kMask);
 
-STATIC_ASSERT(kStringEncodingMask == IsOneByteBit::kMask);
-STATIC_ASSERT(kTwoByteStringTag == IsOneByteBit::encode(false));
-STATIC_ASSERT(kOneByteStringTag == IsOneByteBit::encode(true));
+static_assert(kStringEncodingMask == IsOneByteBit::kMask);
+static_assert(kTwoByteStringTag == IsOneByteBit::encode(false));
+static_assert(kOneByteStringTag == IsOneByteBit::encode(true));
 
-STATIC_ASSERT(kUncachedExternalStringMask == IsUncachedBit::kMask);
-STATIC_ASSERT(kUncachedExternalStringTag == IsUncachedBit::encode(true));
+static_assert(kUncachedExternalStringMask == IsUncachedBit::kMask);
+static_assert(kUncachedExternalStringTag == IsUncachedBit::encode(true));
 
-STATIC_ASSERT(kIsNotInternalizedMask == IsNotInternalizedBit::kMask);
-STATIC_ASSERT(kNotInternalizedTag == IsNotInternalizedBit::encode(true));
-STATIC_ASSERT(kInternalizedTag == IsNotInternalizedBit::encode(false));
+static_assert(kIsNotInternalizedMask == IsNotInternalizedBit::kMask);
+static_assert(kNotInternalizedTag == IsNotInternalizedBit::encode(true));
+static_assert(kInternalizedTag == IsNotInternalizedBit::encode(false));
 }  // namespace
 
 }  // namespace internal
