@@ -2,13 +2,9 @@
 // Use of this source code is governed by a BSD-style license that can be
 // found in the LICENSE file.
 
-#include "src/execution/arguments-inl.h"
 #include "src/heap/factory.h"
 #include "src/heap/heap-inl.h"
-#include "src/logging/counters.h"
 #include "src/objects/js-generator-inl.h"
-#include "src/objects/objects-inl.h"
-#include "src/runtime/runtime-utils.h"
 
 namespace v8 {
 namespace internal {
@@ -46,8 +42,8 @@ RUNTIME_FUNCTION(Runtime_AsyncFunctionResolve) {
 RUNTIME_FUNCTION(Runtime_CreateJSGeneratorObject) {
   HandleScope scope(isolate);
   DCHECK_EQ(2, args.length());
-  CONVERT_ARG_HANDLE_CHECKED(JSFunction, function, 0);
-  CONVERT_ARG_HANDLE_CHECKED(Object, receiver, 1);
+  Handle<JSFunction> function = args.at<JSFunction>(0);
+  Handle<Object> receiver = args.at(1);
   CHECK_IMPLIES(IsAsyncFunction(function->shared().kind()),
                 IsAsyncGeneratorFunction(function->shared().kind()));
   CHECK(IsResumableFunction(function->shared().kind()));
@@ -62,16 +58,18 @@ RUNTIME_FUNCTION(Runtime_CreateJSGeneratorObject) {
 
   Handle<JSGeneratorObject> generator =
       isolate->factory()->NewJSGeneratorObject(function);
-  generator->set_function(*function);
-  generator->set_context(isolate->context());
-  generator->set_receiver(*receiver);
-  generator->set_parameters_and_registers(*parameters_and_registers);
-  generator->set_resume_mode(JSGeneratorObject::ResumeMode::kNext);
-  generator->set_continuation(JSGeneratorObject::kGeneratorExecuting);
-  if (generator->IsJSAsyncGeneratorObject()) {
-    Handle<JSAsyncGeneratorObject>::cast(generator)->set_is_awaiting(0);
+  DisallowGarbageCollection no_gc;
+  auto raw_generator = *generator;
+  raw_generator.set_function(*function);
+  raw_generator.set_context(isolate->context());
+  raw_generator.set_receiver(*receiver);
+  raw_generator.set_parameters_and_registers(*parameters_and_registers);
+  raw_generator.set_resume_mode(JSGeneratorObject::ResumeMode::kNext);
+  raw_generator.set_continuation(JSGeneratorObject::kGeneratorExecuting);
+  if (raw_generator.IsJSAsyncGeneratorObject()) {
+    JSAsyncGeneratorObject::cast(raw_generator).set_is_awaiting(0);
   }
-  return *generator;
+  return raw_generator;
 }
 
 RUNTIME_FUNCTION(Runtime_GeneratorClose) {
@@ -83,7 +81,7 @@ RUNTIME_FUNCTION(Runtime_GeneratorClose) {
 RUNTIME_FUNCTION(Runtime_GeneratorGetFunction) {
   HandleScope scope(isolate);
   DCHECK_EQ(1, args.length());
-  CONVERT_ARG_HANDLE_CHECKED(JSGeneratorObject, generator, 0);
+  Handle<JSGeneratorObject> generator = args.at<JSGeneratorObject>(0);
 
   return generator->function();
 }
@@ -129,7 +127,7 @@ RUNTIME_FUNCTION(Runtime_GeneratorGetResumeMode) {
 RUNTIME_FUNCTION(Runtime_AsyncGeneratorHasCatchHandlerForPC) {
   DisallowGarbageCollection no_gc_scope;
   DCHECK_EQ(1, args.length());
-  CONVERT_ARG_CHECKED(JSAsyncGeneratorObject, generator, 0);
+  auto generator = JSAsyncGeneratorObject::cast(args[0]);
 
   int state = generator.continuation();
   DCHECK_NE(state, JSAsyncGeneratorObject::kGeneratorExecuting);
