@@ -4,19 +4,15 @@
 
 #include <memory>
 
-#include "src/ast/scopes.h"
 #include "src/builtins/accessors.h"
 #include "src/common/message-template.h"
 #include "src/deoptimizer/deoptimizer.h"
 #include "src/execution/arguments-inl.h"
 #include "src/execution/frames-inl.h"
 #include "src/execution/isolate-inl.h"
+#include "src/execution/isolate.h"
 #include "src/heap/heap-inl.h"  // For ToBoolean. TODO(jkummerow): Drop.
-#include "src/init/bootstrapper.h"
-#include "src/logging/counters.h"
 #include "src/objects/arguments-inl.h"
-#include "src/objects/heap-object-inl.h"
-#include "src/objects/module-inl.h"
 #include "src/objects/smi.h"
 #include "src/runtime/runtime-utils.h"
 
@@ -53,7 +49,7 @@ Object DeclareGlobal(Isolate* isolate, Handle<JSGlobalObject> global,
   Handle<ScriptContextTable> script_contexts(
       global->native_context().script_context_table(), isolate);
   VariableLookupResult lookup;
-  if (ScriptContextTable::Lookup(isolate, *script_contexts, *name, &lookup) &&
+  if (script_contexts->Lookup(name, &lookup) &&
       IsLexicalVariableMode(lookup.mode)) {
     // ES#sec-globaldeclarationinstantiation 6.a:
     // If envRec.HasLexicalDeclaration(name) is true, throw a SyntaxError
@@ -124,8 +120,8 @@ RUNTIME_FUNCTION(Runtime_DeclareModuleExports) {
   HandleScope scope(isolate);
   DCHECK_EQ(2, args.length());
 
-  CONVERT_ARG_HANDLE_CHECKED(FixedArray, declarations, 0);
-  CONVERT_ARG_HANDLE_CHECKED(JSFunction, closure, 1);
+  Handle<FixedArray> declarations = args.at<FixedArray>(0);
+  Handle<JSFunction> closure = args.at<JSFunction>(1);
 
   Handle<ClosureFeedbackCellArray> closure_feedback_cell_array =
       Handle<ClosureFeedbackCellArray>::null();
@@ -172,8 +168,8 @@ RUNTIME_FUNCTION(Runtime_DeclareGlobals) {
   HandleScope scope(isolate);
   DCHECK_EQ(2, args.length());
 
-  CONVERT_ARG_HANDLE_CHECKED(FixedArray, declarations, 0);
-  CONVERT_ARG_HANDLE_CHECKED(JSFunction, closure, 1);
+  Handle<FixedArray> declarations = args.at<FixedArray>(0);
+  Handle<JSFunction> closure = args.at<JSFunction>(1);
 
   Handle<JSGlobalObject> global(isolate->global_object());
   Handle<Context> context(isolate->context(), isolate);
@@ -295,7 +291,7 @@ Object DeclareEvalHelper(Isolate* isolate, Handle<String> name,
   } else if (context->has_extension()) {
     object = handle(context->extension_object(), isolate);
     DCHECK(object->IsJSContextExtensionObject());
-  } else {
+  } else if (context->scope_info().HasContextExtensionSlot()) {
     // Sloppy varblock and function contexts might not have an extension object
     // yet. Sloppy eval will never have an extension object, as vars are hoisted
     // out, and lets are known statically.
@@ -306,6 +302,10 @@ Object DeclareEvalHelper(Isolate* isolate, Handle<String> name,
         isolate->factory()->NewJSObject(isolate->context_extension_function());
 
     context->set_extension(*object);
+  } else {
+    THROW_NEW_ERROR_RETURN_FAILURE(
+        isolate,
+        NewEvalError(MessageTemplate::kVarNotAllowedInEvalScope, name));
   }
 
   RETURN_FAILURE_ON_EXCEPTION(isolate, JSObject::SetOwnPropertyIgnoreAttributes(
@@ -319,15 +319,15 @@ Object DeclareEvalHelper(Isolate* isolate, Handle<String> name,
 RUNTIME_FUNCTION(Runtime_DeclareEvalFunction) {
   HandleScope scope(isolate);
   DCHECK_EQ(2, args.length());
-  CONVERT_ARG_HANDLE_CHECKED(String, name, 0);
-  CONVERT_ARG_HANDLE_CHECKED(Object, value, 1);
+  Handle<String> name = args.at<String>(0);
+  Handle<Object> value = args.at(1);
   return DeclareEvalHelper(isolate, name, value);
 }
 
 RUNTIME_FUNCTION(Runtime_DeclareEvalVar) {
   HandleScope scope(isolate);
   DCHECK_EQ(1, args.length());
-  CONVERT_ARG_HANDLE_CHECKED(String, name, 0);
+  Handle<String> name = args.at<String>(0);
   return DeclareEvalHelper(isolate, name,
                            isolate->factory()->undefined_value());
 }
@@ -488,7 +488,7 @@ class ParameterArguments {
 RUNTIME_FUNCTION(Runtime_NewSloppyArguments) {
   HandleScope scope(isolate);
   DCHECK_EQ(1, args.length());
-  CONVERT_ARG_HANDLE_CHECKED(JSFunction, callee, 0);
+  Handle<JSFunction> callee = args.at<JSFunction>(0);
   // This generic runtime function can also be used when the caller has been
   // inlined, we use the slow but accurate {GetCallerArguments}.
   int argument_count = 0;
@@ -501,7 +501,7 @@ RUNTIME_FUNCTION(Runtime_NewSloppyArguments) {
 RUNTIME_FUNCTION(Runtime_NewStrictArguments) {
   HandleScope scope(isolate);
   DCHECK_EQ(1, args.length());
-  CONVERT_ARG_HANDLE_CHECKED(JSFunction, callee, 0);
+  Handle<JSFunction> callee = args.at<JSFunction>(0);
   // This generic runtime function can also be used when the caller has been
   // inlined, we use the slow but accurate {GetCallerArguments}.
   int argument_count = 0;
@@ -526,7 +526,7 @@ RUNTIME_FUNCTION(Runtime_NewStrictArguments) {
 RUNTIME_FUNCTION(Runtime_NewRestParameter) {
   HandleScope scope(isolate);
   DCHECK_EQ(1, args.length());
-  CONVERT_ARG_HANDLE_CHECKED(JSFunction, callee, 0)
+  Handle<JSFunction> callee = args.at<JSFunction>(0);
   int start_index =
       callee->shared().internal_formal_parameter_count_without_receiver();
   // This generic runtime function can also be used when the caller has been
@@ -552,8 +552,8 @@ RUNTIME_FUNCTION(Runtime_NewRestParameter) {
 RUNTIME_FUNCTION(Runtime_NewClosure) {
   HandleScope scope(isolate);
   DCHECK_EQ(2, args.length());
-  CONVERT_ARG_HANDLE_CHECKED(SharedFunctionInfo, shared, 0);
-  CONVERT_ARG_HANDLE_CHECKED(FeedbackCell, feedback_cell, 1);
+  Handle<SharedFunctionInfo> shared = args.at<SharedFunctionInfo>(0);
+  Handle<FeedbackCell> feedback_cell = args.at<FeedbackCell>(1);
   Handle<Context> context(isolate->context(), isolate);
   return *Factory::JSFunctionBuilder{isolate, shared, context}
               .set_feedback_cell(feedback_cell)
@@ -564,8 +564,8 @@ RUNTIME_FUNCTION(Runtime_NewClosure) {
 RUNTIME_FUNCTION(Runtime_NewClosure_Tenured) {
   HandleScope scope(isolate);
   DCHECK_EQ(2, args.length());
-  CONVERT_ARG_HANDLE_CHECKED(SharedFunctionInfo, shared, 0);
-  CONVERT_ARG_HANDLE_CHECKED(FeedbackCell, feedback_cell, 1);
+  Handle<SharedFunctionInfo> shared = args.at<SharedFunctionInfo>(0);
+  Handle<FeedbackCell> feedback_cell = args.at<FeedbackCell>(1);
   Handle<Context> context(isolate->context(), isolate);
   // The caller ensures that we pretenure closures that are assigned
   // directly to properties.
@@ -579,7 +579,7 @@ RUNTIME_FUNCTION(Runtime_NewFunctionContext) {
   HandleScope scope(isolate);
   DCHECK_EQ(1, args.length());
 
-  CONVERT_ARG_HANDLE_CHECKED(ScopeInfo, scope_info, 0);
+  Handle<ScopeInfo> scope_info = args.at<ScopeInfo>(0);
 
   Handle<Context> outer(isolate->context(), isolate);
   return *isolate->factory()->NewFunctionContext(outer, scope_info);
@@ -589,8 +589,8 @@ RUNTIME_FUNCTION(Runtime_NewFunctionContext) {
 RUNTIME_FUNCTION(Runtime_PushWithContext) {
   HandleScope scope(isolate);
   DCHECK_EQ(2, args.length());
-  CONVERT_ARG_HANDLE_CHECKED(JSReceiver, extension_object, 0);
-  CONVERT_ARG_HANDLE_CHECKED(ScopeInfo, scope_info, 1);
+  Handle<JSReceiver> extension_object = args.at<JSReceiver>(0);
+  Handle<ScopeInfo> scope_info = args.at<ScopeInfo>(1);
   Handle<Context> current(isolate->context(), isolate);
   return *isolate->factory()->NewWithContext(current, scope_info,
                                              extension_object);
@@ -600,8 +600,8 @@ RUNTIME_FUNCTION(Runtime_PushWithContext) {
 RUNTIME_FUNCTION(Runtime_PushCatchContext) {
   HandleScope scope(isolate);
   DCHECK_EQ(2, args.length());
-  CONVERT_ARG_HANDLE_CHECKED(Object, thrown_object, 0);
-  CONVERT_ARG_HANDLE_CHECKED(ScopeInfo, scope_info, 1);
+  Handle<Object> thrown_object = args.at(0);
+  Handle<ScopeInfo> scope_info = args.at<ScopeInfo>(1);
   Handle<Context> current(isolate->context(), isolate);
   return *isolate->factory()->NewCatchContext(current, scope_info,
                                               thrown_object);
@@ -611,7 +611,7 @@ RUNTIME_FUNCTION(Runtime_PushCatchContext) {
 RUNTIME_FUNCTION(Runtime_PushBlockContext) {
   HandleScope scope(isolate);
   DCHECK_EQ(1, args.length());
-  CONVERT_ARG_HANDLE_CHECKED(ScopeInfo, scope_info, 0);
+  Handle<ScopeInfo> scope_info = args.at<ScopeInfo>(0);
   Handle<Context> current(isolate->context(), isolate);
   return *isolate->factory()->NewBlockContext(current, scope_info);
 }
@@ -620,7 +620,7 @@ RUNTIME_FUNCTION(Runtime_PushBlockContext) {
 RUNTIME_FUNCTION(Runtime_DeleteLookupSlot) {
   HandleScope scope(isolate);
   DCHECK_EQ(1, args.length());
-  CONVERT_ARG_HANDLE_CHECKED(String, name, 0);
+  Handle<String> name = args.at<String>(0);
 
   int index;
   PropertyAttributes attributes;
@@ -726,7 +726,7 @@ MaybeHandle<Object> LoadLookupSlot(Isolate* isolate, Handle<String> name,
 RUNTIME_FUNCTION(Runtime_LoadLookupSlot) {
   HandleScope scope(isolate);
   DCHECK_EQ(1, args.length());
-  CONVERT_ARG_HANDLE_CHECKED(String, name, 0);
+  Handle<String> name = args.at<String>(0);
   RETURN_RESULT_OR_FAILURE(isolate,
                            LoadLookupSlot(isolate, name, kThrowOnError));
 }
@@ -735,7 +735,7 @@ RUNTIME_FUNCTION(Runtime_LoadLookupSlot) {
 RUNTIME_FUNCTION(Runtime_LoadLookupSlotInsideTypeof) {
   HandleScope scope(isolate);
   DCHECK_EQ(1, args.length());
-  CONVERT_ARG_HANDLE_CHECKED(String, name, 0);
+  Handle<String> name = args.at<String>(0);
   RETURN_RESULT_OR_FAILURE(isolate, LoadLookupSlot(isolate, name, kDontThrow));
 }
 
@@ -743,7 +743,6 @@ RUNTIME_FUNCTION(Runtime_LoadLookupSlotInsideTypeof) {
 RUNTIME_FUNCTION_RETURN_PAIR(Runtime_LoadLookupSlotForCall) {
   HandleScope scope(isolate);
   DCHECK_EQ(1, args.length());
-  DCHECK(args[0].IsString());
   Handle<String> name = args.at<String>(0);
   Handle<Object> value;
   Handle<Object> receiver;
@@ -826,8 +825,8 @@ MaybeHandle<Object> StoreLookupSlot(
 RUNTIME_FUNCTION(Runtime_StoreLookupSlot_Sloppy) {
   HandleScope scope(isolate);
   DCHECK_EQ(2, args.length());
-  CONVERT_ARG_HANDLE_CHECKED(String, name, 0);
-  CONVERT_ARG_HANDLE_CHECKED(Object, value, 1);
+  Handle<String> name = args.at<String>(0);
+  Handle<Object> value = args.at(1);
   Handle<Context> context(isolate->context(), isolate);
   RETURN_RESULT_OR_FAILURE(
       isolate,
@@ -837,8 +836,8 @@ RUNTIME_FUNCTION(Runtime_StoreLookupSlot_Sloppy) {
 RUNTIME_FUNCTION(Runtime_StoreLookupSlot_Strict) {
   HandleScope scope(isolate);
   DCHECK_EQ(2, args.length());
-  CONVERT_ARG_HANDLE_CHECKED(String, name, 0);
-  CONVERT_ARG_HANDLE_CHECKED(Object, value, 1);
+  Handle<String> name = args.at<String>(0);
+  Handle<Object> value = args.at(1);
   Handle<Context> context(isolate->context(), isolate);
   RETURN_RESULT_OR_FAILURE(
       isolate,
@@ -850,8 +849,8 @@ RUNTIME_FUNCTION(Runtime_StoreLookupSlot_Strict) {
 RUNTIME_FUNCTION(Runtime_StoreLookupSlot_SloppyHoisting) {
   HandleScope scope(isolate);
   DCHECK_EQ(2, args.length());
-  CONVERT_ARG_HANDLE_CHECKED(String, name, 0);
-  CONVERT_ARG_HANDLE_CHECKED(Object, value, 1);
+  Handle<String> name = args.at<String>(0);
+  Handle<Object> value = args.at(1);
   const ContextLookupFlags lookup_flags =
       static_cast<ContextLookupFlags>(DONT_FOLLOW_CHAINS);
   Handle<Context> declaration_context(isolate->context().declaration_context(),
@@ -864,16 +863,15 @@ RUNTIME_FUNCTION(Runtime_StoreLookupSlot_SloppyHoisting) {
 RUNTIME_FUNCTION(Runtime_StoreGlobalNoHoleCheckForReplLetOrConst) {
   HandleScope scope(isolate);
   DCHECK_EQ(2, args.length());
-  CONVERT_ARG_HANDLE_CHECKED(String, name, 0);
-  CONVERT_ARG_HANDLE_CHECKED(Object, value, 1);
+  Handle<String> name = args.at<String>(0);
+  Handle<Object> value = args.at(1);
 
   Handle<Context> native_context = isolate->native_context();
   Handle<ScriptContextTable> script_contexts(
       native_context->script_context_table(), isolate);
 
   VariableLookupResult lookup_result;
-  bool found = ScriptContextTable::Lookup(isolate, *script_contexts, *name,
-                                          &lookup_result);
+  bool found = script_contexts->Lookup(name, &lookup_result);
   CHECK(found);
   Handle<Context> script_context = ScriptContextTable::GetContext(
       isolate, script_contexts, lookup_result.context_index);
