@@ -9,26 +9,28 @@
 #ifndef V8_WASM_WASM_MODULE_H_
 #define V8_WASM_WASM_MODULE_H_
 
+#include <map>
 #include <memory>
 
 #include "src/base/optional.h"
-#include "src/base/platform/wrappers.h"
+#include "src/base/platform/mutex.h"
 #include "src/base/vector.h"
+#include "src/codegen/signature.h"
 #include "src/common/globals.h"
 #include "src/handles/handles.h"
 #include "src/wasm/branch-hint-map.h"
+#include "src/wasm/constant-expression.h"
 #include "src/wasm/signature-map.h"
 #include "src/wasm/struct-types.h"
 #include "src/wasm/wasm-constants.h"
 #include "src/wasm/wasm-init-expr.h"
+#include "src/wasm/wasm-limits.h"
 
-namespace v8 {
-
-namespace internal {
-
+namespace v8::internal {
 class WasmModuleObject;
+}
 
-namespace wasm {
+namespace v8::internal::wasm {
 
 using WasmName = base::Vector<const char>;
 
@@ -38,8 +40,8 @@ class ErrorThrower;
 // Reference to a string in the wire bytes.
 class WireBytesRef {
  public:
-  WireBytesRef() : WireBytesRef(0, 0) {}
-  WireBytesRef(uint32_t offset, uint32_t length)
+  constexpr WireBytesRef() = default;
+  constexpr WireBytesRef(uint32_t offset, uint32_t length)
       : offset_(offset), length_(length) {
     DCHECK_IMPLIES(offset_ == 0, length_ == 0);
     DCHECK_LE(offset_, offset_ + length_);  // no uint32_t overflow.
@@ -52,8 +54,8 @@ class WireBytesRef {
   bool is_set() const { return offset_ != 0; }
 
  private:
-  uint32_t offset_;
-  uint32_t length_;
+  uint32_t offset_ = 0;
+  uint32_t length_ = 0;
 };
 
 // Static representation of a wasm function.
@@ -62,6 +64,10 @@ struct WasmFunction {
   uint32_t func_index;     // index into the function table.
   uint32_t sig_index;      // index into the signature table.
   WireBytesRef code;       // code of this function.
+  // Required number of slots in a feedback vector. Marked {mutable} because
+  // this is computed late (by Liftoff compilation), when the rest of the
+  // {WasmFunction} is typically considered {const}.
+  mutable int feedback_slots;
   bool imported;
   bool exported;
   bool declared;
@@ -69,12 +75,15 @@ struct WasmFunction {
 
 // Static representation of a wasm global variable.
 struct WasmGlobal {
-  ValueType type;     // type of the global.
-  bool mutability;    // {true} if mutable.
-  WireBytesRef init;  // the initialization expression of the global.
+  ValueType type;           // type of the global.
+  bool mutability;          // {true} if mutable.
+  ConstantExpression init;  // the initialization expression of the global.
   union {
-    uint32_t index;   // index of imported mutable global.
-    uint32_t offset;  // offset into global memory (if not imported & mutable).
+    // Index of imported mutable global.
+    uint32_t index;
+    // Offset into global memory (if not imported & mutable). Expressed in bytes
+    // for value-typed globals, and in tagged words for reference-typed globals.
+    uint32_t offset;
   };
   bool imported;  // true if imported.
   bool exported;  // true if exported.
@@ -92,61 +101,69 @@ struct WasmTag {
   const WasmTagSig* sig;  // type signature of the tag.
 };
 
+// Static representation of a wasm literal stringref.
+struct WasmStringRefLiteral {
+  explicit WasmStringRefLiteral(const WireBytesRef& source) : source(source) {}
+  WireBytesRef source;  // start offset in the module bytes.
+};
+
 // Static representation of a wasm data segment.
 struct WasmDataSegment {
   // Construct an active segment.
-  explicit WasmDataSegment(WireBytesRef dest_addr)
-      : dest_addr(std::move(dest_addr)), active(true) {}
+  explicit WasmDataSegment(ConstantExpression dest_addr)
+      : dest_addr(dest_addr), active(true) {}
 
   // Construct a passive segment, which has no dest_addr.
   WasmDataSegment() : active(false) {}
 
-  WireBytesRef dest_addr;  // destination memory address of the data.
-  WireBytesRef source;     // start offset in the module bytes.
-  bool active = true;      // true if copied automatically during instantiation.
+  ConstantExpression dest_addr;  // destination memory address of the data.
+  WireBytesRef source;           // start offset in the module bytes.
+  bool active = true;  // true if copied automatically during instantiation.
 };
 
 // Static representation of wasm element segment (table initializer).
 struct WasmElemSegment {
+  enum Status {
+    kStatusActive,      // copied automatically during instantiation.
+    kStatusPassive,     // copied explicitly after instantiation.
+    kStatusDeclarative  // purely declarative and never copied.
+  };
+  enum ElementType { kFunctionIndexElements, kExpressionElements };
+
   // Construct an active segment.
-  WasmElemSegment(ValueType type, uint32_t table_index, WireBytesRef offset)
-      : type(type),
+  WasmElemSegment(ValueType type, uint32_t table_index,
+                  ConstantExpression offset, ElementType element_type)
+      : status(kStatusActive),
+        type(type),
         table_index(table_index),
         offset(std::move(offset)),
-        status(kStatusActive) {}
+        element_type(element_type) {}
 
   // Construct a passive or declarative segment, which has no table index or
   // offset.
-  WasmElemSegment(ValueType type, bool declarative)
-      : type(type),
-        table_index(0),
-        status(declarative ? kStatusDeclarative : kStatusPassive) {}
+  WasmElemSegment(ValueType type, Status status, ElementType element_type)
+      : status(status), type(type), table_index(0), element_type(element_type) {
+    DCHECK_NE(status, kStatusActive);
+  }
 
-  // Construct a passive or declarative segment, which has no table index or
-  // offset.
+  // Default constructor. Constucts an invalid segment.
   WasmElemSegment()
-      : type(kWasmBottom), table_index(0), status(kStatusActive) {}
+      : status(kStatusActive),
+        type(kWasmBottom),
+        table_index(0),
+        element_type(kFunctionIndexElements) {}
 
   WasmElemSegment(const WasmElemSegment&) = delete;
   WasmElemSegment(WasmElemSegment&&) V8_NOEXCEPT = default;
   WasmElemSegment& operator=(const WasmElemSegment&) = delete;
   WasmElemSegment& operator=(WasmElemSegment&&) V8_NOEXCEPT = default;
 
+  Status status;
   ValueType type;
   uint32_t table_index;
-  WireBytesRef offset;
-  struct Entry {
-    enum Kind { kGlobalGetEntry, kRefFuncEntry, kRefNullEntry } kind;
-    uint32_t index;
-    Entry(Kind kind, uint32_t index) : kind(kind), index(index) {}
-    Entry() : kind(kRefNullEntry), index(0) {}
-  };
-  std::vector<Entry> entries;
-  enum Status {
-    kStatusActive,      // copied automatically during instantiation.
-    kStatusPassive,     // copied explicitly after instantiation.
-    kStatusDeclarative  // purely declarative and never copied.
-  } status;
+  ConstantExpression offset;
+  ElementType element_type;
+  std::vector<ConstantExpression> entries;
 };
 
 // Static representation of a wasm import.
@@ -194,22 +211,98 @@ enum ModuleOrigin : uint8_t {
   ((origin) == kWasmOrigin ? (counters)->prefix##_wasm_##suffix() \
                            : (counters)->prefix##_asm_##suffix())
 
+// Uses a map as backing storage when sparsely, or a vector when densely
+// populated. Requires {Value} to implement `bool is_set()` to identify
+// uninitialized objects.
+template <class Value>
+class AdaptiveMap {
+ public:
+  // The technical limitation here is that index+1 must not overflow. Since
+  // we have significantly lower maximums on anything that can be named,
+  // we can have a tighter limit here to reject useless entries early.
+  static constexpr uint32_t kMaxKey = 10'000'000;
+  static_assert(kMaxKey < std::numeric_limits<uint32_t>::max());
+
+  AdaptiveMap() : map_(new MapType()) {}
+
+  explicit AdaptiveMap(const AdaptiveMap&) = delete;
+  AdaptiveMap& operator=(const AdaptiveMap&) = delete;
+
+  AdaptiveMap(AdaptiveMap&& other) V8_NOEXCEPT { *this = std::move(other); }
+
+  AdaptiveMap& operator=(AdaptiveMap&& other) V8_NOEXCEPT {
+    mode_ = other.mode_;
+    vector_.swap(other.vector_);
+    map_.swap(other.map_);
+    return *this;
+  }
+
+  void FinishInitialization();
+
+  bool is_set() const { return mode_ != kInitializing; }
+
+  void Put(uint32_t key, const Value& value) {
+    DCHECK(mode_ == kInitializing);
+    DCHECK_LE(key, kMaxKey);
+    map_->insert(std::make_pair(key, value));
+  }
+
+  void Put(uint32_t key, Value&& value) {
+    DCHECK(mode_ == kInitializing);
+    DCHECK_LE(key, kMaxKey);
+    map_->insert(std::make_pair(key, std::move(value)));
+  }
+
+  const Value* Get(uint32_t key) const {
+    if (mode_ == kDense) {
+      if (key >= vector_.size()) return nullptr;
+      if (!vector_[key].is_set()) return nullptr;
+      return &vector_[key];
+    } else {
+      DCHECK(mode_ == kSparse || mode_ == kInitializing);
+      auto it = map_->find(key);
+      if (it == map_->end()) return nullptr;
+      return &it->second;
+    }
+  }
+
+  bool Has(uint32_t key) const {
+    if (mode_ == kDense) {
+      return key < vector_.size() && vector_[key].is_set();
+    } else {
+      DCHECK(mode_ == kSparse || mode_ == kInitializing);
+      return map_->find(key) != map_->end();
+    }
+  }
+
+ private:
+  static constexpr uint32_t kLoadFactor = 4;
+  using MapType = std::map<uint32_t, Value>;
+  enum Mode { kDense, kSparse, kInitializing };
+
+  Mode mode_{kInitializing};
+  std::vector<Value> vector_;
+  std::unique_ptr<MapType> map_;
+};
+using NameMap = AdaptiveMap<WireBytesRef>;
+using IndirectNameMap = AdaptiveMap<AdaptiveMap<WireBytesRef>>;
+
 struct ModuleWireBytes;
 
 class V8_EXPORT_PRIVATE LazilyGeneratedNames {
  public:
   WireBytesRef LookupFunctionName(const ModuleWireBytes& wire_bytes,
-                                  uint32_t function_index) const;
+                                  uint32_t function_index);
 
   void AddForTesting(int function_index, WireBytesRef name);
+  bool Has(uint32_t function_index);
 
  private:
-  // {function_names_} are populated lazily after decoding, and
-  // therefore need a mutex to protect concurrent modifications
-  // from multiple {WasmModuleObject}.
-  mutable base::Mutex mutex_;
-  mutable std::unique_ptr<std::unordered_map<uint32_t, WireBytesRef>>
-      function_names_;
+  // Lazy loading must guard against concurrent modifications from multiple
+  // {WasmModuleObject}s.
+  base::Mutex mutex_;
+  bool has_functions_{false};
+  NameMap function_names_;
 };
 
 class V8_EXPORT_PRIVATE AsmJsOffsetInformation {
@@ -240,15 +333,47 @@ class V8_EXPORT_PRIVATE AsmJsOffsetInformation {
   std::unique_ptr<AsmJsOffsets> decoded_offsets_;
 };
 
+// Used as the supertype for a type at the top of the type hierarchy.
+constexpr uint32_t kNoSuperType = std::numeric_limits<uint32_t>::max();
+
 struct TypeDefinition {
-  explicit TypeDefinition(const FunctionSig* sig) : function_sig(sig) {}
-  explicit TypeDefinition(const StructType* type) : struct_type(type) {}
-  explicit TypeDefinition(const ArrayType* type) : array_type(type) {}
+  enum Kind { kFunction, kStruct, kArray };
+
+  TypeDefinition(const FunctionSig* sig, uint32_t supertype)
+      : function_sig(sig), supertype(supertype), kind(kFunction) {}
+  TypeDefinition(const StructType* type, uint32_t supertype)
+      : struct_type(type), supertype(supertype), kind(kStruct) {}
+  TypeDefinition(const ArrayType* type, uint32_t supertype)
+      : array_type(type), supertype(supertype), kind(kArray) {}
+  TypeDefinition()
+      : function_sig(nullptr), supertype(kNoSuperType), kind(kFunction) {}
+
   union {
     const FunctionSig* function_sig;
     const StructType* struct_type;
     const ArrayType* array_type;
   };
+
+  bool operator==(const TypeDefinition& other) const {
+    if (supertype != other.supertype || kind != other.kind) {
+      return false;
+    }
+    switch (kind) {
+      case kFunction:
+        return *function_sig == *other.function_sig;
+      case kStruct:
+        return *struct_type == *other.struct_type;
+      case kArray:
+        return *array_type == *other.array_type;
+    }
+  }
+
+  bool operator!=(const TypeDefinition& other) const {
+    return !(*this == other);
+  }
+
+  uint32_t supertype;
+  Kind kind;
 };
 
 struct V8_EXPORT_PRIVATE WasmDebugSymbols {
@@ -257,12 +382,106 @@ struct V8_EXPORT_PRIVATE WasmDebugSymbols {
   WireBytesRef external_url;
 };
 
-struct WasmTable;
+class CallSiteFeedback {
+ public:
+  struct PolymorphicCase {
+    int function_index;
+    int absolute_call_frequency;
+  };
 
-// End of a chain of explicit supertypes.
-constexpr uint32_t kGenericSuperType = 0xFFFFFFFE;
-// Used for types that have no explicit supertype.
-constexpr uint32_t kNoSuperType = 0xFFFFFFFF;
+  // Regular constructor: uninitialized/unknown, monomorphic, or polymorphic.
+  CallSiteFeedback() : index_or_count_(-1), frequency_or_ool_(0) {}
+  CallSiteFeedback(int function_index, int call_count)
+      : index_or_count_(function_index), frequency_or_ool_(call_count) {}
+  CallSiteFeedback(PolymorphicCase* polymorphic_cases, int num_cases)
+      : index_or_count_(-num_cases),
+        frequency_or_ool_(reinterpret_cast<intptr_t>(polymorphic_cases)) {}
+
+  // Copying and assignment: prefer moving, as it's cheaper.
+  // The code below makes sure external polymorphic storage is copied and/or
+  // freed as appropriate.
+  CallSiteFeedback(const CallSiteFeedback& other) V8_NOEXCEPT { *this = other; }
+  CallSiteFeedback(CallSiteFeedback&& other) V8_NOEXCEPT { *this = other; }
+  CallSiteFeedback& operator=(const CallSiteFeedback& other) V8_NOEXCEPT {
+    index_or_count_ = other.index_or_count_;
+    if (other.is_polymorphic()) {
+      int num_cases = other.num_cases();
+      PolymorphicCase* polymorphic = new PolymorphicCase[num_cases];
+      for (int i = 0; i < num_cases; i++) {
+        polymorphic[i].function_index = other.function_index(i);
+        polymorphic[i].absolute_call_frequency = other.call_count(i);
+      }
+      frequency_or_ool_ = reinterpret_cast<intptr_t>(polymorphic);
+    } else {
+      frequency_or_ool_ = other.frequency_or_ool_;
+    }
+    return *this;
+  }
+  CallSiteFeedback& operator=(CallSiteFeedback&& other) V8_NOEXCEPT {
+    if (this != &other) {
+      index_or_count_ = other.index_or_count_;
+      frequency_or_ool_ = other.frequency_or_ool_;
+      other.frequency_or_ool_ = 0;
+    }
+    return *this;
+  }
+
+  ~CallSiteFeedback() {
+    if (is_polymorphic()) delete[] polymorphic_storage();
+  }
+
+  int num_cases() const {
+    if (is_monomorphic()) return 1;
+    if (is_invalid()) return 0;
+    return -index_or_count_;
+  }
+  int function_index(int i) const {
+    DCHECK(!is_invalid());
+    if (is_monomorphic()) return index_or_count_;
+    return polymorphic_storage()[i].function_index;
+  }
+  int call_count(int i) const {
+    if (index_or_count_ >= 0) return static_cast<int>(frequency_or_ool_);
+    return polymorphic_storage()[i].absolute_call_frequency;
+  }
+
+ private:
+  bool is_monomorphic() const { return index_or_count_ >= 0; }
+  bool is_polymorphic() const { return index_or_count_ <= -2; }
+  bool is_invalid() const { return index_or_count_ == -1; }
+  const PolymorphicCase* polymorphic_storage() const {
+    DCHECK(is_polymorphic());
+    return reinterpret_cast<PolymorphicCase*>(frequency_or_ool_);
+  }
+
+  int index_or_count_;
+  intptr_t frequency_or_ool_;
+};
+
+struct FunctionTypeFeedback {
+  // {feedback_vector} is computed from {call_targets} and the instance-specific
+  // feedback vector by {TransitiveTypeFeedbackProcessor}.
+  std::vector<CallSiteFeedback> feedback_vector;
+
+  // {call_targets} has one entry per "call" and "call_ref" in the function.
+  // For "call", it holds the index of the called function, for "call_ref" the
+  // value will be {kNonDirectCall}.
+  base::OwnedVector<uint32_t> call_targets;
+
+  // {tierup_priority} is updated and used when triggering tier-up.
+  // TODO(clemensb): This does not belong here; find a better place.
+  int tierup_priority = 0;
+
+  static constexpr uint32_t kNonDirectCall = 0xFFFFFFFF;
+};
+
+struct TypeFeedbackStorage {
+  std::unordered_map<uint32_t, FunctionTypeFeedback> feedback_for_function;
+  // Accesses to {feedback_for_function} are guarded by this mutex.
+  mutable base::Mutex mutex;
+};
+
+struct WasmTable;
 
 // Static representation of a module.
 struct V8_EXPORT_PRIVATE WasmModule {
@@ -276,7 +495,6 @@ struct V8_EXPORT_PRIVATE WasmModule {
   bool mem_export = false;         // true if the memory is exported
   int start_function_index = -1;   // start function, >= 0 if any
 
-  std::vector<WasmGlobal> globals;
   // Size of the buffer required for all globals that are not imported and
   // mutable.
   uint32_t untagged_globals_buffer_size = 0;
@@ -291,25 +509,29 @@ struct V8_EXPORT_PRIVATE WasmModule {
   // ID and length).
   WireBytesRef code = {0, 0};
   WireBytesRef name = {0, 0};
-  std::vector<TypeDefinition> types;  // by type index
-  std::vector<uint8_t> type_kinds;    // by type index
-  std::vector<uint32_t> supertypes;   // by type index
-  // Map from each type index to the index of its corresponding canonical type.
-  // Note: right now, only functions are canonicalized, and arrays and structs
-  // map to themselves.
-  std::vector<uint32_t> canonicalized_type_ids;
+  // Position and size of the name section (payload only, i.e. without section
+  // ID and length).
+  WireBytesRef name_section = {0, 0};
+
+  void add_type(TypeDefinition type) {
+    types.push_back(type);
+    uint32_t canonical_id = type.kind == TypeDefinition::kFunction
+                                ? signature_map.FindOrInsert(*type.function_sig)
+                                : 0;
+    per_module_canonical_type_ids.push_back(canonical_id);
+    // Isorecursive canonical type will be computed later.
+    isorecursive_canonical_type_ids.push_back(kNoSuperType);
+  }
 
   bool has_type(uint32_t index) const { return index < types.size(); }
 
   void add_signature(const FunctionSig* sig, uint32_t supertype) {
-    types.push_back(TypeDefinition(sig));
-    type_kinds.push_back(kWasmFunctionTypeCode);
-    supertypes.push_back(supertype);
-    uint32_t canonical_id = sig ? signature_map.FindOrInsert(*sig) : 0;
-    canonicalized_type_ids.push_back(canonical_id);
+    DCHECK_NOT_NULL(sig);
+    add_type(TypeDefinition(sig, supertype));
   }
   bool has_signature(uint32_t index) const {
-    return index < types.size() && type_kinds[index] == kWasmFunctionTypeCode;
+    return index < types.size() &&
+           types[index].kind == TypeDefinition::kFunction;
   }
   const FunctionSig* signature(uint32_t index) const {
     DCHECK(has_signature(index));
@@ -317,14 +539,11 @@ struct V8_EXPORT_PRIVATE WasmModule {
   }
 
   void add_struct_type(const StructType* type, uint32_t supertype) {
-    types.push_back(TypeDefinition(type));
-    type_kinds.push_back(kWasmStructTypeCode);
-    supertypes.push_back(supertype);
-    // No canonicalization for structs.
-    canonicalized_type_ids.push_back(0);
+    DCHECK_NOT_NULL(type);
+    add_type(TypeDefinition(type, supertype));
   }
   bool has_struct(uint32_t index) const {
-    return index < types.size() && type_kinds[index] == kWasmStructTypeCode;
+    return index < types.size() && types[index].kind == TypeDefinition::kStruct;
   }
   const StructType* struct_type(uint32_t index) const {
     DCHECK(has_struct(index));
@@ -332,14 +551,11 @@ struct V8_EXPORT_PRIVATE WasmModule {
   }
 
   void add_array_type(const ArrayType* type, uint32_t supertype) {
-    types.push_back(TypeDefinition(type));
-    type_kinds.push_back(kWasmArrayTypeCode);
-    supertypes.push_back(supertype);
-    // No canonicalization for arrays.
-    canonicalized_type_ids.push_back(0);
+    DCHECK_NOT_NULL(type);
+    add_type(TypeDefinition(type, supertype));
   }
   bool has_array(uint32_t index) const {
-    return index < types.size() && type_kinds[index] == kWasmArrayTypeCode;
+    return index < types.size() && types[index].kind == TypeDefinition::kArray;
   }
   const ArrayType* array_type(uint32_t index) const {
     DCHECK(has_array(index));
@@ -347,26 +563,39 @@ struct V8_EXPORT_PRIVATE WasmModule {
   }
 
   uint32_t supertype(uint32_t index) const {
-    DCHECK(index < supertypes.size());
-    return supertypes[index];
+    DCHECK(index < types.size());
+    return types[index].supertype;
   }
   bool has_supertype(uint32_t index) const {
     return supertype(index) != kNoSuperType;
   }
 
+  std::vector<TypeDefinition> types;  // by type index
+  // TODO(7748): Unify the following two arrays.
+  // Maps each type index to a canonical index for purposes of call_indirect.
+  std::vector<uint32_t> per_module_canonical_type_ids;
+  // Maps each type index to its global (cross-module) canonical index as per
+  // isorecursive type canonicalization.
+  std::vector<uint32_t> isorecursive_canonical_type_ids;
+  // Canonicalizing map for signature indexes.
+  SignatureMap signature_map;
   std::vector<WasmFunction> functions;
+  std::vector<WasmGlobal> globals;
   std::vector<WasmDataSegment> data_segments;
   std::vector<WasmTable> tables;
   std::vector<WasmImport> import_table;
   std::vector<WasmExport> export_table;
   std::vector<WasmTag> tags;
+  std::vector<WasmStringRefLiteral> stringref_literals;
   std::vector<WasmElemSegment> elem_segments;
   std::vector<WasmCompilationHint> compilation_hints;
   BranchHintInfo branch_hints;
-  SignatureMap signature_map;  // canonicalizing map for signature indexes.
+  // Pairs of module offsets and mark id.
+  std::vector<std::pair<uint32_t, uint32_t>> inst_traces;
+  mutable TypeFeedbackStorage type_feedback;
 
   ModuleOrigin origin = kWasmOrigin;  // origin of the module
-  LazilyGeneratedNames lazily_generated_names;
+  mutable LazilyGeneratedNames lazily_generated_names;
   WasmDebugSymbols debug_symbols;
 
   // Asm.js source position information. Only available for modules compiled
@@ -375,7 +604,6 @@ struct V8_EXPORT_PRIVATE WasmModule {
 
   explicit WasmModule(std::unique_ptr<Zone> signature_zone = nullptr);
   WasmModule(const WasmModule&) = delete;
-  ~WasmModule();
   WasmModule& operator=(const WasmModule&) = delete;
 };
 
@@ -383,24 +611,13 @@ struct V8_EXPORT_PRIVATE WasmModule {
 struct WasmTable {
   MOVE_ONLY_WITH_DEFAULT_CONSTRUCTORS(WasmTable);
 
-  // 'module' can be nullptr
-  // TODO(9495): Update this function as more table types are supported, or
-  // remove it completely when all reference types are allowed.
-  static bool IsValidTableType(ValueType type, const WasmModule* module) {
-    if (!type.is_object_reference()) return false;
-    HeapType heap_type = type.heap_type();
-    return heap_type == HeapType::kFunc || heap_type == HeapType::kExtern ||
-           (module != nullptr && heap_type.is_index() &&
-            module->has_signature(heap_type.ref_index()));
-  }
-
   ValueType type = kWasmVoid;     // table type.
   uint32_t initial_size = 0;      // initial table size.
   uint32_t maximum_size = 0;      // maximum table size.
   bool has_maximum_size = false;  // true if there is a maximum size.
   bool imported = false;          // true if imported.
   bool exported = false;          // true if exported.
-  WireBytesRef initial_value;
+  ConstantExpression initial_value;
 };
 
 inline bool is_asmjs_module(const WasmModule* module) {
@@ -439,7 +656,8 @@ int GetNearestWasmFunction(const WasmModule* module, uint32_t byte_offset);
 // Returns 0 if the type has no explicit supertype.
 // The result is capped to {kV8MaxRttSubtypingDepth + 1}.
 // Invalid cyclic hierarchies will return -1.
-int GetSubtypingDepth(const WasmModule* module, uint32_t type_index);
+V8_EXPORT_PRIVATE int GetSubtypingDepth(const WasmModule* module,
+                                        uint32_t type_index);
 
 // Interface to the storage (wire bytes) of a wasm module.
 // It is illegal for anyone receiving a ModuleWireBytes to store pointers based
@@ -490,10 +708,13 @@ struct WasmFunctionName {
   const WasmName name_;
 };
 
-std::ostream& operator<<(std::ostream& os, const WasmFunctionName& name);
+V8_EXPORT_PRIVATE std::ostream& operator<<(std::ostream& os,
+                                           const WasmFunctionName& name);
 
 V8_EXPORT_PRIVATE bool IsWasmCodegenAllowed(Isolate* isolate,
                                             Handle<Context> context);
+V8_EXPORT_PRIVATE Handle<String> ErrorStringForCodegen(Isolate* isolate,
+                                                       Handle<Context> context);
 
 Handle<JSObject> GetTypeForFunction(Isolate* isolate, const FunctionSig* sig,
                                     bool for_exception = false);
@@ -525,6 +746,9 @@ inline int declared_function_index(const WasmModule* module, int func_index) {
   return declared_idx;
 }
 
+// Translate from function index to jump table offset.
+int JumpTableOffset(const WasmModule* module, int func_index);
+
 // TruncatedUserString makes it easy to output names up to a certain length, and
 // output a truncation followed by '...' if they exceed a limit.
 // Use like this:
@@ -568,8 +792,15 @@ class TruncatedUserString {
 size_t PrintSignature(base::Vector<char> buffer, const wasm::FunctionSig*,
                       char delimiter = ':');
 
-}  // namespace wasm
-}  // namespace internal
-}  // namespace v8
+V8_EXPORT_PRIVATE size_t
+GetWireBytesHash(base::Vector<const uint8_t> wire_bytes);
+
+void DumpProfileToFile(const WasmModule* module,
+                       base::Vector<const uint8_t> wire_bytes);
+
+void LoadProfileFromFile(WasmModule* module,
+                         base::Vector<const uint8_t> wire_bytes);
+
+}  // namespace v8::internal::wasm
 
 #endif  // V8_WASM_WASM_MODULE_H_
