@@ -2,8 +2,6 @@
 # Use of this source code is governed by a BSD-style license that can be
 # found in the LICENSE file.
 
-from collections import namedtuple
-import time
 
 from . import base
 
@@ -11,46 +9,52 @@ from . import base
 # Extra flags randomly added to all fuzz tests with numfuzz. List of tuples
 # (probability, flag).
 EXTRA_FLAGS = [
-  (0.1, '--always-opt'),
-  (0.1, '--assert-types'),
-  # TODO(mythria): https://crbug.com/v8/10243
-  # (0.1, '--budget-for-feedback-vector-allocation=0'),
-  (0.1, '--cache=code'),
-  (0.1, '--force-slow-path'),
-  (0.2, '--future'),
-  (0.1, '--interrupt-budget=100'),
-  (0.1, '--liftoff'),
-  (0.2, '--no-analyze-environment-liveness'),
-  # TODO(machenbach): Enable when it doesn't collide with crashing on missing
-  # simd features.
-  #(0.1, '--no-enable-sse3'),
-  #(0.1, '--no-enable-ssse3'),
-  #(0.1, '--no-enable-sse4_1'),
-  (0.1, '--no-enable-sse4_2'),
-  (0.1, '--no-enable-sahf'),
-  (0.1, '--no-enable-avx'),
-  (0.1, '--no-enable-fma3'),
-  (0.1, '--no-enable-bmi1'),
-  (0.1, '--no-enable-bmi2'),
-  (0.1, '--no-enable-lzcnt'),
-  (0.1, '--no-enable-popcnt'),
-  (0.3, '--no-lazy-feedback-allocation'),
-  (0.1, '--no-liftoff'),
-  (0.1, '--no-opt'),
-  (0.2, '--no-regexp-tier-up'),
-  (0.1, '--no-wasm-tier-up'),
-  (0.1, '--regexp-interpret-all'),
-  (0.1, '--regexp-tier-up-ticks=10'),
-  (0.1, '--regexp-tier-up-ticks=100'),
-  (0.1, '--stress-background-compile'),
-  (0.1, '--stress-concurrent-inlining'),
-  (0.1, '--stress-flush-code'),
-  (0.1, '--stress-lazy-source-positions'),
-  (0.1, '--stress-wasm-code-gc'),
-  (0.1, '--turbo-instruction-scheduling'),
-  (0.1, '--turbo-stress-instruction-scheduling'),
+    (0.1, '--always-turbofan'),
+    (0.1, '--assert-types'),
+    (0.1, '--interrupt-budget-for-feedback-allocation=0'),
+    (0.1, '--cache=code'),
+    (0.25, '--compact-maps'),
+    (0.1, '--force-slow-path'),
+    (0.2, '--future'),
+    (0.1, '--interrupt-budget=100'),
+    # TODO(almuthanna): enable again when the FYI bots are greener
+    # (0.1, '--interrupt-budget-for-maglev=100'),
+    (0.1, '--liftoff'),
+    (0.1, '--maglev'),
+    (0.1, '--minor-mc'),
+    (0.2, '--no-analyze-environment-liveness'),
+    # TODO(machenbach): Enable when it doesn't collide with crashing on missing
+    # simd features.
+    #(0.1, '--no-enable-sse3'),
+    #(0.1, '--no-enable-ssse3'),
+    #(0.1, '--no-enable-sse4_1'),
+    (0.1, '--no-enable-sse4_2'),
+    (0.1, '--no-enable-sahf'),
+    (0.1, '--no-enable-avx'),
+    (0.1, '--no-enable-fma3'),
+    (0.1, '--no-enable-bmi1'),
+    (0.1, '--no-enable-bmi2'),
+    (0.1, '--no-enable-lzcnt'),
+    (0.1, '--no-enable-popcnt'),
+    (0.3, '--no-lazy-feedback-allocation'),
+    (0.1, '--no-liftoff'),
+    (0.1, '--no-turbofan'),
+    (0.2, '--no-regexp-tier-up'),
+    (0.25, '--no-use-map-space'),
+    (0.1, '--no-wasm-tier-up'),
+    (0.1, '--regexp-interpret-all'),
+    (0.1, '--regexp-tier-up-ticks=10'),
+    (0.1, '--regexp-tier-up-ticks=100'),
+    (0.1, '--stress-background-compile'),
+    (0.1, '--stress-flush-code'),
+    (0.1, '--stress-lazy-source-positions'),
+    (0.1, '--stress-wasm-code-gc'),
+    (0.1, '--turbo-instruction-scheduling'),
+    (0.1, '--turbo-stress-instruction-scheduling'),
+    (0.1, '--turbo-force-mid-tier-regalloc'),
 ]
 
+
 def random_extra_flags(rng):
   """Returns a random list of flags chosen from the configurations in
   EXTRA_FLAGS.
@@ -58,6 +62,49 @@ def random_extra_flags(rng):
   return [flag for prob, flag in EXTRA_FLAGS if rng.random() < prob]
 
 
+def _flag_prefix(flag):
+  """Returns the flag part before an equal sign."""
+  if '=' not in flag:
+    return flag
+  else:
+    return flag[0:flag.index('=')]
+
+
+def _invert_flag(flag):
+  """Flips a --flag and its --no-flag counterpart."""
+  assert flag.startswith('--')
+  if flag.startswith('--no-'):
+    return '--' + flag[len('--no-'):]
+  else:
+    return '--no-' + flag[2:]
+
+
+def _drop_contradictory_flags(new_flags, existing_flags):
+  """Drops flags that have a simple contradiction with an existing flag.
+
+  Contradictions checked for:
+  - Repetition: --flag --flag
+  - Repetition with param: --flag=foo --flag=bar
+  - Negation: --flag --no-flag
+  - Inverse negation: --no-flag --flag
+  - For simplicity also drops combinations of negation and param, which don't
+    occur in practice.
+
+  Args:
+    new_flags: new flags to filter from
+    existing_flags: existing flags checked against
+  Returns: A list of flags without contradictions.
+  """
+  existing_flag_prefixes = set(_flag_prefix(flag) for flag in existing_flags)
+
+  def contradictory_flag(flag):
+    flag_prefix = _flag_prefix(flag)
+    return (flag_prefix in existing_flag_prefixes or
+            _invert_flag(flag_prefix) in existing_flag_prefixes)
+
+  return [flag for flag in new_flags if not contradictory_flag(flag)]
+
+
 class FuzzerConfig(object):
   def __init__(self, probability, analyzer, fuzzer):
     """
@@ -95,6 +142,15 @@ class Fuzzer(object):
 
 # TODO(majeski): Allow multiple subtests to run at once.
 class FuzzerProc(base.TestProcProducer):
+  @staticmethod
+  def create(options):
+    return FuzzerProc(
+        options.fuzzer_rng(),
+        options.fuzzer_tests_count(),
+        options.fuzzer_configs(),
+        options.combine_tests,
+    )
+
   def __init__(self, rng, count, fuzzers, disable_analysis=False):
     """
     Args:
@@ -105,18 +161,12 @@ class FuzzerProc(base.TestProcProducer):
         set, processor passes None as analysis result to fuzzers
     """
     super(FuzzerProc, self).__init__('Fuzzer')
-
     self._rng = rng
     self._count = count
     self._fuzzer_configs = fuzzers
     self._disable_analysis = disable_analysis
     self._gens = {}
 
-  def setup(self, requirement=base.DROP_RESULT):
-    # Fuzzer is optimized to not store the results
-    assert requirement == base.DROP_RESULT
-    super(FuzzerProc, self).setup(requirement)
-
   def _next_test(self, test):
     if self.is_stopped:
       return False
@@ -189,15 +239,18 @@ class FuzzerProc(base.TestProcProducer):
           flags += next(gen)
 
       flags.append('--fuzzer-random-seed=%s' % self._next_seed())
+
+      flags = _drop_contradictory_flags(flags, test.get_flags())
       yield self._create_subtest(test, str(i), flags=flags)
 
       i += 1
 
   def _try_send_next_test(self, test):
-    if not self.is_stopped:
-      for subtest in self._gens[test.procid]:
-        if self._send_test(subtest):
-          return True
+    for subtest in self._gens[test.procid]:
+      if self._send_test(subtest):
+        return True
+      elif self.is_stopped:
+        return False
 
     del self._gens[test.procid]
     return False
@@ -266,11 +319,30 @@ class CompactionFuzzer(Fuzzer):
     while True:
       yield ['--stress-compaction-random']
 
+
+class InterruptBudgetFuzzer(Fuzzer):
+  def create_flags_generator(self, rng, test, analysis_value):
+    while True:
+      # Half with, half without lazy feedback allocation. The first flag
+      # overwrites potential flag negations from the extra flags list.
+      flag1 = rng.choice(
+          ['--lazy-feedback-allocation', '--no-lazy-feedback-allocation'])
+      flag2 = '--interrupt-budget=%d' % rng.randint(0, 135168)
+      flag3 = '--interrupt-budget-for-maglev=%d' % rng.randint(0, 40960)
+      flag4 = '--interrupt-budget-for-feedback-allocation=%d' % rng.randint(
+          0, 940)
+      flag5 = '--interrupt-budget-factor-for-feedback-allocation=%d' % rng.randint(
+          1, 8)
+
+      yield [flag1, flag2, flag3, flag4, flag5]
+
+
 class StackSizeFuzzer(Fuzzer):
   def create_flags_generator(self, rng, test, analysis_value):
     while True:
       yield ['--stack-size=%d' % rng.randint(54, 983)]
 
+
 class TaskDelayFuzzer(Fuzzer):
   def create_flags_generator(self, rng, test, analysis_value):
     while True:
@@ -321,14 +393,15 @@ class DeoptFuzzer(Fuzzer):
 
 
 FUZZERS = {
-  'compaction': (None, CompactionFuzzer),
-  'delay': (None, TaskDelayFuzzer),
-  'deopt': (DeoptAnalyzer, DeoptFuzzer),
-  'gc_interval': (GcIntervalAnalyzer, GcIntervalFuzzer),
-  'marking': (MarkingAnalyzer, MarkingFuzzer),
-  'scavenge': (ScavengeAnalyzer, ScavengeFuzzer),
-  'stack': (None, StackSizeFuzzer),
-  'threads': (None, ThreadPoolSizeFuzzer),
+    'compaction': (None, CompactionFuzzer),
+    'delay': (None, TaskDelayFuzzer),
+    'deopt': (DeoptAnalyzer, DeoptFuzzer),
+    'gc_interval': (GcIntervalAnalyzer, GcIntervalFuzzer),
+    'interrupt': (None, InterruptBudgetFuzzer),
+    'marking': (MarkingAnalyzer, MarkingFuzzer),
+    'scavenge': (ScavengeAnalyzer, ScavengeFuzzer),
+    'stack': (None, StackSizeFuzzer),
+    'threads': (None, ThreadPoolSizeFuzzer),
 }
 
 
