@@ -84,11 +84,6 @@ const {
   validateInteger,
 } = require('internal/validators');
 const { getTimerDuration } = require('internal/timers');
-const {
-  DTRACE_HTTP_CLIENT_REQUEST,
-  DTRACE_HTTP_CLIENT_RESPONSE
-} = require('internal/dtrace');
-
 const {
   hasObserver,
   startPerf,
@@ -178,10 +173,8 @@ function ClientRequest(input, options, cb) {
 
   if (options.path) {
     const path = String(options.path);
-    if (RegExpPrototypeExec(INVALID_PATH_REGEX, path) !== null) {
-      debug('Path contains unescaped characters: "%s"', path);
+    if (RegExpPrototypeExec(INVALID_PATH_REGEX, path) !== null)
       throw new ERR_UNESCAPED_CHARACTERS('Request path');
-    }
   }
 
   if (protocol !== expectedProtocol) {
@@ -367,7 +360,6 @@ ObjectSetPrototypeOf(ClientRequest.prototype, OutgoingMessage.prototype);
 ObjectSetPrototypeOf(ClientRequest, OutgoingMessage);
 
 ClientRequest.prototype._finish = function _finish() {
-  DTRACE_HTTP_CLIENT_REQUEST(this, this.socket);
   FunctionPrototypeCall(OutgoingMessage.prototype._finish, this);
   if (hasObserver('http')) {
     startPerf(this, kClientRequestStatistics, {
@@ -657,7 +649,6 @@ function parserOnIncomingClient(res, shouldKeepAlive) {
     req.shouldKeepAlive = false;
   }
 
-  DTRACE_HTTP_CLIENT_RESPONSE(socket, req);
   if (req[kClientRequestStatistics] && hasObserver('http')) {
     stopPerf(req, kClientRequestStatistics, {
       detail: {
