@@ -8,7 +8,6 @@ const {
   ObjectDefineProperties,
   ObjectSetPrototypeOf,
   ObjectDefineProperty,
-  PromiseResolve,
   SafeFinalizationRegistry,
   SafeSet,
   Symbol,
@@ -23,13 +22,11 @@ const {
   kTrustEvent,
   kNewListener,
   kRemoveListener,
-  kWeakHandler,
 } = require('internal/event_target');
 const {
-  createDeferredPromise,
   customInspectSymbol,
-  kEmptyObject,
   kEnumerableProperty,
+  kEmptyObject,
 } = require('internal/util');
 const { inspect } = require('internal/util/inspect');
 const {
@@ -37,12 +34,10 @@ const {
     ERR_ILLEGAL_CONSTRUCTOR,
     ERR_INVALID_ARG_TYPE,
     ERR_INVALID_THIS,
-  },
+  }
 } = require('internal/errors');
 
 const {
-  validateAbortSignal,
-  validateObject,
   validateUint32,
 } = require('internal/validators');
 
@@ -58,7 +53,7 @@ const {
 const {
   messaging_deserialize_symbol: kDeserialize,
   messaging_transfer_symbol: kTransfer,
-  messaging_transfer_list_symbol: kTransferList,
+  messaging_transfer_list_symbol: kTransferList
 } = internalBinding('symbols');
 
 let _MessageChannel;
@@ -93,13 +88,13 @@ function customInspect(self, obj, depth, options) {
     return self;
 
   const opts = ObjectAssign({}, options, {
-    depth: options.depth === null ? null : options.depth - 1,
+    depth: options.depth === null ? null : options.depth - 1
   });
 
   return `${self.constructor.name} ${inspect(obj, opts)}`;
 }
 
-function validateThisAbortSignal(obj) {
+function validateAbortSignal(obj) {
   if (obj?.[kAborted] === undefined)
     throw new ERR_INVALID_THIS('AbortSignal');
 }
@@ -137,7 +132,7 @@ class AbortSignal extends EventTarget {
    * @type {boolean}
    */
   get aborted() {
-    validateThisAbortSignal(this);
+    validateAbortSignal(this);
     return !!this[kAborted];
   }
 
@@ -145,20 +140,19 @@ class AbortSignal extends EventTarget {
    * @type {any}
    */
   get reason() {
-    validateThisAbortSignal(this);
+    validateAbortSignal(this);
     return this[kReason];
   }
 
   throwIfAborted() {
-    validateThisAbortSignal(this);
-    if (this[kAborted]) {
-      throw this[kReason];
+    if (this.aborted) {
+      throw this.reason;
     }
   }
 
   [customInspectSymbol](depth, options) {
     return customInspect(this, {
-      aborted: this.aborted,
+      aborted: this.aborted
     }, depth, options);
   }
 
@@ -176,7 +170,7 @@ class AbortSignal extends EventTarget {
    * @returns {AbortSignal}
    */
   static timeout(delay) {
-    validateUint32(delay, 'delay', false);
+    validateUint32(delay, 'delay', true);
     const signal = createAbortSignal();
     signal[kTimeout] = true;
     clearTimeoutRegistry.register(
@@ -188,10 +182,10 @@ class AbortSignal extends EventTarget {
   [kNewListener](size, type, listener, once, capture, passive, weak) {
     super[kNewListener](size, type, listener, once, capture, passive, weak);
     if (this[kTimeout] &&
-      type === 'abort' &&
-      !this.aborted &&
-      !weak &&
-      size === 1) {
+        type === 'abort' &&
+        !this.aborted &&
+        !weak &&
+        size === 1) {
       // If this is a timeout signal, and we're adding a non-weak abort
       // listener, then we don't want it to be gc'd while the listener
       // is attached and the timer still hasn't fired. So, we retain a
@@ -208,7 +202,7 @@ class AbortSignal extends EventTarget {
   }
 
   [kTransfer]() {
-    validateThisAbortSignal(this);
+    validateAbortSignal(this);
     const aborted = this.aborted;
     if (aborted) {
       const reason = this.reason;
@@ -267,7 +261,7 @@ class AbortSignal extends EventTarget {
 function ClonedAbortSignal() {
   return createAbortSignal({ transferable: true });
 }
-ClonedAbortSignal.prototype[kDeserialize] = () => { };
+ClonedAbortSignal.prototype[kDeserialize] = () => {};
 
 ObjectDefineProperties(AbortSignal.prototype, {
   aborted: kEnumerableProperty,
@@ -309,50 +303,37 @@ function abortSignal(signal, reason) {
   signal[kAborted] = true;
   signal[kReason] = reason;
   const event = new Event('abort', {
-    [kTrustEvent]: true,
+    [kTrustEvent]: true
   });
   signal.dispatchEvent(event);
 }
 
-// TODO(joyeecheung): use private fields and we'll get invalid access
-// validation from V8 instead of throwing ERR_INVALID_THIS ourselves.
-const kSignal = Symbol('signal');
-
-function validateAbortController(obj) {
-  if (obj?.[kSignal] === undefined)
-    throw new ERR_INVALID_THIS('AbortController');
-}
-
 class AbortController {
-  constructor() {
-    this[kSignal] = createAbortSignal();
-  }
+  #signal = createAbortSignal();
 
   /**
    * @type {AbortSignal}
    */
   get signal() {
-    validateAbortController(this);
-    return this[kSignal];
+    return this.#signal;
   }
 
   /**
    * @param {any} reason
    */
   abort(reason = new DOMException('This operation was aborted', 'AbortError')) {
-    validateAbortController(this);
-    abortSignal(this[kSignal], reason);
+    abortSignal(this.#signal, reason);
   }
 
   [customInspectSymbol](depth, options) {
     return customInspect(this, {
-      signal: this.signal,
+      signal: this.signal
     }, depth, options);
   }
 
   static [kMakeTransferable]() {
     const controller = new AbortController();
-    controller[kSignal] = transferableAbortSignal(controller[kSignal]);
+    controller.#signal = transferableAbortSignal(controller.#signal);
     return controller;
   }
 }
@@ -375,24 +356,6 @@ function transferableAbortController() {
   return AbortController[kMakeTransferable]();
 }
 
-/**
- * @param {AbortSignal} signal
- * @param {any} resource
- * @returns {Promise<void>}
- */
-async function aborted(signal, resource) {
-  if (signal === undefined) {
-    throw new ERR_INVALID_ARG_TYPE('signal', 'AbortSignal', signal);
-  }
-  validateAbortSignal(signal, 'signal');
-  validateObject(resource, 'resource', { nullable: false, allowFunction: true, allowArray: true });
-  if (signal.aborted)
-    return PromiseResolve();
-  const abortPromise = createDeferredPromise();
-  signal.addEventListener('abort', abortPromise.resolve, { [kWeakHandler]: resource, once: true });
-  return abortPromise.promise;
-}
-
 ObjectDefineProperties(AbortController.prototype, {
   signal: kEnumerableProperty,
   abort: kEnumerableProperty,
@@ -407,10 +370,10 @@ ObjectDefineProperty(AbortController.prototype, SymbolToStringTag, {
 });
 
 module.exports = {
+  kAborted,
   AbortController,
   AbortSignal,
   ClonedAbortSignal,
-  aborted,
   transferableAbortSignal,
   transferableAbortController,
 };
