@@ -15,6 +15,8 @@ const EventEmitter = require('events');
 const { owner_symbol } = require('internal/async_hooks').symbols;
 const Worker = require('internal/cluster/worker');
 const { internal, sendHelper } = require('internal/cluster/utils');
+const { exitCodes: { kNoFailure } } = internalBinding('errors');
+
 const cluster = new EventEmitter();
 const handles = new SafeMap();
 const indexes = new SafeMap();
@@ -23,7 +25,8 @@ const noop = FunctionPrototype;
 module.exports = cluster;
 
 cluster.isWorker = true;
-cluster.isMaster = false;
+cluster.isMaster = false; // Deprecated alias. Must be same as isPrimary.
+cluster.isPrimary = false;
 cluster.worker = null;
 cluster.Worker = Worker;
 
@@ -40,9 +43,9 @@ cluster._setupWorker = function() {
     worker.emit('disconnect');
 
     if (!worker.exitedAfterDisconnect) {
-      // Unexpected disconnect, master exited, or some such nastiness, so
+      // Unexpected disconnect, primary exited, or some such nastiness, so
       // worker exits immediately.
-      process.exit(0);
+      process.exit(kNoFailure);
     }
   });
 
@@ -110,6 +113,10 @@ cluster._getServer = function(obj, options, cb) {
   });
 
   obj.once('listening', () => {
+    // short-lived sockets might have been closed
+    if (!indexes.has(indexesKey)) {
+      return;
+    }
     cluster.worker.state = 'listening';
     const address = obj.address();
     message.act = 'listening';
@@ -156,7 +163,7 @@ function rr(message, { indexesKey, index }, cb) {
   let key = message.key;
 
   function listen(backlog) {
-    // TODO(bnoordhuis) Send a message to the master that tells it to
+    // TODO(bnoordhuis) Send a message to the primary that tells it to
     // update the backlog size. The actual backlog should probably be
     // the largest requested size by any worker.
     return 0;
@@ -165,9 +172,9 @@ function rr(message, { indexesKey, index }, cb) {
   function close() {
     // lib/net.js treats server._handle.close() as effectively synchronous.
     // That means there is a time window between the call to close() and
-    // the ack by the master process in which we can still receive handles.
+    // the ack by the primary process in which we can still receive handles.
     // onconnection() below handles that by sending those handles back to
-    // the master.
+    // the primary.
     if (key === undefined)
       return;
 
@@ -203,19 +210,28 @@ function rr(message, { indexesKey, index }, cb) {
 function onconnection(message, handle) {
   const key = message.key;
   const server = handles.get(key);
-  const accepted = server !== undefined;
+  let accepted = server !== undefined;
+
+  if (accepted && server[owner_symbol]) {
+    const self = server[owner_symbol];
+    if (self.maxConnections && self._connections >= self.maxConnections) {
+      accepted = false;
+    }
+  }
 
   send({ ack: message.seq, accepted });
 
   if (accepted)
     server.onconnection(0, handle);
+  else
+    handle.close();
 }
 
 function send(message, cb) {
   return sendHelper(process, message, null, cb);
 }
 
-function _disconnect(masterInitiated) {
+function _disconnect(primaryInitiated) {
   this.exitedAfterDisconnect = true;
   let waitingCount = 1;
 
@@ -224,10 +240,10 @@ function _disconnect(masterInitiated) {
 
     if (waitingCount === 0) {
       // If disconnect is worker initiated, wait for ack to be sure
-      // exitedAfterDisconnect is properly set in the master, otherwise, if
-      // it's master initiated there's no need to send the
+      // exitedAfterDisconnect is properly set in the primary, otherwise, if
+      // it's primary initiated there's no need to send the
       // exitedAfterDisconnect message
-      if (masterInitiated) {
+      if (primaryInitiated) {
         process.disconnect();
       } else {
         send({ act: 'exitedAfterDisconnect' }, () => process.disconnect());
@@ -264,10 +280,10 @@ Worker.prototype.destroy = function() {
 
   this.exitedAfterDisconnect = true;
   if (!this.isConnected()) {
-    process.exit(0);
+    process.exit(kNoFailure);
   } else {
     this.state = 'destroying';
     send({ act: 'exitedAfterDisconnect' }, () => process.disconnect());
-    process.once('disconnect', () => process.exit(0));
+    process.once('disconnect', () => process.exit(kNoFailure));
   }
 };
