@@ -143,10 +143,6 @@ const kNamedCurveAliases = {
   'P-256': 'prime256v1',
   'P-384': 'secp384r1',
   'P-521': 'secp521r1',
-  'NODE-ED25519': 'ed25519',
-  'NODE-ED448': 'ed448',
-  'NODE-X25519': 'x25519',
-  'NODE-X448': 'x448',
 };
 
 const kAesKeyLengths = [128, 192, 256];
@@ -170,13 +166,10 @@ const kAlgorithms = {
   'sha-512': 'SHA-512',
   'hkdf': 'HKDF',
   'pbkdf2': 'PBKDF2',
-  // Following here are Node.js specific extensions. All
-  // should be prefixed with 'node-'
-  'node-dsa': 'NODE-DSA',
-  'node-dh': 'NODE-DH',
-  'node-scrypt': 'NODE-SCRYPT',
-  'node-ed25519': 'NODE-ED25519',
-  'node-ed448': 'NODE-ED448',
+  'ed25519': 'Ed25519',
+  'ed448': 'Ed448',
+  'x25519': 'X25519',
+  'x448': 'X448',
 };
 const kAlgorithmsKeys = ObjectKeys(kAlgorithms);
 
@@ -186,8 +179,7 @@ const kExportFormats = [
   'raw',
   'pkcs8',
   'spki',
-  'jwk',
-  'node.keyObject'];
+  'jwk'];
 
 // These are the only hash algorithms we currently support via
 // the Web Crypto API.
@@ -286,7 +278,11 @@ const validateByteSource = hideStackFrames((val, name) => {
 });
 
 function onDone(resolve, reject, err, result) {
-  if (err) return reject(err);
+  if (err) {
+    return reject(lazyDOMException(
+      'The operation failed for an operation-specific reason',
+      { name: 'OperationError', cause: err }));
+  }
   resolve(result);
 }
 
@@ -344,12 +340,12 @@ function getUsagesUnion(usageSet, ...usages) {
   return newset;
 }
 
-function getHashLength(name) {
+function getBlockSize(name) {
   switch (name) {
-    case 'SHA-1': return 160;
-    case 'SHA-256': return 256;
-    case 'SHA-384': return 384;
-    case 'SHA-512': return 512;
+    case 'SHA-1': return 512;
+    case 'SHA-256': return 512;
+    case 'SHA-384': return 1024;
+    case 'SHA-512': return 1024;
   }
 }
 
@@ -434,8 +430,8 @@ module.exports = {
   validateMaxBufferLength,
   bigIntArrayToUnsignedBigInt,
   bigIntArrayToUnsignedInt,
+  getBlockSize,
   getStringOption,
   getUsagesUnion,
-  getHashLength,
   secureHeapUsed,
 };
