@@ -6,14 +6,16 @@ const {
   ArrayPrototypeMap,
   ArrayPrototypePushApply,
   ArrayPrototypeSlice,
-  StringPrototypeStartsWith,
 } = primordials;
 
 const {
   prepareMainThreadExecution,
-  markBootstrapComplete,
+  markBootstrapComplete
 } = require('internal/process/pre_execution');
-const { triggerUncaughtException } = internalBinding('errors');
+const {
+  triggerUncaughtException,
+  exitCodes: { kNoFailure },
+} = internalBinding('errors');
 const { getOptionValue } = require('internal/options');
 const { emitExperimentalWarning } = require('internal/util');
 const { FilesWatcher } = require('internal/watch_mode/files_watcher');
@@ -25,7 +27,6 @@ const { setTimeout, clearTimeout } = require('timers');
 const { resolve } = require('path');
 const { once, on } = require('events');
 
-
 prepareMainThreadExecution(false, false);
 markBootstrapComplete();
 
@@ -33,13 +34,10 @@ markBootstrapComplete();
 const kKillSignal = 'SIGTERM';
 const kShouldFilterModules = getOptionValue('--watch-path').length === 0;
 const kWatchedPaths = ArrayPrototypeMap(getOptionValue('--watch-path'), (path) => resolve(path));
-const kPreserveOutput = getOptionValue('--watch-preserve-output');
 const kCommand = ArrayPrototypeSlice(process.argv, 1);
 const kCommandStr = inspect(ArrayPrototypeJoin(kCommand, ' '));
 const args = ArrayPrototypeFilter(process.execArgv, (arg, i, arr) =>
-  !StringPrototypeStartsWith(arg, '--watch-path') &&
-  (!arr[i - 1] || !StringPrototypeStartsWith(arr[i - 1], '--watch-path')) &&
-  arg !== '--watch' && arg !== '--watch-preserve-output');
+  arg !== '--watch-path' && arr[i - 1] !== '--watch-path' && arg !== '--watch');
 ArrayPrototypePushApply(args, kCommand);
 
 const watcher = new FilesWatcher({ throttle: 500, mode: kShouldFilterModules ? 'filter' : 'all' });
@@ -51,7 +49,7 @@ let exited;
 
 function start() {
   exited = false;
-  const stdio = kShouldFilterModules ? ['inherit', 'inherit', 'inherit', 'ipc'] : 'inherit';
+  const stdio = kShouldFilterModules ? ['inherit', 'inherit', 'inherit', 'ipc'] : undefined;
   child = spawn(process.execPath, args, { stdio, env: { ...process.env, WATCH_REPORT_DEPENDENCIES: '1' } });
   watcher.watchChildProcessModules(child);
   child.once('exit', (code) => {
@@ -102,8 +100,7 @@ async function stop() {
 }
 
 async function restart() {
-  if (!kPreserveOutput) process.stdout.write(clear);
-  process.stdout.write(`${green}Restarting ${kCommandStr}${white}\n`);
+  process.stdout.write(`${clear}${green}Restarting ${kCommandStr}${white}\n`);
   await stop();
   start();
 }
@@ -130,7 +127,7 @@ function signalHandler(signal) {
   return async () => {
     watcher.clear();
     const exitCode = await killAndWait(signal, true);
-    process.exit(exitCode ?? 0);
+    process.exit(exitCode ?? kNoFailure);
   };
 }
 process.on('SIGTERM', signalHandler('SIGTERM'));
