@@ -8,7 +8,8 @@ const {
   ArrayPrototypePushApply,
   ArrayPrototypeSplice,
   ObjectDefineProperty,
-  PromisePrototypeCatch,
+  PromisePrototypeThen,
+  RegExpPrototypeExec,
   globalThis: { Atomics },
 } = primordials;
 
@@ -18,6 +19,8 @@ const {
   setupInspectorHooks,
   setupWarningHandler,
   setupFetch,
+  setupWebCrypto,
+  setupCustomEvent,
   setupDebugEnv,
   setupPerfHooks,
   initializeDeprecations,
@@ -28,8 +31,9 @@ const {
   initializeReport,
   initializeSourceMapsHandlers,
   loadPreloadModules,
-  setupTraceCategoryState
-} = require('internal/bootstrap/pre_execution');
+  setupTraceCategoryState,
+  markBootstrapComplete
+} = require('internal/process/pre_execution');
 
 const {
   threadId,
@@ -62,6 +66,7 @@ let debug = require('internal/util/debuglog').debuglog('worker', (fn) => {
 });
 
 const assert = require('internal/assert');
+const { exitCodes: { kGenericUserError } } = internalBinding('errors');
 
 patchProcessObject();
 setupInspectorHooks();
@@ -69,6 +74,8 @@ setupDebugEnv();
 
 setupWarningHandler();
 setupFetch();
+setupWebCrypto();
+setupCustomEvent();
 initializeSourceMapsHandlers();
 
 // Since worker threads cannot switch cwd, we do not need to
@@ -88,11 +95,13 @@ const port = getEnvMessagePort();
 if (process.env.NODE_CHANNEL_FD) {
   const workerThreadSetup = require('internal/process/worker_thread_only');
   ObjectDefineProperty(process, 'channel', {
+    __proto__: null,
     enumerable: false,
     get: workerThreadSetup.unavailable('process.channel')
   });
 
   ObjectDefineProperty(process, 'connected', {
+    __proto__: null,
     enumerable: false,
     get: workerThreadSetup.unavailable('process.connected')
   });
@@ -126,6 +135,9 @@ port.on('message', (message) => {
     }
     initializeDeprecations();
     initializeWASI();
+
+    require('internal/dns/utils').initializeDns();
+
     initializeCJSLoader();
     initializeESMLoader();
 
@@ -162,7 +174,7 @@ port.on('message', (message) => {
       process.stdin.push(null);
 
     debug(`[${threadId}] starts worker script ${filename} ` +
-          `(eval = ${eval}) at cwd = ${process.cwd()}`);
+          `(eval = ${doEval}) at cwd = ${process.cwd()}`);
     port.postMessage({ type: UP_AND_RUNNING });
     if (doEval === 'classic') {
       const { evalScript } = require('internal/process/execution');
@@ -170,6 +182,7 @@ port.on('message', (message) => {
       // This is necessary for CJS module compilation.
       // TODO: pass this with something really internal.
       ObjectDefineProperty(process, '_eval', {
+        __proto__: null,
         configurable: true,
         enumerable: true,
         value: filename,
@@ -178,7 +191,7 @@ port.on('message', (message) => {
       evalScript(name, filename);
     } else if (doEval === 'module') {
       const { evalModule } = require('internal/process/execution');
-      PromisePrototypeCatch(evalModule(filename), (e) => {
+      PromisePrototypeThen(evalModule(filename), undefined, (e) => {
         workerOnGlobalUncaughtException(e, true);
       });
     } else {
@@ -222,7 +235,7 @@ function workerOnGlobalUncaughtException(error, fromPromise) {
   if (!process._exiting) {
     try {
       process._exiting = true;
-      process.exitCode = 1;
+      process.exitCode = kGenericUserError;
       if (!handlerThrew) {
         process.emit('exit', process.exitCode);
       }
@@ -259,4 +272,7 @@ process._fatalException = workerOnGlobalUncaughtException;
 
 markBootstrapComplete();
 
+// Necessary to reset RegExp statics before user code runs.
+RegExpPrototypeExec(/^/, '');
+
 port.start();
