@@ -11,7 +11,6 @@ const {
   SafeArrayIterator,
   SafeMap,
   SafeSet,
-  StringPrototypeIncludes,
   StringPrototypeReplaceAll,
   StringPrototypeSlice,
   StringPrototypeStartsWith,
@@ -19,13 +18,9 @@ const {
   globalThis: { WebAssembly },
 } = primordials;
 
-/** @type {import('internal/util/types')} */
 let _TYPES = null;
-/**
- * Lazily loads and returns the internal/util/types module.
- */
 function lazyTypes() {
-  if (_TYPES !== null) { return _TYPES; }
+  if (_TYPES !== null) return _TYPES;
   return _TYPES = require('internal/util/types');
 }
 
@@ -35,33 +30,31 @@ const {
   hasEsmSyntax,
   loadBuiltinModule,
   stripBOM,
-} = require('internal/modules/helpers');
+} = require('internal/modules/cjs/helpers');
 const {
   Module: CJSModule,
-  cjsParseCache,
+  cjsParseCache
 } = require('internal/modules/cjs/loader');
-const { fileURLToPath, URL } = require('internal/url');
+const internalURLModule = require('internal/url');
+const createDynamicModule = require(
+  'internal/modules/esm/create_dynamic_module');
+const { fileURLToPath, URL } = require('url');
 let debug = require('internal/util/debuglog').debuglog('esm', (fn) => {
   debug = fn;
 });
 const { emitExperimentalWarning } = require('internal/util');
 const {
   ERR_UNKNOWN_BUILTIN_MODULE,
-  ERR_INVALID_RETURN_PROPERTY_VALUE,
+  ERR_INVALID_RETURN_PROPERTY_VALUE
 } = require('internal/errors').codes;
 const { maybeCacheSourceMap } = require('internal/source_map/source_map_cache');
 const moduleWrap = internalBinding('module_wrap');
 const { ModuleWrap } = moduleWrap;
 const asyncESM = require('internal/process/esm_loader');
 const { emitWarningSync } = require('internal/process/warning');
+const { TextDecoder } = require('internal/encoding');
 
-/** @type {import('deps/cjs-module-lexer/lexer.js').parse} */
 let cjsParse;
-/**
- * Initializes the CommonJS module lexer parser.
- * If WebAssembly is available, it uses the optimized version from the dist folder.
- * Otherwise, it falls back to the JavaScript version from the lexer folder.
- */
 async function initCJSParse() {
   if (typeof WebAssembly === 'undefined') {
     cjsParse = require('internal/deps/cjs-module-lexer/lexer').parse;
@@ -82,14 +75,6 @@ exports.translators = translators;
 exports.enrichCJSError = enrichCJSError;
 
 let DECODER = null;
-/**
- * Asserts that the given body is a buffer source (either a string, array buffer, or typed array).
- * Throws an error if the body is not a buffer source.
- * @param {string | ArrayBufferView | ArrayBuffer} body - The body to check.
- * @param {boolean} allowString - Whether or not to allow a string as a valid buffer source.
- * @param {string} hookName - The name of the hook being called.
- * @throws {ERR_INVALID_RETURN_PROPERTY_VALUE} If the body is not a buffer source.
- */
 function assertBufferSource(body, allowString, hookName) {
   if (allowString && typeof body === 'string') {
     return;
@@ -102,27 +87,17 @@ function assertBufferSource(body, allowString, hookName) {
     `${allowString ? 'string, ' : ''}array buffer, or typed array`,
     hookName,
     'source',
-    body,
+    body
   );
 }
 
-/**
- * Converts a buffer or buffer-like object to a string.
- * @param {string | ArrayBuffer | ArrayBufferView} body - The buffer or buffer-like object to convert to a string.
- * @returns {string} The resulting string.
- */
 function stringify(body) {
-  if (typeof body === 'string') { return body; }
-  assertBufferSource(body, false, 'load');
-  const { TextDecoder } = require('internal/encoding');
+  if (typeof body === 'string') return body;
+  assertBufferSource(body, false, 'transformSource');
   DECODER = DECODER === null ? new TextDecoder() : DECODER;
   return DECODER.decode(body);
 }
 
-/**
- * Converts a URL to a file path if the URL protocol is 'file:'.
- * @param {string} url - The URL to convert.
- */
 function errPath(url) {
   const parsed = new URL(url);
   if (parsed.protocol === 'file:') {
@@ -131,16 +106,8 @@ function errPath(url) {
   return url;
 }
 
-/**
- * Dynamically imports a module using the ESM loader.
- * @param {string} specifier - The module specifier to import.
- * @param {object} options - An object containing options for the import.
- * @param {string} options.url - The URL of the module requesting the import.
- * @param {Record<string, string>} [attributes] - An object containing attributes for the import.
- * @returns {Promise<import('internal/modules/esm/loader.js').ModuleExports>} The imported module.
- */
-async function importModuleDynamically(specifier, { url }, attributes) {
-  return asyncESM.esmLoader.import(specifier, url, attributes);
+async function importModuleDynamically(specifier, { url }, assertions) {
+  return asyncESM.esmLoader.import(specifier, url, assertions);
 }
 
 // Strategy for loading a standard JavaScript module.
@@ -150,9 +117,7 @@ translators.set('module', async function moduleStrategy(url, source, isMain) {
   maybeCacheSourceMap(url, source);
   debug(`Translating StandardModule ${url}`);
   const module = new ModuleWrap(url, undefined, source, 0, 0);
-  const { registerModule } = require('internal/modules/esm/utils');
-  registerModule(module, {
-    __proto__: null,
+  moduleWrap.callbackMap.set(module, {
     initializeImportMeta: (meta, wrap) => this.importMetaInitialize(meta, { url }),
     importModuleDynamically,
   });
@@ -160,7 +125,6 @@ translators.set('module', async function moduleStrategy(url, source, isMain) {
 });
 
 /**
- * Provide a more informative error for CommonJS imports.
  * @param {Error | any} err
  * @param {string} [content] Content of the file, if known.
  * @param {string} [filename] Useful only if `content` is unknown.
@@ -173,7 +137,7 @@ function enrichCJSError(err, content, filename) {
     // asynchronous warning would be emitted.
     emitWarningSync(
       'To load an ES module, set "type": "module" in the package.json or use ' +
-      'the .mjs extension.',
+      'the .mjs extension.'
     );
   }
 }
@@ -184,9 +148,11 @@ translators.set('commonjs', async function commonjsStrategy(url, source,
                                                             isMain) {
   debug(`Translating CJSModule ${url}`);
 
-  const filename = fileURLToPath(new URL(url));
+  let filename = internalURLModule.fileURLToPath(new URL(url));
+  if (isWindows)
+    filename = StringPrototypeReplaceAll(filename, '/', '\\');
 
-  if (!cjsParse) { await initCJSParse(); }
+  if (!cjsParse) await initCJSParse();
   const { module, exportNames } = cjsPreparseModuleExports(filename);
   const namesWithDefault = exportNames.has('default') ?
     [...exportNames] : ['default', ...exportNames];
@@ -209,9 +175,8 @@ translators.set('commonjs', async function commonjsStrategy(url, source,
 
     for (const exportName of exportNames) {
       if (!ObjectPrototypeHasOwnProperty(exports, exportName) ||
-          exportName === 'default') {
+          exportName === 'default')
         continue;
-      }
       // We might trigger a getter -> dont fail.
       let value;
       try {
@@ -225,17 +190,12 @@ translators.set('commonjs', async function commonjsStrategy(url, source,
   });
 });
 
-/**
- * Pre-parses a CommonJS module's exports and re-exports.
- * @param {string} filename - The filename of the module.
- */
 function cjsPreparseModuleExports(filename) {
   let module = CJSModule._cache[filename];
   if (module) {
     const cached = cjsParseCache.get(module);
-    if (cached) {
+    if (cached)
       return { module, exportNames: cached.exportNames };
-    }
   }
   const loaded = Boolean(module);
   if (!loaded) {
@@ -280,9 +240,8 @@ function cjsPreparseModuleExports(filename) {
     if ((ext === '.js' || ext === '.cjs' || !CJSModule._extensions[ext]) &&
         isAbsolute(resolved)) {
       const { exportNames: reexportNames } = cjsPreparseModuleExports(resolved);
-      for (const name of reexportNames) {
+      for (const name of reexportNames)
         exportNames.add(name);
-      }
     }
   });
 
@@ -304,17 +263,15 @@ translators.set('builtin', async function builtinStrategy(url) {
 });
 
 // Strategy for loading a JSON file
-translators.set('json', function jsonStrategy(url, source) {
+translators.set('json', async function jsonStrategy(url, source) {
+  emitExperimentalWarning('Importing JSON modules');
   assertBufferSource(source, true, 'load');
   debug(`Loading JSONModule ${url}`);
   const pathname = StringPrototypeStartsWith(url, 'file:') ?
     fileURLToPath(url) : null;
-  const shouldCheckAndPopulateCJSModuleCache =
-    // We want to involve the CJS loader cache only for `file:` URL with no search query and no hash.
-    pathname && !StringPrototypeIncludes(url, '?') && !StringPrototypeIncludes(url, '#');
   let modulePath;
   let module;
-  if (shouldCheckAndPopulateCJSModuleCache) {
+  if (pathname) {
     modulePath = isWindows ?
       StringPrototypeReplaceAll(pathname, '/', '\\') : pathname;
     module = CJSModule._cache[modulePath];
@@ -326,7 +283,7 @@ translators.set('json', function jsonStrategy(url, source) {
     }
   }
   source = stringify(source);
-  if (shouldCheckAndPopulateCJSModuleCache) {
+  if (pathname) {
     // A require call could have been called on the same file during loading and
     // that resolves synchronously. To make sure we always return the identical
     // export, we have to check again if the module already exists or not.
@@ -342,7 +299,7 @@ translators.set('json', function jsonStrategy(url, source) {
     const exports = JSONParse(stripBOM(source));
     module = {
       exports,
-      loaded: true,
+      loaded: true
     };
   } catch (err) {
     // TODO (BridgeAR): We could add a NodeCore error that wraps the JSON
@@ -352,7 +309,7 @@ translators.set('json', function jsonStrategy(url, source) {
     err.message = errPath(url) + ': ' + err.message;
     throw err;
   }
-  if (shouldCheckAndPopulateCJSModuleCache) {
+  if (pathname) {
     CJSModule._cache[modulePath] = module;
   }
   return new ModuleWrap(url, undefined, ['default'], function() {
@@ -384,12 +341,9 @@ translators.set('wasm', async function(url, source) {
     ArrayPrototypeMap(WebAssembly.Module.exports(compiled),
                       ({ name }) => name);
 
-  const createDynamicModule = require(
-    'internal/modules/esm/create_dynamic_module');
   return createDynamicModule(imports, exports, url, (reflect) => {
     const { exports } = new WebAssembly.Instance(compiled, reflect.imports);
-    for (const expt of ObjectKeys(exports)) {
+    for (const expt of ObjectKeys(exports))
       reflect.exports[expt].set(exports[expt]);
-    }
   }).module;
 });
