@@ -19,12 +19,13 @@ const {
 const { reconnectZeroFillToggle } = require('internal/buffer');
 const {
   defineOperation,
+  emitExperimentalWarning,
   exposeInterface,
-  setupCoverageHooks,
 } = require('internal/util');
 
 const {
   ERR_MANIFEST_ASSERT_INTEGRITY,
+  ERR_NO_CRYPTO,
 } = require('internal/errors').codes;
 const assert = require('internal/assert');
 
@@ -33,26 +34,12 @@ const {
   isBuildingSnapshot,
 } = require('v8').startupSnapshot;
 
-function prepareMainThreadExecution(expandArgv1 = false, initializeModules = true) {
-  prepareExecution({
-    expandArgv1,
-    initializeModules,
-    isMainThread: true,
-  });
-}
-
-function prepareWorkerThreadExecution() {
-  prepareExecution({
-    expandArgv1: false,
-    initializeModules: false,  // Will need to initialize it after policy setup
-    isMainThread: false,
-  });
-}
-
-function prepareExecution(options) {
-  const { expandArgv1, initializeModules, isMainThread } = options;
-
+function prepareMainThreadExecution(expandArgv1 = false,
+                                    initialzeModules = true) {
   refreshRuntimeOptions();
+
+  // TODO(joyeecheung): this is also necessary for workers when they deserialize
+  // this toggle from the snapshot.
   reconnectZeroFillToggle();
 
   // Patch the process object with legacy properties and normalizations
@@ -64,59 +51,58 @@ function prepareExecution(options) {
   setupFetch();
   setupWebCrypto();
   setupCustomEvent();
-  setupCodeCoverage();
+
+  // Resolve the coverage directory to an absolute path, and
+  // overwrite process.env so that the original path gets passed
+  // to child processes even when they switch cwd.
+  if (process.env.NODE_V8_COVERAGE) {
+    process.env.NODE_V8_COVERAGE =
+      setupCoverageHooks(process.env.NODE_V8_COVERAGE);
+  }
+
   setupDebugEnv();
+
+  // Print stack trace on `SIGINT` if option `--trace-sigint` presents.
+  setupStacktracePrinterOnSigint();
+
   // Process initial diagnostic reporting configuration, if present.
   initializeReport();
+  initializeReportSignalHandlers();  // Main-thread-only.
+
+  initializeHeapSnapshotSignalHandlers();
+
+  // If the process is spawned with env NODE_CHANNEL_FD, it's probably
+  // spawned by our child_process module, then initialize IPC.
+  // This attaches some internal event listeners and creates:
+  // process.send(), process.channel, process.connected,
+  // process.disconnect().
+  setupChildProcessIpcChannel();
+
+  // Load policy from disk and parse it.
+  initializePolicy();
+
+  // If this is a worker in cluster mode, start up the communication
+  // channel. This needs to be done before any user code gets executed
+  // (including preload modules).
+  initializeClusterIPC();
+
   initializeSourceMapsHandlers();
   initializeDeprecations();
   initializeWASI();
+
   require('internal/dns/utils').initializeDns();
 
-  if (isMainThread) {
-    assert(internalBinding('worker').isMainThread);
-    // Worker threads will get the manifest in the message handler.
-    const policy = readPolicyFromDisk();
-    if (policy) {
-      require('internal/process/policy')
-        .setup(policy.manifestSrc, policy.manifestURL);
-    }
+  require('internal/v8/startup_snapshot').runDeserializeCallbacks();
 
-    // Print stack trace on `SIGINT` if option `--trace-sigint` presents.
-    setupStacktracePrinterOnSigint();
-    initializeReportSignalHandlers();  // Main-thread-only.
-    initializeHeapSnapshotSignalHandlers();
-    // If the process is spawned with env NODE_CHANNEL_FD, it's probably
-    // spawned by our child_process module, then initialize IPC.
-    // This attaches some internal event listeners and creates:
-    // process.send(), process.channel, process.connected,
-    // process.disconnect().
-    setupChildProcessIpcChannel();
-    // If this is a worker in cluster mode, start up the communication
-    // channel. This needs to be done before any user code gets executed
-    // (including preload modules).
-    initializeClusterIPC();
-
-    // TODO(joyeecheung): do this for worker threads as well.
-    require('internal/v8/startup_snapshot').runDeserializeCallbacks();
-  } else {
-    assert(!internalBinding('worker').isMainThread);
-    // The setup should be called in LOAD_SCRIPT message handler.
-    assert(!initializeModules);
+  if (!initialzeModules) {
+    return;
   }
 
-  if (initializeModules) {
-    setupUserModules();
-  }
-}
-
-function setupUserModules() {
   initializeCJSLoader();
   initializeESMLoader();
   const CJSLoader = require('internal/modules/cjs/loader');
   assert(!CJSLoader.hasLoadedAnyUserCJSModule);
   loadPreloadModules();
-  // Need to be done after --require setup.
   initializeFrozenIntrinsics();
 }
 
@@ -135,7 +121,7 @@ function patchProcessObject(expandArgv1) {
     enumerable: true,
     // Only set it to true during snapshot building.
     configurable: getOptionValue('--build-snapshot'),
-    value: process.argv[0],
+    value: process.argv[0]
   });
 
   process.exitCode = undefined;
@@ -184,7 +170,7 @@ function addReadOnlyProcessAlias(name, option, enumerable = true) {
       writable: false,
       configurable: true,
       enumerable,
-      value,
+      value
     });
   }
 }
@@ -192,7 +178,7 @@ function addReadOnlyProcessAlias(name, option, enumerable = true) {
 function setupWarningHandler() {
   const {
     onWarning,
-    resetForSerialization,
+    resetForSerialization
   } = require('internal/process/warning');
   if (getOptionValue('--warnings') &&
     process.env.NODE_NO_WARNINGS !== '1') {
@@ -227,6 +213,7 @@ function setupFetch() {
   }
 
   async function fetch(input, init = undefined) {
+    emitExperimentalWarning('The Fetch API');
     return lazyUndici().fetch(input, init);
   }
 
@@ -241,7 +228,7 @@ function setupFetch() {
       },
       set(value) {
         exposeInterface(globalThis, name, value);
-      },
+      }
     };
   }
 
@@ -261,36 +248,29 @@ function setupFetch() {
 //               removed.
 function setupWebCrypto() {
   if (process.config.variables.node_no_browser_globals ||
-      !getOptionValue('--experimental-global-webcrypto')) {
+      getOptionValue('--no-experimental-global-webcrypto')) {
     return;
   }
 
-  let webcrypto;
-  ObjectDefineProperty(globalThis, 'crypto',
-                       { __proto__: null, ...ObjectGetOwnPropertyDescriptor({
-                         get crypto() {
-                           webcrypto ??= require('internal/crypto/webcrypto');
-                           return webcrypto.crypto;
-                         },
-                       }, 'crypto') });
   if (internalBinding('config').hasOpenSSL) {
-    webcrypto ??= require('internal/crypto/webcrypto');
+    const webcrypto = require('internal/crypto/webcrypto');
+    ObjectDefineProperty(globalThis, 'crypto',
+                         { __proto__: null, ...ObjectGetOwnPropertyDescriptor({
+                           get crypto() {
+                             return webcrypto.crypto;
+                           }
+                         }, 'crypto') });
     exposeInterface(globalThis, 'Crypto', webcrypto.Crypto);
     exposeInterface(globalThis, 'CryptoKey', webcrypto.CryptoKey);
     exposeInterface(globalThis, 'SubtleCrypto', webcrypto.SubtleCrypto);
-  }
-}
+  } else {
+    ObjectDefineProperty(globalThis, 'crypto',
+                         { __proto__: null, ...ObjectGetOwnPropertyDescriptor({
+                           get crypto() {
+                             throw new ERR_NO_CRYPTO();
+                           }
+                         }, 'crypto') });
 
-function setupCodeCoverage() {
-  // Resolve the coverage directory to an absolute path, and
-  // overwrite process.env so that the original path gets passed
-  // to child processes even when they switch cwd. Don't do anything if the
-  // --experimental-test-coverage flag is present, as the test runner will
-  // handle coverage.
-  if (process.env.NODE_V8_COVERAGE &&
-      !getOptionValue('--experimental-test-coverage')) {
-    process.env.NODE_V8_COVERAGE =
-      setupCoverageHooks(process.env.NODE_V8_COVERAGE);
   }
 }
 
@@ -298,13 +278,34 @@ function setupCodeCoverage() {
 //                removed.
 function setupCustomEvent() {
   if (process.config.variables.node_no_browser_globals ||
-      !getOptionValue('--experimental-global-customevent')) {
+      getOptionValue('--no-experimental-global-customevent')) {
     return;
   }
   const { CustomEvent } = require('internal/event_target');
   exposeInterface(globalThis, 'CustomEvent', CustomEvent);
 }
 
+// Setup User-facing NODE_V8_COVERAGE environment variable that writes
+// ScriptCoverage to a specified file.
+function setupCoverageHooks(dir) {
+  const cwd = require('internal/process/execution').tryGetCwd();
+  const { resolve } = require('path');
+  const coverageDirectory = resolve(cwd, dir);
+  const { sourceMapCacheToObject } =
+    require('internal/source_map/source_map_cache');
+
+  if (process.features.inspector) {
+    internalBinding('profiler').setCoverageDirectory(coverageDirectory);
+    internalBinding('profiler').setSourceMapCacheGetter(sourceMapCacheToObject);
+  } else {
+    process.emitWarning('The inspector is disabled, ' +
+                        'coverage could not be collected',
+                        'Warning');
+    return '';
+  }
+  return coverageDirectory;
+}
+
 function setupStacktracePrinterOnSigint() {
   if (!getOptionValue('--trace-sigint')) {
     return;
@@ -323,7 +324,7 @@ function initializeReport() {
     configurable: true,
     get() {
       return report;
-    },
+    }
   });
 }
 
@@ -371,7 +372,6 @@ function setupTraceCategoryState() {
 }
 
 function setupPerfHooks() {
-  require('internal/perf/performance').refreshTimeOrigin();
   require('internal/perf/utils').refreshTimeOrigin();
 }
 
@@ -384,7 +384,7 @@ function setupInspectorHooks() {
   if (internalBinding('config').hasInspector) {
     const {
       enable,
-      disable,
+      disable
     } = require('internal/inspector_async_hook');
     internalBinding('inspector').registerAsyncHook(enable, disable);
   }
@@ -439,7 +439,7 @@ function initializeDeprecations() {
       writable: false,
       enumerable: true,
       configurable: true,
-      value: noBrowserGlobals,
+      value: noBrowserGlobals
     });
   }
 
@@ -482,7 +482,7 @@ function initializeClusterIPC() {
   }
 }
 
-function readPolicyFromDisk() {
+function initializePolicy() {
   const experimentalPolicy = getOptionValue('--experimental-policy');
   if (experimentalPolicy) {
     process.emitWarning('Policies are experimental.',
@@ -510,7 +510,7 @@ function readPolicyFromDisk() {
       for (let i = 0; i < integrityEntries.length; i++) {
         const {
           algorithm,
-          value: expected,
+          value: expected
         } = integrityEntries[i];
         const hash = createHash(algorithm);
         hash.update(src);
@@ -526,9 +526,8 @@ function readPolicyFromDisk() {
         throw new ERR_MANIFEST_ASSERT_INTEGRITY(manifestURL, realIntegrities);
       }
     }
-    return {
-      manifestSrc: src, manifestURL: manifestURL.href,
-    };
+    require('internal/process/policy')
+      .setup(src, manifestURL.href);
   }
 }
 
@@ -557,7 +556,7 @@ function initializeESMLoader() {
 
   const {
     setImportModuleDynamicallyCallback,
-    setInitializeImportMetaObjectCallback,
+    setInitializeImportMetaObjectCallback
   } = internalBinding('module_wrap');
   const esm = require('internal/process/esm_loader');
   // Setup per-isolate callbacks that locate data or callbacks that we keep
@@ -579,11 +578,9 @@ function initializeESMLoader() {
 }
 
 function initializeSourceMapsHandlers() {
-  const { setSourceMapsEnabled, getSourceMapsEnabled } =
+  const { setSourceMapsEnabled } =
     require('internal/source_map/source_map_cache');
   process.setSourceMapsEnabled = setSourceMapsEnabled;
-  // Initialize the environment flag of source maps.
-  getSourceMapsEnabled();
 }
 
 function initializeFrozenIntrinsics() {
@@ -600,7 +597,7 @@ function loadPreloadModules() {
   if (preloadModules && preloadModules.length > 0) {
     const {
       Module: {
-        _preloadModules,
+        _preloadModules
       },
     } = require('internal/modules/cjs/loader');
     _preloadModules(preloadModules);
@@ -612,8 +609,25 @@ function markBootstrapComplete() {
 }
 
 module.exports = {
-  setupUserModules,
+  refreshRuntimeOptions,
+  patchProcessObject,
+  setupCoverageHooks,
+  setupWarningHandler,
+  setupFetch,
+  setupWebCrypto,
+  setupCustomEvent,
+  setupDebugEnv,
+  setupPerfHooks,
   prepareMainThreadExecution,
-  prepareWorkerThreadExecution,
-  markBootstrapComplete,
+  initializeDeprecations,
+  initializeESMLoader,
+  initializeFrozenIntrinsics,
+  initializeSourceMapsHandlers,
+  loadPreloadModules,
+  setupTraceCategoryState,
+  setupInspectorHooks,
+  initializeReport,
+  initializeCJSLoader,
+  initializeWASI,
+  markBootstrapComplete
 };
