@@ -211,6 +211,7 @@ WritableState.prototype.getBuffer = function getBuffer() {
 };
 
 ObjectDefineProperty(WritableState.prototype, 'bufferedRequestCount', {
+  __proto__: null,
   get() {
     return this.buffered.length - this.bufferedIndex;
   }
@@ -268,6 +269,7 @@ function Writable(options) {
 }
 
 ObjectDefineProperty(Writable, SymbolHasInstance, {
+  __proto__: null,
   value: function(object) {
     if (FunctionPrototypeSymbolHasInstance(this, object)) return true;
     if (this !== Writable) return false;
@@ -495,7 +497,7 @@ function afterWrite(stream, state, count, cb) {
 
   while (count-- > 0) {
     state.pendingcb--;
-    cb();
+    cb(null);
   }
 
   if (state.destroyed) {
@@ -638,8 +640,10 @@ Writable.prototype.end = function(chunk, encoding, cb) {
   }
 
   if (typeof cb === 'function') {
-    if (err || state.finished) {
+    if (err) {
       process.nextTick(cb, err);
+    } else if (state.finished) {
+      process.nextTick(cb, null);
     } else {
       state[kOnFinished].push(cb);
     }
@@ -740,7 +744,7 @@ function finish(stream, state) {
 
   const onfinishCallbacks = state[kOnFinished].splice(0);
   for (let i = 0; i < onfinishCallbacks.length; i++) {
-    onfinishCallbacks[i]();
+    onfinishCallbacks[i](null);
   }
 
   stream.emit('finish');
@@ -764,12 +768,14 @@ function finish(stream, state) {
 ObjectDefineProperties(Writable.prototype, {
 
   closed: {
+    __proto__: null,
     get() {
       return this._writableState ? this._writableState.closed : false;
     }
   },
 
   destroyed: {
+    __proto__: null,
     get() {
       return this._writableState ? this._writableState.destroyed : false;
     },
@@ -782,6 +788,7 @@ ObjectDefineProperties(Writable.prototype, {
   },
 
   writable: {
+    __proto__: null,
     get() {
       const w = this._writableState;
       // w.writable === false means that this is part of a Duplex stream
@@ -800,30 +807,35 @@ ObjectDefineProperties(Writable.prototype, {
   },
 
   writableFinished: {
+    __proto__: null,
     get() {
       return this._writableState ? this._writableState.finished : false;
     }
   },
 
   writableObjectMode: {
+    __proto__: null,
     get() {
       return this._writableState ? this._writableState.objectMode : false;
     }
   },
 
   writableBuffer: {
+    __proto__: null,
     get() {
       return this._writableState && this._writableState.getBuffer();
     }
   },
 
   writableEnded: {
+    __proto__: null,
     get() {
       return this._writableState ? this._writableState.ending : false;
     }
   },
 
   writableNeedDrain: {
+    __proto__: null,
     get() {
       const wState = this._writableState;
       if (!wState) return false;
@@ -832,24 +844,28 @@ ObjectDefineProperties(Writable.prototype, {
   },
 
   writableHighWaterMark: {
+    __proto__: null,
     get() {
       return this._writableState && this._writableState.highWaterMark;
     }
   },
 
   writableCorked: {
+    __proto__: null,
     get() {
       return this._writableState ? this._writableState.corked : 0;
     }
   },
 
   writableLength: {
+    __proto__: null,
     get() {
       return this._writableState && this._writableState.length;
     }
   },
 
   errored: {
+    __proto__: null,
     enumerable: false,
     get() {
       return this._writableState ? this._writableState.errored : null;
@@ -857,6 +873,7 @@ ObjectDefineProperties(Writable.prototype, {
   },
 
   writableAborted: {
+    __proto__: null,
     enumerable: false,
     get: function() {
       return !!(
