@@ -39,7 +39,6 @@ using v8::Array;
 using v8::ArrayBuffer;
 using v8::ArrayBufferView;
 using v8::BackingStore;
-using v8::Boolean;
 using v8::Context;
 using v8::DontDelete;
 using v8::Exception;
@@ -58,6 +57,7 @@ using v8::PropertyAttribute;
 using v8::ReadOnly;
 using v8::Signature;
 using v8::String;
+using v8::True;
 using v8::Uint32;
 using v8::Value;
 
@@ -96,14 +96,15 @@ void OnClientHello(
 
   if ((buf.IsEmpty() ||
        hello_obj->Set(env->context(), env->session_id_string(), buf)
-           .IsNothing()) ||
+          .IsNothing()) ||
       hello_obj->Set(env->context(), env->servername_string(), servername)
           .IsNothing() ||
-      hello_obj
-          ->Set(env->context(),
-                env->tls_ticket_string(),
-                Boolean::New(env->isolate(), hello.has_ticket()))
-          .IsNothing()) {
+      hello_obj->Set(
+          env->context(),
+          env->tls_ticket_string(),
+          hello.has_ticket()
+              ? True(env->isolate())
+              : False(env->isolate())).IsNothing()) {
     return;
   }
 
@@ -201,8 +202,9 @@ int SSLCertCallback(SSL* s, void* arg) {
       ? String::Empty(env->isolate())
       : OneByteString(env->isolate(), servername, strlen(servername));
 
-  Local<Value> ocsp = Boolean::New(
-      env->isolate(), SSL_get_tlsext_status_type(s) == TLSEXT_STATUSTYPE_ocsp);
+  Local<Value> ocsp = (SSL_get_tlsext_status_type(s) == TLSEXT_STATUSTYPE_ocsp)
+      ? True(env->isolate())
+      : False(env->isolate());
 
   if (info->Set(env->context(), env->servername_string(), servername_str)
           .IsNothing() ||
@@ -223,25 +225,34 @@ int SelectALPNCallback(
     const unsigned char* in,
     unsigned int inlen,
     void* arg) {
-  TLSWrap* w = static_cast<TLSWrap*>(arg);
-  const std::vector<unsigned char>& alpn_protos = w->alpn_protos_;
+  TLSWrap* w = static_cast<TLSWrap*>(SSL_get_app_data(s));
+  Environment* env = w->env();
+  HandleScope handle_scope(env->isolate());
+  Context::Scope context_scope(env->context());
 
-  if (alpn_protos.empty()) return SSL_TLSEXT_ERR_NOACK;
+  Local<Value> alpn_buffer =
+      w->object()->GetPrivate(
+          env->context(),
+          env->alpn_buffer_private_symbol()).FromMaybe(Local<Value>());
+  if (UNLIKELY(alpn_buffer.IsEmpty()) || !alpn_buffer->IsArrayBufferView())
+    return SSL_TLSEXT_ERR_NOACK;
 
-  int status = SSL_select_next_proto(const_cast<unsigned char**>(out),
-                                     outlen,
-                                     alpn_protos.data(),
-                                     alpn_protos.size(),
-                                     in,
-                                     inlen);
+  ArrayBufferViewContents<unsigned char> alpn_protos(alpn_buffer);
+  int status = SSL_select_next_proto(
+      const_cast<unsigned char**>(out),
+      outlen,
+      alpn_protos.data(),
+      alpn_protos.length(),
+      in,
+      inlen);
 
-  // According to 3.2. Protocol Selection of RFC7301, fatal
-  // no_application_protocol alert shall be sent but OpenSSL 1.0.2 does not
-  // support it yet. See
-  // https://rt.openssl.org/Ticket/Display.html?id=3463&user=guest&pass=guest
-  return status == OPENSSL_NPN_NEGOTIATED
-      ? SSL_TLSEXT_ERR_OK
-      : SSL_TLSEXT_ERR_NOACK;
+  // Previous versions of Node.js returned SSL_TLSEXT_ERR_NOACK if no protocol
+  // match was found. This would neither cause a fatal alert nor would it result
+  // in a useful ALPN response as part of the Server Hello message.
+  // We now return SSL_TLSEXT_ERR_ALERT_FATAL in that case as per Section 3.2
+  // of RFC 7301, which causes a fatal no_application_protocol alert.
+  return status == OPENSSL_NPN_NEGOTIATED ? SSL_TLSEXT_ERR_OK
+                                          : SSL_TLSEXT_ERR_ALERT_FATAL;
 }
 
 int TLSExtStatusCallback(SSL* s, void* arg) {
@@ -319,15 +330,12 @@ TLSWrap::TLSWrap(Environment* env,
                  Local<Object> obj,
                  Kind kind,
                  StreamBase* stream,
-                 SecureContext* sc,
-                 UnderlyingStreamWriteStatus under_stream_ws)
+                 SecureContext* sc)
     : AsyncWrap(env, obj, AsyncWrap::PROVIDER_TLSWRAP),
       StreamBase(env),
       env_(env),
       kind_(kind),
-      sc_(sc),
-      has_active_write_issued_by_prev_listener_(
-          under_stream_ws == UnderlyingStreamWriteStatus::kHasActive) {
+      sc_(sc) {
   MakeWeak();
   CHECK(sc_);
   ssl_ = sc_->CreateSSL();
@@ -437,19 +445,14 @@ void TLSWrap::InitSSL() {
 void TLSWrap::Wrap(const FunctionCallbackInfo<Value>& args) {
   Environment* env = Environment::GetCurrent(args);
 
-  CHECK_EQ(args.Length(), 4);
+  CHECK_EQ(args.Length(), 3);
   CHECK(args[0]->IsObject());
   CHECK(args[1]->IsObject());
   CHECK(args[2]->IsBoolean());
-  CHECK(args[3]->IsBoolean());
 
   Local<Object> sc = args[1].As<Object>();
   Kind kind = args[2]->IsTrue() ? Kind::kServer : Kind::kClient;
 
-  UnderlyingStreamWriteStatus under_stream_ws =
-      args[3]->IsTrue() ? UnderlyingStreamWriteStatus::kHasActive
-                        : UnderlyingStreamWriteStatus::kVacancy;
-
   StreamBase* stream = StreamBase::FromObject(args[0].As<Object>());
   CHECK_NOT_NULL(stream);
 
@@ -460,8 +463,7 @@ void TLSWrap::Wrap(const FunctionCallbackInfo<Value>& args) {
     return;
   }
 
-  TLSWrap* res = new TLSWrap(
-      env, obj, kind, stream, Unwrap<SecureContext>(sc), under_stream_ws);
+  TLSWrap* res = new TLSWrap(env, obj, kind, stream, Unwrap<SecureContext>(sc));
 
   args.GetReturnValue().Set(res->object());
 }
@@ -567,13 +569,6 @@ void TLSWrap::EncOut() {
     return;
   }
 
-  if (UNLIKELY(has_active_write_issued_by_prev_listener_)) {
-    Debug(this,
-          "Returning from EncOut(), "
-          "has_active_write_issued_by_prev_listener_ is true");
-    return;
-  }
-
   // Split-off queue
   if (established_ && current_write_) {
     Debug(this, "EncOut() write is scheduled");
@@ -644,15 +639,6 @@ void TLSWrap::EncOut() {
 
 void TLSWrap::OnStreamAfterWrite(WriteWrap* req_wrap, int status) {
   Debug(this, "OnStreamAfterWrite(status = %d)", status);
-
-  if (UNLIKELY(has_active_write_issued_by_prev_listener_)) {
-    Debug(this, "Notify write finish to the previous_listener_");
-    CHECK_EQ(write_size_, 0);  // we must have restrained writes
-
-    previous_listener_->OnStreamAfterWrite(req_wrap, status);
-    return;
-  }
-
   if (current_empty_write_) {
     Debug(this, "Had empty write");
     BaseObjectPtr<AsyncWrap> current_empty_write =
@@ -691,6 +677,11 @@ void TLSWrap::OnStreamAfterWrite(WriteWrap* req_wrap, int status) {
   EncOut();
 }
 
+int TLSWrap::GetSSLError(int status) const {
+  // ssl_ might already be destroyed for reading EOF from a close notify alert.
+  return ssl_ != nullptr ? SSL_get_error(ssl_.get(), status) : 0;
+}
+
 void TLSWrap::ClearOut() {
   Debug(this, "Trying to read cleartext output");
   // Ignore cycling data if ClientHello wasn't yet parsed
@@ -744,23 +735,25 @@ void TLSWrap::ClearOut() {
     }
   }
 
+  int flags = SSL_get_shutdown(ssl_.get());
+  if (!eof_ && flags & SSL_RECEIVED_SHUTDOWN) {
+    eof_ = true;
+    EmitRead(UV_EOF);
+  }
+
   // We need to check whether an error occurred or the connection was
   // shutdown cleanly (SSL_ERROR_ZERO_RETURN) even when read == 0.
-  // See node#1642 and SSL_read(3SSL) for details. SSL_get_error must be
-  // called immediately after SSL_read, without calling into JS, which may
-  // change OpenSSL's error queue, modify ssl_, or even destroy ssl_
-  // altogether.
+  // See node#1642 and SSL_read(3SSL) for details.
   if (read <= 0) {
     HandleScope handle_scope(env()->isolate());
     Local<Value> error;
-    int err = SSL_get_error(ssl_.get(), read);
+    int err = GetSSLError(read);
     switch (err) {
       case SSL_ERROR_ZERO_RETURN:
-        if (!eof_) {
-          eof_ = true;
-          EmitRead(UV_EOF);
-        }
-        return;
+        // Ignore ZERO_RETURN after EOF, it is basically not an error.
+        if (eof_) return;
+        error = env()->zero_return_string();
+        break;
 
       case SSL_ERROR_SSL:
       case SSL_ERROR_SYSCALL:
@@ -845,7 +838,7 @@ void TLSWrap::ClearIn() {
   }
 
   // Error or partial write
-  int err = SSL_get_error(ssl_.get(), written);
+  int err = GetSSLError(written);
   if (err == SSL_ERROR_SSL || err == SSL_ERROR_SYSCALL) {
     Debug(this, "Got SSL error (%d)", err);
     write_callback_scheduled_ = true;
@@ -1021,7 +1014,7 @@ int TLSWrap::DoWrite(WriteWrap* w,
 
   if (written == -1) {
     // If we stopped writing because of an error, it's fatal, discard the data.
-    int err = SSL_get_error(ssl_.get(), written);
+    int err = GetSSLError(written);
     if (err == SSL_ERROR_SSL || err == SSL_ERROR_SYSCALL) {
       // TODO(@jasnell): What are we doing with the error?
       Debug(this, "Got SSL error (%d), returning UV_EPROTO", err);
@@ -1375,7 +1368,8 @@ unsigned int TLSWrap::PskServerCallback(
 
   // Make sure there are no utf8 replacement symbols.
   Utf8Value identity_utf8(env->isolate(), identity_str);
-  if (identity_utf8 != identity) return 0;
+  if (strcmp(*identity_utf8, identity) != 0)
+    return 0;
 
   Local<Value> argv[] = {
     identity_str,
@@ -1535,14 +1529,19 @@ void TLSWrap::SetALPNProtocols(const FunctionCallbackInfo<Value>& args) {
   if (args.Length() < 1 || !Buffer::HasInstance(args[0]))
     return env->ThrowTypeError("Must give a Buffer as first argument");
 
-  ArrayBufferViewContents<uint8_t> protos(args[0].As<ArrayBufferView>());
-  SSL* ssl = w->ssl_.get();
   if (w->is_client()) {
-    CHECK_EQ(0, SSL_set_alpn_protos(ssl, protos.data(), protos.length()));
+    ArrayBufferViewContents<char> protos(args[0].As<ArrayBufferView>());
+    CHECK(SetALPN(w->ssl_, {protos.data(), protos.length()}));
   } else {
-    w->alpn_protos_ = std::vector<unsigned char>(
-        protos.data(), protos.data() + protos.length());
-    SSL_CTX_set_alpn_select_cb(SSL_get_SSL_CTX(ssl), SelectALPNCallback, w);
+    CHECK(
+        w->object()->SetPrivate(
+            env->context(),
+            env->alpn_buffer_private_symbol(),
+            args[0]).FromJust());
+    // Server should select ALPN protocol from list of advertised by client
+    SSL_CTX_set_alpn_select_cb(SSL_get_SSL_CTX(w->ssl_.get()),
+                               SelectALPNCallback,
+                               nullptr);
   }
 }
 
@@ -1999,16 +1998,6 @@ void TLSWrap::GetALPNNegotiatedProto(const FunctionCallbackInfo<Value>& args) {
   args.GetReturnValue().Set(result);
 }
 
-void TLSWrap::WritesIssuedByPrevListenerDone(
-    const FunctionCallbackInfo<Value>& args) {
-  TLSWrap* w;
-  ASSIGN_OR_RETURN_UNWRAP(&w, args.Holder());
-
-  Debug(w, "WritesIssuedByPrevListenerDone is called");
-  w->has_active_write_issued_by_prev_listener_ = false;
-  w->EncOut();  // resume all of our restrained writes
-}
-
 void TLSWrap::Cycle() {
   // Prevent recursion
   if (++cycle_depth_ > 1)
@@ -2085,10 +2074,6 @@ void TLSWrap::Initialize(
   SetProtoMethod(isolate, t, "setSession", SetSession);
   SetProtoMethod(isolate, t, "setVerifyMode", SetVerifyMode);
   SetProtoMethod(isolate, t, "start", Start);
-  SetProtoMethod(isolate,
-                 t,
-                 "writesIssuedByPrevListenerDone",
-                 WritesIssuedByPrevListenerDone);
 
   SetProtoMethodNoSideEffect(
       isolate, t, "exportKeyingMaterial", ExportKeyingMaterial);
@@ -2170,7 +2155,6 @@ void TLSWrap::RegisterExternalReferences(ExternalReferenceRegistry* registry) {
   registry->Register(GetSharedSigalgs);
   registry->Register(GetTLSTicket);
   registry->Register(VerifyError);
-  registry->Register(WritesIssuedByPrevListenerDone);
 
 #ifdef SSL_set_max_send_fragment
   registry->Register(SetMaxSendFragment);
@@ -2185,6 +2169,6 @@ void TLSWrap::RegisterExternalReferences(ExternalReferenceRegistry* registry) {
 }  // namespace crypto
 }  // namespace node
 
-NODE_BINDING_CONTEXT_AWARE_INTERNAL(tls_wrap, node::crypto::TLSWrap::Initialize)
-NODE_BINDING_EXTERNAL_REFERENCE(
+NODE_MODULE_CONTEXT_AWARE_INTERNAL(tls_wrap, node::crypto::TLSWrap::Initialize)
+NODE_MODULE_EXTERNAL_REFERENCE(
     tls_wrap, node::crypto::TLSWrap::RegisterExternalReferences)
