@@ -1,8 +1,8 @@
 #include "env-inl.h"
 #include "memory_tracker.h"
 #include "node.h"
+#include "node_builtins.h"
 #include "node_i18n.h"
-#include "node_native_module.h"
 #include "node_options.h"
 #include "util-inl.h"
 
@@ -75,13 +75,7 @@ static void Initialize(Local<Object> target,
   READONLY_FALSE_PROPERTY(target, "noBrowserGlobals");
 #endif  // NODE_NO_BROWSER_GLOBALS
 
-  READONLY_PROPERTY(target,
-                    "bits",
-                    Number::New(isolate, 8 * sizeof(intptr_t)));
-
-#if defined HAVE_DTRACE || defined HAVE_ETW
-  READONLY_TRUE_PROPERTY(target, "hasDtrace");
-#endif
+  READONLY_PROPERTY(target, "bits", Number::New(isolate, 8 * sizeof(intptr_t)));
 }  // InitConfig
 
 }  // namespace node
