@@ -5,6 +5,7 @@ const fixtures = require('../common/fixtures');
 const fs = require('fs');
 const fsPromises = fs.promises;
 const path = require('path');
+const events = require('events');
 const { inspect } = require('util');
 const { Worker } = require('worker_threads');
 
@@ -78,7 +79,7 @@ class ResourceLoader {
 }
 
 class StatusRule {
-  constructor(key, value, pattern = undefined) {
+  constructor(key, value, pattern) {
     this.key = key;
     this.requires = value.requires || [];
     this.fail = value.fail;
@@ -152,7 +153,8 @@ class WPTTestSpec {
     this.filename = filename;
 
     this.requires = new Set();
-    this.failReasons = [];
+    this.failedTests = [];
+    this.flakyTests = [];
     this.skipReasons = [];
     for (const item of rules) {
       if (item.requires.length) {
@@ -160,13 +162,21 @@ class WPTTestSpec {
           this.requires.add(req);
         }
       }
-      if (item.fail) {
-        this.failReasons.push(item.fail);
+      if (Array.isArray(item.fail?.expected)) {
+        this.failedTests.push(...item.fail.expected);
+      }
+      if (Array.isArray(item.fail?.flaky)) {
+        this.failedTests.push(...item.fail.flaky);
+        this.flakyTests.push(...item.fail.flaky);
       }
       if (item.skip) {
         this.skipReasons.push(item.skip);
       }
     }
+
+    this.failedTests = [...new Set(this.failedTests)];
+    this.flakyTests = [...new Set(this.flakyTests)];
+    this.skipReasons = [...new Set(this.skipReasons)];
   }
 
   getRelativePath() {
@@ -288,7 +298,7 @@ class WPTRunner {
     this.resource = new ResourceLoader(path);
 
     this.flags = [];
-    this.dummyGlobalThisScript = null;
+    this.globalThisInitScripts = [];
     this.initScript = null;
 
     this.status = new StatusLoader(path);
@@ -330,17 +340,17 @@ class WPTRunner {
   }
 
   get fullInitScript() {
-    if (this.initScript === null && this.dummyGlobalThisScript === null) {
-      return null;
-    }
-
-    if (this.initScript === null) {
-      return this.dummyGlobalThisScript;
-    } else if (this.dummyGlobalThisScript === null) {
+    if (this.globalThisInitScripts.length === null) {
       return this.initScript;
     }
 
-    return `${this.dummyGlobalThisScript}\n\n//===\n${this.initScript}`;
+    const globalThisInitScript = this.globalThisInitScripts.join('\n\n//===\n');
+
+    if (this.initScript === null) {
+      return globalThisInitScript;
+    }
+
+    return `${globalThisInitScript}\n\n//===\n${this.initScript}`;
   }
 
   /**
@@ -351,8 +361,9 @@ class WPTRunner {
   pretendGlobalThisAs(name) {
     switch (name) {
       case 'Window': {
-        this.dummyGlobalThisScript =
-          'global.Window = Object.getPrototypeOf(globalThis).constructor;';
+        this.globalThisInitScripts.push(
+          `global.Window = Object.getPrototypeOf(globalThis).constructor;
+          self.GLOBAL.isWorker = () => false;`);
         break;
       }
 
@@ -366,9 +377,39 @@ class WPTRunner {
     }
   }
 
+  brandCheckGlobalScopeAttribute(name) {
+    // TODO(legendecas): idlharness GlobalScope attribute receiver validation.
+    const script = `
+      const desc = Object.getOwnPropertyDescriptor(globalThis, '${name}');
+      function getter() {
+        // Mimic GlobalScope instance brand check.
+        if (this !== globalThis) {
+          throw new TypeError('Illegal invocation');
+        }
+        return desc.get();
+      }
+      Object.defineProperty(getter, 'name', { value: 'get ${name}' });
+
+      function setter(value) {
+        // Mimic GlobalScope instance brand check.
+        if (this !== globalThis) {
+          throw new TypeError('Illegal invocation');
+        }
+        desc.set(value);
+      }
+      Object.defineProperty(setter, 'name', { value: 'set ${name}' });
+
+      Object.defineProperty(globalThis, '${name}', {
+        get: getter,
+        set: setter,
+      });
+    `;
+    this.globalThisInitScripts.push(script);
+  }
+
   // TODO(joyeecheung): work with the upstream to port more tests in .html
   // to .js.
-  runJsTests() {
+  async runJsTests() {
     let queue = [];
 
     // If the tests are run as `node test/wpt/test-something.js subset.any.js`,
@@ -459,44 +500,86 @@ class WPTRunner {
         );
         this.inProgress.delete(testFileName);
       });
+
+      await events.once(worker, 'exit').catch(() => {});
     }
 
     process.on('exit', () => {
-      const total = this.specMap.size;
       if (this.inProgress.size > 0) {
         for (const filename of this.inProgress) {
           this.fail(filename, { name: 'Unknown' }, kIncomplete);
         }
       }
       inspect.defaultOptions.depth = Infinity;
-      console.log(this.results);
+      // Sorts the rules to have consistent output
+      console.log(JSON.stringify(Object.keys(this.results).sort().reduce(
+        (obj, key) => {
+          obj[key] = this.results[key];
+          return obj;
+        },
+        {}
+      ), null, 2));
 
       const failures = [];
       let expectedFailures = 0;
       let skipped = 0;
-      for (const key of Object.keys(this.results)) {
-        const item = this.results[key];
-        if (item.fail && item.fail.unexpected) {
+      for (const [key, item] of Object.entries(this.results)) {
+        if (item.fail?.unexpected) {
           failures.push(key);
         }
-        if (item.fail && item.fail.expected) {
+        if (item.fail?.expected) {
           expectedFailures++;
         }
         if (item.skip) {
           skipped++;
         }
       }
-      const ran = total - skipped;
+
+      const unexpectedPasses = [];
+      for (const specMap of queue) {
+        const key = specMap.filename;
+
+        // File has no expected failures
+        if (!specMap.failedTests.length) {
+          continue;
+        }
+
+        // File was (maybe even conditionally) skipped
+        if (this.results[key]?.skip) {
+          continue;
+        }
+
+        // Full check: every expected to fail test is present
+        if (specMap.failedTests.some((expectedToFail) => {
+          if (specMap.flakyTests.includes(expectedToFail)) {
+            return false;
+          }
+          return this.results[key]?.fail?.expected?.includes(expectedToFail) !== true;
+        })) {
+          unexpectedPasses.push(key);
+          continue;
+        }
+      }
+
+      const ran = queue.length;
+      const total = ran + skipped;
       const passed = ran - expectedFailures - failures.length;
       console.log(`Ran ${ran}/${total} tests, ${skipped} skipped,`,
                   `${passed} passed, ${expectedFailures} expected failures,`,
-                  `${failures.length} unexpected failures`);
+                  `${failures.length} unexpected failures,`,
+                  `${unexpectedPasses.length} unexpected passes`);
       if (failures.length > 0) {
         const file = path.join('test', 'wpt', 'status', `${this.path}.json`);
         throw new Error(
           `Found ${failures.length} unexpected failures. ` +
           `Consider updating ${file} for these files:\n${failures.join('\n')}`);
       }
+      if (unexpectedPasses.length > 0) {
+        const file = path.join('test', 'wpt', 'status', `${this.path}.json`);
+        throw new Error(
+          `Found ${unexpectedPasses.length} unexpected passes. ` +
+          `Consider updating ${file} for these files:\n${unexpectedPasses.join('\n')}`);
+      }
     });
   }
 
@@ -577,8 +660,9 @@ class WPTRunner {
       if (!result[item.status][key]) {
         result[item.status][key] = [];
       }
-      if (result[item.status][key].indexOf(item.reason) === -1) {
-        result[item.status][key].push(item.reason);
+      const hasName = result[item.status][key].includes(item.name);
+      if (!hasName) {
+        result[item.status][key].push(item.name);
       }
     }
   }
@@ -589,10 +673,10 @@ class WPTRunner {
 
   fail(filename, test, status) {
     const spec = this.specMap.get(filename);
-    const expected = !!(spec.failReasons.length);
+    const expected = spec.failedTests.includes(test.name);
     if (expected) {
       console.log(`[EXPECTED_FAILURE][${status.toUpperCase()}] ${test.name}`);
-      console.log(spec.failReasons.join('; '));
+      console.log(test.message || status);
     } else {
       console.log(`[UNEXPECTED_FAILURE][${status.toUpperCase()}] ${test.name}`);
     }
@@ -604,6 +688,7 @@ class WPTRunner {
                     ` ${require.main.filename} ${filename}`;
     console.log(`Command: ${command}\n`);
     this.addTestResult(filename, {
+      name: test.name,
       expected,
       status: kFail,
       reason: test.message || status
