@@ -101,3 +101,21 @@ tmpdir.refresh();
   const content = fs.readFileSync(file, { encoding: 'utf8' });
   assert.strictEqual(content, 'hello world!');
 }
+
+// Test writeFileSync with an invalid input
+{
+  const file = path.join(tmpdir.path, 'testWriteFileSyncInvalid.txt');
+  for (const data of [
+    false, 5, {}, [], null, undefined, true, 5n, () => {}, Symbol(), new Map(),
+    new String('notPrimitive'),
+    { [Symbol.toPrimitive]: (hint) => 'amObject' },
+    { toString() { return 'amObject'; } },
+    Promise.resolve('amPromise'),
+    common.mustNotCall(),
+  ]) {
+    assert.throws(
+      () => fs.writeFileSync(file, data, { encoding: 'utf8', flag: 'a' }),
+      { code: 'ERR_INVALID_ARG_TYPE' }
+    );
+  }
+}
